from vec3 import Vec3

class AABoundingBox:
    def __init__(self, minCorner, maxCorner):
        self._corner1, self._corner2 = minCorner, maxCorner
        self._centre = None
        self._infinite = False

    @classmethod
    def infinite(cls):
        box = cls(Vec3.zeroVector(), Vec3.zeroVector())
        box._infinite = True
        return box

    @classmethod
    def fromPoints(cls, points, fromFrame=None):
        points = list(points)
        if fromFrame:
            for n, point in enumerate(points):
                points[n] = fromFrame.toWorld(point)
        points.sort(key=lambda x: x[0])
        minX = points[0].x
        maxX = points[-1].x
        points.sort(key=lambda x: x[1])
        minY = points[0].y
        maxY = points[-1].y
        points.sort(key=lambda x: x[2])
        minZ = points[0].z
        maxZ = points[-1].z
        corner1 = Vec3((minX, minY, minZ))
        corner2 = Vec3((maxX, maxY, maxZ))
        return cls(corner1, corner2)

    @classmethod
    def fromAABoundingBoxes(cls, boxes):
        vertices = []
        for box in boxes:
            if box.isInfinite:
                return cls.infinite()
            vertices += box.corners
        return cls.fromPoints(vertices)

    def isHitBy(self, ray):
        if self._infinite:
            return True
        try:
            dirfrac = ray.direction.reciprocated
        except:
            return False
        sub1 = self._corner1 - ray.origin
        sub2 = self._corner2 - ray.origin
        t1 = sub1.x * dirfrac.x
        t2 = sub2.x * dirfrac.x
        t3 = sub1.y * dirfrac.y
        t4 = sub2.y * dirfrac.y
        t5 = sub1.z * dirfrac.z
        t6 = sub2.z * dirfrac.z
        tmin = max(max(min(t1, t2), min(t3, t4)), min(t5, t6))
        tmax = min(min(max(t1, t2), max(t3, t4)), max(t5, t6))
        if tmax < 0:
            return False
        if tmin > tmax:
            return False
        return True

    def isHitByNEW(self, ray, tMin=0, tMax=1e15):
        invD = ray.direction.reciprocated
        for index in range(3):
            t0 = (self._corner1[index] - ray.origin[index]) * invD[index]
            t1 = (self._corner2[index] - ray.origin[index]) * invD[index]
            if invD[index] < 0:
                tTemp = t1
                t1 = t0
                t0 = tTemp
            tMin = t0 if t0 > tMin else tMin
            tMax = t1 if t1 < tMax else tMax
            if tMax <= tMin:
                #print('failed on index', index)
                #print(ray.origin[index], invD[index])
                #print(tMin, tMax)
                return False
        return True

    @property
    def isInfinite(self):
        return self._infinite

    @property
    def corners(self):
        return (self._corner1, self._corner2)

    @property
    def centre(self):
        if self._centre:
            return self._centre
        self._centre = 0.5 * (self._corner1 + self._corner2)
        return self._centre

    @property
    def limits(self):
        return list(self._corner1.tup) + list(self._corner2.tup)