import math
import random
import numpy as np
from vec3 import Vec3
from frame import Frame
from sensor import Sensor, Polariser, Sensor_old
from ray import Ray
import copy

class Camera:
    def __init__(self, FOV, aperture, horizontalRes=None, verticalRes=None, polarisationAng=None):
        self._frame = Frame(Vec3.j().negated, Vec3.k().negated, Vec3.i(), Vec3.zeroVector())
        self._FOV = FOV
        self._aperture = aperture
        self._f = 0.1
        detectorWidth = 2 * self._inwardProjection(math.radians(FOV[0]) * 0.5)
        detectorHeight = 2 * self._inwardProjection(math.radians(FOV[1]) * 0.5)
        detectorOrigin = self.position + (self._frame.u * self._f) + 0.5 * (detectorWidth * self._frame.v + detectorHeight * self._frame.w)
        if horizontalRes is None and verticalRes is None:
            # Neither resolution is defined:
            # use default horizontal resolution
            horizontalRes = 1024
        elif verticalRes is not None and horizontalRes is not None:
            # Both resolutions have been defined and system is over constrained:
            # discard vertical resolution
            verticalRes = None
        if verticalRes is not None:
            # Vertical resolution is defined: calculate corresponding horizontal resolution
            self._pixelPitch = detectorHeight / verticalRes
            horizontalRes = int(detectorWidth / self._pixelPitch)
        else:
            # Horizontal resolution is defined: calculate corresponding vertical resolution
            self._pixelPitch = detectorWidth / horizontalRes
            verticalRes = int(detectorHeight / self._pixelPitch)
        principalPoint = self.position + (self._f * self._frame.w)
        sensorFrame = Frame(self.frame.u.negated, self.frame.v, self.frame.w.negated, principalPoint)
        self._sensor = Sensor(horizontalRes, verticalRes, detectorWidth, detectorHeight, sensorFrame, 0.3, 20000, 25, 5)
        self._pixelFractX = 1 / horizontalRes
        self._pixelFractY = 1 / verticalRes
        self._polarisationAng = polarisationAng
        self.type = 'Pinhole'

    def _inwardProjection(self, angleOfIncidence):
        if type(angleOfIncidence) is not np.ndarray:
            return self._f * math.tan(angleOfIncidence)
        return self._f * np.tan(angleOfIncidence)

    def _outwardProjection(self, magnitude):
        if type(magnitude) is not np.ndarray:
            return math.atan(magnitude / self._f)
        return np.arctan(magnitude / self._f)
 
    def _inwardRayFocalPlaneInterceptDEPRECATED(self, rayOrigin, rayVector):
        # TODO: remove deprecated function
        return self._focalPlane.intercept(rayOrigin, rayVector)

    def castRayIn(self, point, cropToImageBoundaries=True):
        # TODO: clean up
        point = self.frame.fromWorld(point).norm
        if type(point.z) is not np.ndarray:
            # scalar arithmetic required
            if point.z < 0:
                return None
            angleOfIncidence = math.acos(point.z)
            # get the point's radius from principal point on the detector
            radius = self._inwardProjection(angleOfIncidence)
            #print(radius * Vec3((point.x, point.y, 0)).norm)
            # calculate the world coordinate of the point on the detector
            point = self._frame.toWorld((radius * Vec3((point.x, point.y, 0)).norm) + Vec3((0, 0, self._f)))
            return self._sensor.fromWorldToImage(point, cropToImageBoundaries=cropToImageBoundaries)
        else:
            # numpy arithmetic required
            shape = point.z.shape
            frontField = point.z >= 0
            #print(point.z)
            #return frontField
            if not frontField.any():
                return None
            angleOfIncidence = np.arccos(point.z[frontField])
            radius = self._inwardProjection(angleOfIncidence)
            localPoint = Vec3((point.x[frontField], point.y[frontField], 0)).norm * radius + Vec3((0, 0, self._f))
            point = self._frame.toWorld(localPoint)
            components = tuple()
            for n in range(3):
                component = np.full(shape, np.nan)
                component[frontField] = point[n]
                components += (component,)
            return self._sensor.fromWorldToImage(Vec3(components), cropToImageBoundaries=cropToImageBoundaries)

    def castRayOut(self, xfract, yfract, antialias=False, clipToForwardRays=True):
        # TODO: clean up
        """Returns a normalised vector representing a ray cast out of the camera from a position on the focal plane in fractional coordinates."""
        if type(xfract) is not np.ndarray and type(yfract) is not np.ndarray:
            # scalar arithmetic required
            pixelPosition = self._sensor.fromImageToWorld(xfract, yfract)
            radiusOnDetector = pixelPosition - self._sensor.principalPoint
            angleOfIncidence = self._outwardProjection(radiusOnDetector.length)
            if angleOfIncidence > 1.5708 and clipToForwardRays:
                return None
            try:
                z = 1 / math.tan(angleOfIncidence)
                return ((z * self._frame.w) + radiusOnDetector.norm).norm
            except ZeroDivisionError:
                # angle of incidence = 0
                return self._frame.w
            except ValueError:
                print('failed on angle of incidence: {0}'.format(angleOfIncidence))
                1 / 0
        else:
            # numpy arithmetic required:
            pixelPosition = self._sensor.fromImageToWorld(xfract, yfract)
            radiusOnDetector = pixelPosition - self._sensor.principalPoint
            angleOfIncidence = self._outwardProjection(radiusOnDetector.length)
            if clipToForwardRays:
                angleOfIncidence[angleOfIncidence > 1.5708] = np.nan
            z = 1 / np.tan(angleOfIncidence)
            return ((self._frame.w * z) + radiusOnDetector.norm).norm

    def captureFlux(self, radiance):
        # TODO: add front end optics (NOT including filters) effect to this
        return self._aperture * (self.angularResolution() ** 2) * radiance

    def digitiseLight(self, lightRay, exposureTime, xfract, yfract):
        # TODO: remove this function
        #pixelAngularResolution = self.angularResolution(xfract, yfract)
        pixelAngularResolution = self.angularResolution(0.5, 0.5)
        horizontalRes, verticalRes = pixelAngularResolution
        pixelSolidAngle = 4 * math.tan(0.5 * horizontalRes) * math.tan(0.5 * verticalRes)
        photonsCaptured = lightRay.intensity * pixelSolidAngle * self._aperture * exposureTime
        if self._polarisationAng is not None:
            pol = Polariser(self.frame.rotateInPlace(self.frame.u, math.radians(self._polarisationAng)))
            photonsCaptured = pol.transmit(lightRay.mapTo(self.position, self.frame.u)) * photonsCaptured
        return self._sensor.pixelValue(photonsCaptured)

    def angularResolution(self):
        # TODO: variation of angular resolution over field of view
        return math.radians(self._FOV[0]) / self.resolution[0]

    def spatialResolution(self, atDistance, xfract=0.5, yfract=0.5):
        # TODO: clean up and match with angularResolution function
        r = atDistance
        IFOV = self.angularResolution(xfract=xfract, yfract=yfract)
        return (r * IFOV[0], r * IFOV[1])

    def areaSampled(self, atDistance, xfract=0.5, yfract=0.5):
        # TODO: clean up
        r = atDistance
        IFOV = self.angularResolution(xfract=xfract, yfract=yfract)
        return IFOV[0] * IFOV[1] * r ** 2

    def pointCloud(self, radius, scene=None, columns=None, rows=None):
        # TODO: clean up
        matrix = []
        if columns is None:
            columns = list(range(self.resolution[0]))
        if rows is None:
            rows = list(range(self.resolution[1]))
        for y in rows:
            row = []
            yfract = y / self._sensor.pixelsY
            for x in columns:
                xfract = x / self._sensor.pixelsX
                ray = Ray(self.position, self.castRayOut(xfract, yfract))
                if scene is None:
                    point = self.position + radius * ray.direction
                else:
                    sceneTraversalResult = scene.traverse(ray, returnXYZ=True)
                    if len(sceneTraversalResult) != 0:
                        point = sceneTraversalResult[0]['XYZs'][0]
                    else:
                        point = None
                row += [point]
            matrix += [row]
        return matrix

    @property
    def position(self):
        return self._frame.origin

    @position.setter
    def position(self, newPosition):
        self._frame.origin = newPosition
        self._sensor.principalPoint = newPosition + (self._f * self._frame.w)

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, newFrame):
        self._frame = newFrame
        self._sensor.frame = Frame(newFrame.u.negated, newFrame.v, newFrame.w.negated, newFrame.origin)

    @property
    def FOV(self):
        return self._FOV

    @property
    def aperture(self):
        return self._aperture

    @property
    def resolution(self):
        return self._sensor.resolution

    @property
    def sensor(self):
        return self._sensor

class EquidistanceCamera(Camera):
    def __init__(self, FOV, aperture, horizontalRes=None, verticalRes=None):
        super().__init__(FOV, aperture, horizontalRes=horizontalRes, verticalRes=verticalRes)
        self.type = "Equidistance"

    def _inwardProjection(self, angleOfIncidence):
        return self._f * angleOfIncidence

    def _outwardProjection(self, magnitude):
        return magnitude / self._f

class Pushbroom:
    def __init__(self, scanAngle, acrossTrackRes, spinPeriod, descentV=0):
        self._frame = Frame.universal()
        self._scanAngle = math.radians(scanAngle)
        self._resolution = (int((scanAngle / 90) * acrossTrackRes), acrossTrackRes)
        self.spinPeriod = spinPeriod
        self.descentV = descentV

    def castRayIn(self, point, cropToImageBoundaries=True):
        localSphericalCoords = self.frame.fromWorldToSpherical(point)
        if localSphericalCoords is None:
            return None
        if self.descentV != 0:
            _, az, _ = localSphericalCoords
            dZ = self.spinPeriod * self.descentV * az / (2 * math.pi)
            point = point + self.frame.w * dZ
            localSphericalCoords = self.frame.fromWorldToSpherical(point)
            if localSphericalCoords is None:
                return None
        _, az, pol = localSphericalCoords
        xfract = az / self._scanAngle
        yfract = (pol / 1.5708) - 1
        return xfract, yfract

    @property
    def scanAngle(self):
        return math.degrees(self._scanAngle)

    @scanAngle.setter
    def scanAngle(self, newScanAngle):
        self._scanAngle = math.radians(newScanAngle)

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, newFrame):
        self._frame = Frame(newFrame.w, newFrame.u.negated, newFrame.v.negated, newFrame.origin)

    @property
    def resolution(self):
        return self._resolution

class Pushbroom1:
    def __init__(self, scanAngle, camera, cameraTilt, spinPeriod, descentV=0, scanLineXfract=0.5):
        self._frame = Frame.universal()
        self._scanAngle = math.radians(scanAngle)
        acrossTrackRes = camera.resolution[1]
        self._resolution = (int((scanAngle / 90) * acrossTrackRes), acrossTrackRes)

        camFrame = Frame(self._frame.v.negated, self._frame.w.negated, self._frame.u, self._frame.origin)
        camFrame = camFrame.rotateInPlace(self._frame.v, math.radians(cameraTilt))
        cameraDuplicate = copy.deepcopy(camera)
        cameraDuplicate.frame = camFrame
        midRay = cameraDuplicate.castRayOut(scanLineXfract, 0.5)
        _, self._aziOffset, _ = self._frame.fromWorldToSpherical(midRay)
        topRay = cameraDuplicate.castRayOut(scanLineXfract, 0)
        bottomRay = cameraDuplicate.castRayOut(scanLineXfract, 1)
        _, _, self._minVisiblePolar = self._frame.fromWorldToSpherical(topRay)
        _, _, self._maxVisiblePolar = self._frame.fromWorldToSpherical(bottomRay)

        self.spinPeriod = spinPeriod
        self.descentV = descentV

    def castRayIn(self, point, cropToImageBoundaries=True):
        frameDuplicate = copy.deepcopy(self._frame)
        frameDuplicate = frameDuplicate.rotateInPlace(self._frame.w, self._aziOffset)
        localSphericalCoords = frameDuplicate.fromWorldToSpherical(point)
        if localSphericalCoords is None:
            return None
        if self.descentV != 0:
            _, az, _ = localSphericalCoords
            dZ = self.spinPeriod * self.descentV * az / (2 * math.pi)
            point = point + frameDuplicate.w * dZ
            localSphericalCoords = frameDuplicate.fromWorldToSpherical(point)
            if localSphericalCoords is None:
                return None
        _, az, pol = localSphericalCoords
        xfract = az / self._scanAngle
        #yfract = (pol / 1.5708) - 1
        yfract = (pol - self._minVisiblePolar) / (self._maxVisiblePolar - self._minVisiblePolar)
        return xfract, yfract

    @property
    def scanAngle(self):
        return math.degrees(self._scanAngle)

    @scanAngle.setter
    def scanAngle(self, newScanAngle):
        self._scanAngle = math.radians(newScanAngle)

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, newFrame):
        self._frame = Frame(newFrame.w, newFrame.u.negated, newFrame.v.negated, newFrame.origin)

    @property
    def resolution(self):
        return self._resolution

class SpinningPushbroom:
    def __init__(self, scanAngle, camera, cameraTilt, spinPeriod, descentV=0, scanLineXfract=0.5):
        self._frame = Frame.universal()
        self._scanAngle = math.radians(scanAngle)
        acrossTrackRes = camera.resolution[1]
        self._resolution = (int((scanAngle / 90) * acrossTrackRes), acrossTrackRes)

        from matplotlib import pyplot as plt

        camFrame = Frame(self._frame.v.negated, self._frame.w.negated, self._frame.u, self._frame.origin)
        camFrame = camFrame.rotateInPlace(self._frame.v, math.radians(cameraTilt))
        cameraDuplicate = copy.deepcopy(camera)
        cameraDuplicate.frame = camFrame
        self._refYfract = np.linspace(0, 1, 2 * acrossTrackRes, endpoint=False)
        midRay = cameraDuplicate.castRayOut(scanLineXfract, self._refYfract)
        _, self._refAzimuth, self._refPolar = self._frame.fromWorldToSpherical(midRay)
        #test = np.interp(np.linspace(1.6, 3.1, 10), pol, self._aziOffset)
        #plt.plot(pol, self._aziOffset)
        #plt.plot(np.linspace(1.6, 3.1, 10), test)
        #plt.show()
        #1/0
        self._minVisiblePolar = np.nanmin(self._refPolar)
        self._maxVisiblePolar = np.nanmax(self._refPolar)
        self.spinPeriod = spinPeriod
        self.descentV = descentV

    def castRayIn(self, point, cropToImageBoundaries=True):
        frameDuplicate = copy.deepcopy(self._frame)
        localSphericalCoords = frameDuplicate.fromWorldToSpherical(point)
        if localSphericalCoords is None:
            return None
        if self.descentV != 0:
            _, az, _ = localSphericalCoords
            dZ = self.spinPeriod * self.descentV * az / (2 * math.pi)
            point = point + frameDuplicate.w * dZ
            localSphericalCoords = frameDuplicate.fromWorldToSpherical(point)
            if localSphericalCoords is None:
                return None
        _, az, pol = localSphericalCoords
        unseenPolars = (pol < self._minVisiblePolar) * (pol > self._maxVisiblePolar)
        azOffset = np.interp(pol, self._refPolar, self._refAzimuth)
        adjustedAz = np.mod(az - azOffset, 2 * np.pi)
        xfract = adjustedAz / self._scanAngle
        yfract = np.interp(pol, self._refPolar, self._refYfract)
        xfract[unseenPolars] = np.nan
        yfract[unseenPolars] = np.nan
        return xfract, yfract

    @property
    def scanAngle(self):
        return math.degrees(self._scanAngle)

    @scanAngle.setter
    def scanAngle(self, newScanAngle):
        self._scanAngle = math.radians(newScanAngle)

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, newFrame):
        self._frame = Frame(newFrame.w, newFrame.u.negated, newFrame.v.negated, newFrame.origin)

    @property
    def resolution(self):
        return self._resolution


'''f = Frame.universal()
f = f.rotateInPlace(Vec3.k(), math.radians(-13))
p = Vec3.i()
x = f.fromWorldToSpherical(p)
print(math.degrees(x[1]))

c1 = Camera([45, 45], 0, horizontalRes=1000)
p = Vec3((1000, 0, -100))
print(c1.castRayIn(p))
c2 = Pushbroom(0, 0)
print(c2.castRayIn(p, 180))'''