from cameras import Camera
from primitives import Cone, _DoubleInfiniteCone
from frame import Frame
from vec3 import Vec3
from ray import Ray
from spectra import SpectralDensityCurve
import math
import numpy as np
import visualise
from matplotlib import pyplot as plt
import time


class Spheroid:
    def __init__(self, frame, radii):
        self._radii = radii
        self._r1, self._r2, self._r3 = radii
        self._inverseSquareRadii = (1 / (self._r1 ** 2), 1 / (self._r2 ** 2), 1 / (self._r3 ** 2))
        self._frame = frame

    def intersect(self, ray):
        transformedRay = ray.toFrame(self._frame)
        RInvSq = self._inverseSquareRadii
        o = transformedRay.origin
        d = transformedRay.direction
        A = ((d.x ** 2) * RInvSq[0]) + ((d.y ** 2) * RInvSq[1]) + ((d.z ** 2) * RInvSq[2])
        B1 = o.x * d.x * RInvSq[0]
        B2 = o.y * d.y * RInvSq[1]
        B3 = o.z * d.z * RInvSq[2]
        B = 2 * (B1 + B2 + B3)
        C1 = (o.x ** 2) * RInvSq[0]
        C2 = (o.y ** 2) * RInvSq[1]
        C3 = (o.z ** 2) * RInvSq[2]
        C = C1 + C2 + C3 - 1
        determinant = ((B ** 2) - (4 * A * C))
        if type(o.x) is np.ndarray or type(d.x) is np.ndarray:
            root1 = np.full_like(d.x, np.nan)
            root2 = np.full_like(d.x, np.nan)
            root1[determinant > 0] = ((-B - (determinant ** 0.5)) / (2 * A))[determinant > 0]
            root2[determinant > 0] = ((-B + (determinant ** 0.5)) / (2 * A))[determinant > 0]
            root1[root1 < 0] = np.nan
            root2[root2 < 0] = np.nan
            return root1, root2


class Conic:
    def __init__(self, frame, angle):
        self._angle = math.radians(angle)
        self._tanSquaredAngle = math.tan(self._angle) ** 2
        self._frame = frame

    def intersect(self, ray):
        transformedRay = ray.toFrame(self._frame)
        r = transformedRay.origin
        d = transformedRay.direction
        A = (d.x ** 2) + (d.y ** 2) - ((d.z ** 2) * self._tanSquaredAngle)
        B = 2 * (r.x * d.x + r.y * d.y - r.z * d.z * self._tanSquaredAngle)
        C = (r.x ** 2) + (r.y ** 2) - ((r.z ** 2) * self._tanSquaredAngle)
        determinant = ((B ** 2) - (4 * A * C))
        if type(r.x) is np.ndarray or type(d.x) is np.ndarray:
            root1 = np.full_like(d.x, np.nan)
            root2 = np.full_like(d.x, np.nan)
            root1[determinant > 0] = ((-B - (determinant ** 0.5)) / (2 * A))[determinant > 0]
            root2[determinant > 0] = ((-B + (determinant ** 0.5)) / (2 * A))[determinant > 0]
            root1[root1 < 0] = np.nan
            root2[root2 < 0] = np.nan
            z = transformedRay.origin.z + root1 * transformedRay.direction.z
            root1[z < 0] = np.nan
            z = transformedRay.origin.z + root2 * transformedRay.direction.z
            root2[z < 0] = np.nan
            return root1, root2


def allSkyLOS(scFrame, resolution):
    azimuth = np.linspace(0, 2 * math.pi, 2 * resolution)
    polar = np.linspace(0, math.pi, resolution)
    azimuth, polar = np.meshgrid(azimuth, polar)
    return (scFrame.fromSphericalToWorld(1, azimuth, polar) - scFrame.origin).norm


def sphericalComaColumnDen(LOS, observer, nucPos, refRadius, refDensity, bMin=0):
    LOS = LOS.norm
    obsToNuc = nucPos - observer
    L = obsToNuc.projectOnto(LOS)
    anti = L.antiparallelWith(LOS)
    L = np.where(anti, -1 * L.length, L.length)
    b = obsToNuc.projectPerpTo(LOS).length
    b[b < bMin] = bMin
    colDens = ((0.5 * math.pi) + np.arctan(L / b)) * refDensity * (refRadius ** 2) / b
    return colDens


def sphericalDistColumnDen(LOS, observer, s1, s2, nucPos, refRadius, refDensity, bMin=0):
    LOS = LOS.norm
    obsToNuc = nucPos - observer
    L = obsToNuc.projectOnto(LOS)
    anti = L.antiparallelWith(LOS)
    L = np.where(anti, -1 * L.length, L.length)
    b = obsToNuc.projectPerpTo(LOS).length
    b[b < bMin] = bMin
    k = refDensity * (refRadius ** 2)

    def f(x):
        return k * np.arctan((L - x) / b) / b
    colDens = f(s2) - f(s1)
    return colDens


def micronDustPhaseFunc(ang):
    phaseCurve = SpectralDensityCurve.load('dust_phasefunc_micron').curve
    indices = (400 * ang / math.pi).astype(int)
    return phaseCurve[indices]


def solarFluxDen(atPos, sunPos):
    p = atPos
    return 1.7 * (1.496e11 ** 2) / ((sunPos - p).length ** 2)  # W m^-2 nm^-1


def solarFluxDenObserved(LOS, sunPos, scPos):
    solarRadius = 6.96e8  # m
    solarDist = (sunPos - scPos).length
    solarDiskAng = solarRadius / solarDist
    viewSun = (sunPos - scPos).angleWith(LOS) <= 1.5 * solarDiskAng
    return np.where(viewSun, solarFluxDen(scPos, sunPos) / np.sum(viewSun), 0)  # W m^-2 nm^-1


def radianceFromColDen(colDen, LOS, sunPos, scPos, nucPos, IFOV=1.53e-3):
    scatEff = 0.455
    xSection = math.pi * (1e-6 ** 2)  # m^2
    phaseAng = (scPos - sunPos).angleWith(LOS)
    dustColDen = colDen
    Fsol = solarFluxDen(nucPos, sunPos)  # W m^-2 (nm^-1)
    dustRadiance = Fsol * dustColDen * xSection * scatEff * micronDustPhaseFunc(phaseAng) / (4 * math.pi)
    fluxDens = solarFluxDenObserved(LOS, sunPos, scPos)  # + nucleusFluxDenObserved(LOS, scPos, nucPos, 5e3, phaseAng)
    return dustRadiance + (fluxDens / (IFOV ** 2))  # W m^-2 sr^-1 (nm^-1)


def sphericalColDenAlongFlybyA(refRadius, refDensity):
    flybyStart = Vec3((-5e9, 1e6, 0))
    los = Vec3((np.array((1,)), np.array((0,)), np.array((0,))))
    return sphericalComaColumnDen(los, flybyStart, Vec3((0, 0, 0)), refRadius, refDensity)


def plotSolarIrradianceOfActiveRegion():
    rotAng = np.radians(np.linspace(0, 360, 1000) * 4)
    irrad = np.cos(rotAng)
    irrad[irrad < 0] = 0
    plt.plot(np.degrees(rotAng) / 360, irrad, color=visualise.valueColour(0))
    plt.xlim(0, 4)
    plt.ylim(0, 1.2)
    plt.xlabel('time [nucleus rotation periods]')
    plt.ylabel('relative ejection rate of active region')
    plt.grid(alpha=0.3)
    plt.tick_params(direction='inout', length=5)
    path = '../data/products/figures/chapter 5/{0}.png'.format('spinning nucleus dust production')
    plt.savefig(path, dpi=200)
    plt.show()


def plotSphericalComaRenderTimes():
    scf = Frame.principal(Vec3((1, 0, 0)))

    def getTime(reps, res):
        los = allSkyLOS(scf, res)
        T = 0

        for _ in range(reps):
            t1 = time.time()
            c = sphericalComaColumnDen(los, scf.origin, Vec3.zeroVector(), 1, 1)
            t2 = time.time()
            T += (t2 - t1)
        t = T / reps
        return los.x.size, t
    sizes = []
    times = []
    for res in np.geomspace(1, 3200, 15):
        rep = 1000
        if res > 50:
            rep = 100
        if res > 100:
            rep = 10
        if res > 1000:
            rep = 1
        result = getTime(rep, int(res))
        sizes += [result[0]]
        times += [result[1]]
    plt.plot(sizes, times, color=visualise.valueColour(0))
    plt.xscale('log')
    plt.yscale('log')
    plt.xlim(1e2, 1e7)
    plt.ylim(1e-4, 1e1)
    plt.xlabel('no. of pixels in image')
    plt.ylabel('image render time [s]')
    plt.grid(color='gray', alpha=0.15, which='both')
    plt.grid(color='gray', alpha=0.4)
    plt.tick_params(direction='inout', length=5)
    path = '../data/products/figures/chapter 5/{0}.png'.format('numpy coma cast speed')
    plt.savefig(path, dpi=200)
    plt.show()


def plotNumericalStructureRenderTimes():
    scf = Frame.principal(Vec3((1, 0, 0)))
    sph = Spheroid(Frame.universal(), [3e-1, 3e-1, 3e-1])

    def getTime(reps, res, nSamp):
        los = allSkyLOS(scf, res)
        T = 0

        for _ in range(reps):
            t1 = time.time()
            depth1, depth2 = sph.intersect(Ray(scf.origin, los))
            depths = np.linspace(depth1, depth2, nSamp)
            pos = scf.origin + (los * depths)
            d = (pos.length ** -2) * np.sin(pos.length)
            t2 = time.time()
            T += (t2 - t1)
        t = T / reps
        return los.x.size, t
    for sampN, colVal in zip([10, 100], [0, 0.4]):
        sizes = []
        times = []
        for res in np.geomspace(5, 1000, 15):
            rep = 100
            if res > 50:
                rep = 10
            if res > 100:
                rep = 1
            result = getTime(rep, int(res), sampN)
            sizes += [result[0]]
            times += [result[1]]
        plt.plot(sizes, times, label=sampN, color=visualise.valueColour(colVal))
    plt.xscale('log')
    plt.yscale('log')
    plt.xlim(1e2, 1e7)
    plt.ylim(1e-3, 1e2)
    plt.xlabel('no. of pixels in image')
    plt.ylabel('image render time [s]')
    plt.grid(color='gray', alpha=0.15, which='both')
    plt.grid(color='gray', alpha=0.4)
    plt.tick_params(direction='inout', length=5)
    plt.legend(title='samples per pixel')
    path = '../data/products/figures/chapter 5/{0}.png'.format('numpy numeric structure cast speed')
    plt.savefig(path, dpi=200)
    plt.show()

plotNumericalStructureRenderTimes()
1/0


#plotSolarIrradianceOfActiveRegion()


N = 500
cam = Camera([45, 45], 0, horizontalRes=N)

sun = Vec3((0, -1, 0)).norm * 1.496e11
nucleus = Vec3((0, 0, 0))
sc = Vec3((0, -5e4, 5e5))
scFrame = Frame.principal(sc)

cam.position = Vec3((-5e5, 0, 0))

xfract, yfract = np.meshgrid(np.linspace(0, 1, N), np.linspace(0, 1, N))
rayFrame = Ray(cam.position, cam.castRayOut(xfract, yfract))
rayAllSky = Ray(scFrame.origin, allSkyLOS(scFrame, N))

cdComa = sphericalComaColumnDen(rayAllSky.direction, rayAllSky.origin, nucleus, 1e6, 22, bMin=7000)
cd = cdComa

cone1 = Conic(Frame.withW(Vec3((1, 1, 1))), 10)
cone2 = Conic(Frame.withW(Vec3((-0.2, 1, 1))), 5)
cone3 = Conic(Frame.withW(sun.norm), 5)
for cone in [cone1, cone2, cone3]:
    root1, root2 = cone.intersect(rayAllSky)
    #snip = root2 - root1 > 3e5
    #root1[snip] = np.nan
    #root2[snip] = np.nan
    cdJet = sphericalDistColumnDen(rayAllSky.direction, rayAllSky.origin, root2, root1, nucleus, 1e6, 22 * 3, bMin=0)
    cdJet[np.isnan(cdJet)] = 0
    cd += cdJet

'''sphere = Spheroid(Frame.universal(), [1e5, 1e5, 1e5])
root1, root2 = sphere.intersect(rayFrame)
root1, root2 = sphere.intersect(rayAllSky)
scInside = np.isnan(root1) * ~np.isnan(root2)
root1[scInside] = 0
arcCone = Conic(Frame.withW(Vec3((0, 0, 1))), 60)
#root1, root2 = arcCone.intersect(rayFrame)'''

#visualise.display(root1)
#visualise.display(root2)

#cdComa = sphericalComaColumnDen(rayAllSky.direction, rayAllSky.origin, nucleus, 1e6, 22, bMin=7000)
#cdComa = sphericalComaColumnDen(rayFrame.direction, rayFrame.origin, nucleus, 1e6, 22, bMin=7000)
#cdJet = sphericalDistColumnDen(rayAllSky.direction, rayAllSky.origin, root2, root1, nucleus, 1e6, 22 * 3, bMin=0)
#cdJet[np.isnan(cdJet)] = 0
#cd = cdComa + cdJet
#cd = cdComa
#visualise.display(cdComa, map='grey')


'''ns = 50
ds = (root2 - root1) / ns
depths = np.linspace(root1, root2, ns)
#pos = cam.position + (rayFrame.direction * depths)
pos = scFrame.origin + (rayAllSky.direction * depths)
rad = (pos - nucleus).length
refRad = 1e6
refDen = 22
#density = ((refRad / rad) ** 2) * refDen
#density = ((refRad / rad) ** 2) * refDen * (np.sin(rad / 7e3) ** 2)
#density *= pos.z
density = ((refRad / rad) ** 2) * refDen * np.sin(rad / 3e3)
density[density < 0] = 0
density[pos.z < 0] = 0
cd = np.sum(density, axis=0) * ds
cd[np.isnan(cd)] = 0
#visualise.display(cd + (cdComa * 5e4), map='grey')
#visualise.display(cd + (cdComa * 4e0), map='grey', vlims=(0, 1.6e10))'''

rad = radianceFromColDen(cd, rayAllSky.direction, sun, sc, nucleus)
raw = np.array([1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9])
visualise.display(np.log10(rad), vlims=(-9, -4), cb=1, cbticks=np.log10(raw), cbticklabs=('10$^{-4}$', '10$^{-5}$', '10$^{-6}$', '10$^{-7}$', '10$^{-8}$', '10$^{-9}$'), cbextend='both',
                  xticks=np.linspace(0, 360, 5), yticks=np.linspace(0, 180, 5), xtitle='azimuth [$^o$]',
                  ytitle='polar [$^o$]', cbtitle='radiance [W m$^{-2}$ sr$^{-1}$ nm$^{-1}$]',
                  saveWithName='chapter 5/enviss_SSDCM_plus_jets_raw_rad_map_5e2km_widelims', scaleFactor=1.5, cbshrink=0.7)
