import math
import numpy as np
from vec3 import Vec3
from frame import Frame
from spectra import SpectralDensityCurve
from cameras import EquidistanceCamera as EQCamera
import visualise


def allSkyLOS(scFrame, resolution):
    azimuth = np.linspace(0, 2 * math.pi, 2 * resolution)
    polar = np.linspace(0, math.pi, resolution)
    azimuth, polar = np.meshgrid(azimuth, polar)
    return (scFrame.fromSphericalToWorld(1, azimuth, polar) - scFrame.origin).norm


def allSkyAziPol(resolution):
    return np.meshgrid(np.linspace(0, 2 * math.pi, 2 * resolution), np.linspace(0, math.pi, resolution))


def frameletAziPolLOS(resolution, filterColBounds, nExposures, dAz):
    cam = EQCamera([180, 180], 1, verticalRes=resolution)
    xfract = np.arange(filterColBounds[0], filterColBounds[1]) / resolution
    yfract = np.linspace(0, 1, resolution)
    xfract, yfract = np.meshgrid(xfract, yfract)
    point = cam.castRayOut(xfract, yfract)
    _, az, pol = Frame.universal().fromWorldToSpherical(point)
    dAz = np.arange(nExposures) * dAz
    dAz = np.repeat(dAz, az.shape[1])
    dAz = np.repeat(dAz[None, :], az.shape[0], axis=0)
    az = np.tile(az, nExposures) + dAz
    az = np.mod(az, 2 * math.pi)
    pol = np.tile(pol, nExposures)
    return az, pol


def micronDustPhaseFunc(ang):
    phaseCurve = SpectralDensityCurve.load('dust_phasefunc_micron').curve
    indices = (400 * ang / math.pi).astype(int)
    return phaseCurve[indices]


def nucleusPhaseFunc(ang):
    return np.cos(ang * 0.5) ** 3.42


def sphericalComaColumnDen(LOS, observer, nucPos, refRadius, refDensity):
    LOS = LOS.norm
    obsToNuc = nucPos - observer
    L = obsToNuc.projectOnto(LOS)
    anti = L.antiparallelWith(LOS)
    L = np.where(anti, -1 * L.length, L.length)
    b = obsToNuc.projectPerpTo(LOS).length
    bMin = 5e3
    b[b < bMin] = bMin
    colDens = ((0.5 * math.pi) + np.arctan(L / b)) * refDensity * (refRadius ** 2) / b
    return colDens


def solarFluxDen(atPos, sunPos):
    p = atPos
    return 1.7 * (1.496e11 ** 2) / ((sunPos - p).length ** 2)  # W m^-2 nm^-1


def solarFluxDenObserved(LOS, sunPos, scPos):
    solarRadius = 6.96e8  # m
    solarDist = (sunPos - scPos).length
    solarDiskAng = solarRadius / solarDist
    viewSun = (sunPos - scPos).angleWith(LOS) <= 1.5 * solarDiskAng
    return np.where(viewSun, solarFluxDen(scPos, sunPos) / np.sum(viewSun), 0)  # W m^-2 nm^-1


def nucleusFluxDenObserved(LOS, scPos, nucPos, nucRadius, phaseAngle):
    nucDist = (scPos - nucPos).length
    nucDiskAng = nucRadius / nucDist
    losAngFromNuc = (nucPos - scPos).angleWith(LOS)
    viewNuc = losAngFromNuc <= 1.1 * nucDiskAng
    if not np.any(viewNuc):
        viewNuc = losAngFromNuc == np.nanmin(losAngFromNuc)
    refFluxDen = 5e-13  # W m^-2 nm^-1 at 10^9 metres away, 0 degree phase angle, 5 km diam nucleus 1AU from Sun
    nucSizeFactor = (2 * nucRadius / 5e3) ** 2
    nucDistFactor = (1e9 / (scPos - nucPos).length) ** 2
    fluxDen = refFluxDen * nucSizeFactor * nucDistFactor * nucleusPhaseFunc(phaseAngle[viewNuc]) / np.sum(viewNuc)
    obsFluxDen = np.zeros_like(phaseAngle)
    obsFluxDen[viewNuc] = fluxDen
    return obsFluxDen  # W m^-2 nm^-1


def losRadianceGet(LOS, sunPos, scPos, nucPos, IFOV=1.53e-3):
    scatEff = 0.455
    xSection = math.pi * (1e-6 ** 2)  # m^2
    phaseAng = (scPos - sunPos).angleWith(LOS)
    dustColDen = sphericalComaColumnDen(LOS, scPos, nucPos, 1e6, 22)
    Fsol = solarFluxDen(nucPos, sunPos)  # W m^-2 (nm^-1)
    dustRadiance = Fsol * dustColDen * xSection * scatEff * micronDustPhaseFunc(phaseAng) / (4 * math.pi)
    fluxDens = solarFluxDenObserved(LOS, sunPos, scPos) + nucleusFluxDenObserved(LOS, scPos, nucPos, 5e3, phaseAng)
    return dustRadiance + (fluxDens / (IFOV ** 2))  # W m^-2 sr^-1 (nm^-1)


def losElectronRateGet(losRadiance, IFOV, apDiam, opTrans, filtTrans, filtWidth, filtCentre, qe):
    apArea = math.pi * ((0.5 * apDiam) ** 2)  # m^2
    energyPerPhoton = 3e8 * 6.6e-34 / (filtCentre / 1e9)
    pixIrradiantPower = losRadiance * (IFOV ** 2) * apArea * filtWidth * filtTrans * opTrans
    return qe * pixIrradiantPower / energyPerPhoton  # electrons per second per pixel


def sampleSky(losElecRate, losAzi, losPol, tExp, nSamp, spinP, fullWell):
    nanMap = np.isnan(losAzi) * np.isnan(losPol)
    sky = losElecRate
    col = (sky.shape[1] * losAzi / (2 * math.pi)).astype(int)
    row = (sky.shape[0] * losPol / math.pi).astype(int)
    col[nanMap] = 0
    row[nanMap] = 0
    data = np.zeros_like(losAzi)
    expRot = 2 * math.pi * tExp / spinP
    dPix = expRot * sky.shape[1] / (2 * math.pi * nSamp)
    for n in range(nSamp):
        data += (tExp / nSamp) * sky[row, col]
        sky = np.roll(losElecRate, int(n * dPix), axis=1)
    data[data > fullWell] = fullWell
    return data


def addNoise(image, darkCount, Nr):
    return np.random.normal(np.random.poisson(image) + np.random.poisson(darkCount) - int(darkCount), Nr)


def noiseFloor(nBinX, nBinY, Nr):
    return Nr * ((nBinX * nBinY) ** 0.5)


def eCountLims(radLims, nBinX, nBinY, t):
    rads = np.array(radLims)
    return tuple(losElectronRateGet(rads, 1.53e-3, 1.3e-3, 0.8, 0.9, 250, 675, 0.55) * t * nBinX * nBinY)


def binImage(image, hFactor, vFactor):
    def sumChunks(array, factor, axis=-1):
        shape = array.shape
        if axis < 0:
            axis += array.ndim
        shape = shape[:axis] + (-1, factor) + shape[axis + 1:]
        array = array.reshape(shape)
        return array.sum(axis=axis+1)
    return sumChunks(sumChunks(image, hFactor), vFactor, axis=0)


def saveArray(array, path):
    np.save(path, array)


def saveImage(image, path):
    pass


def makeRadianceMap(sunPos, scPos, nucPos, resolution, savePath):
    radianceMap = losRadianceGet(allSkyLOS(Frame.universal(), resolution), sunPos, scPos, nucPos)
    saveArray(radianceMap, savePath)


nucleus = Vec3((0, 0, 0))
sun = Vec3((0, -1, 0)).norm * 1.496e11
sc = Vec3((0, -5e5, 1e5))

radiance = losRadianceGet(allSkyLOS(Frame.universal(), 2000), sun, sc, nucleus)
eRate = losElectronRateGet(radiance, 1.53e-3, 1.3e-3, 0.8, 0.9, 250, 675, 0.55)
a, p = frameletAziPolLOS(2000, [975, 1025], 80, math.pi / 40)
a, p = allSkyAziPol(2000)
im = sampleSky(eRate, a, p, 1e-3, 1, 4, 13500)
l = eCountLims([1e-9, 1e-4], 1, 1, 1e-3)
visualise.display(np.log10(im), vlims=(np.log10(l[0]), np.log10(l[1])))
