import math
import numpy as np
from vec3 import Vec3, Mat3
#from mat3 import Mat3

class Frame:

    @staticmethod
    def _convertBetweenFrames(pointInA, frameA, frameB):
        """Returns the result of converting a point from one frame (frameA) to another (frameB)."""
        return frameB.matrixInverse * (frameA.origin - frameB.origin + (frameA.matrix * pointInA))

    @staticmethod
    def changeBasis(vector, frameA, frameB):
        """Returns the result of changing the vector's basis from frameA's axes to frameB's."""
        return frameB.matrixInverse * (frameA.matrix * vector)

    @staticmethod
    def toSpherical(point):
        """Returns the result of converting a cartesian coordinate to a spherical polar coordinate in the same frame.

        :param point: cartesian coordinate.
        :return: spherical polar coordinate as (radius, azimuth, polar) tuple.
        """
        r = point.length
        try:
            polar = math.acos(point.z / r)
            azimuth = Vec3.i().anticlockAngleWith(point.onPlaneXY, Vec3.k())
            return r, azimuth, polar
        except TypeError:
            polar = np.arccos(point.z / r)
            azimuth = Vec3.i().anticlockAngleWith(point.onPlaneXY, Vec3.k())
            return r, azimuth, polar

    @staticmethod
    def fromSpherical(radius, azimuth, polar):
        """Returns the cartesian coordinate of the input spherical coordinate, both with respect to the same frame.

        :param radius: spherical radius of point.
        :param azimuth: spherical azimuth angle (in radians) of point, measured from x axis around z axis.
        :param polar: spherical polar angle (in radians) of point, measured from z axis.
        :return: cartesian coordinate.
        """
        try:
            x = radius * math.sin(polar) * math.cos(azimuth)
            y = radius * math.sin(polar) * math.sin(azimuth)
            z = radius * math.cos(polar)
        except TypeError:
            x = radius * np.sin(polar) * np.cos(azimuth)
            y = radius * np.sin(polar) * np.sin(azimuth)
            z = radius * np.cos(polar)
        return Vec3((x, y, z))

    def __init__(self, u, v, w, origin):
        self._origin = origin
        self._u = u.norm
        self._v = v.norm
        self._w = w.norm
        top = (u[0], v[0], w[0])
        middle = (u[1], v[1], w[1])
        bottom = (u[2], v[2], w[2])
        self._matrix = Mat3(top, middle, bottom)
        self._matrixInverse = self.matrix.transpose

    @classmethod
    def universal(cls):
        return cls(Vec3.i(), Vec3.j(), Vec3.k(), Vec3.zeroVector())

    @classmethod
    def principal(cls, origin):
        return cls(Vec3.i(), Vec3.j(), Vec3.k(), origin)

    @classmethod
    def withU(cls, u, origin=Vec3.zeroVector()):
        v = Vec3.perpendicularTo(u)
        w = u.cross(v).norm
        return cls(u.norm, v, w, origin)

    @classmethod
    def withV(cls, v, origin=Vec3.zeroVector()):
        u = Vec3.perpendicularTo(v)
        w = u.cross(v).norm
        return cls(u, v.norm, w, origin)

    @classmethod
    def withW(cls, w, origin=Vec3.zeroVector()):
        u = Vec3.perpendicularTo(w)
        v = w.cross(u).norm
        return cls(u, v, w.norm, origin)

    @classmethod
    def fromUV(cls, u, v, origin=Vec3.zeroVector()):
        w = u.cross(v)
        return cls(u, v, w, origin)

    @classmethod
    def fromUW(cls, u, w, origin=Vec3.zeroVector()):
        v = w.cross(u)
        return cls(u, v, w, origin)

    @classmethod
    def fromVW(cls, v, w, origin=Vec3.zeroVector()):
        u = v.cross(w)
        return cls(u, v, w, origin)

    def fromWorld(self, point):
        """Returns the result of converting a coordinate from the universal frame to this frame."""
        return self._matrixInverse * (point - self._origin)

    def toWorld(self, point):
        """Returns the result of converting a coordinate from this frame to the universal frame."""
        return self._origin + self._matrix * point

    def toFrame(self, point, newFrame):
        """Returns the result of converting a coordinate from this frame to another."""
        return Frame._convertBetweenFrames(point, self, newFrame)

    def fromFrame(self, point, originFrame):
        """Returns the result of converting a coordinate to this frame from another."""
        return Frame._convertBetweenFrames(point, originFrame, self)

    def fromWorldToSpherical(self, point):
        """Returns the spherical polar coordinate of a point in the world frame, with respect to this frame.

        :param point: cartesian point in world frame.
        :return: radius, azimuth, polar tuple of spherical polar coordinate in this frame.
        """
        point = self.fromWorld(point)
        return Frame.toSpherical(point)

    def fromSphericalToWorld(self, radius, azimuth, polar):
        """Returns the world cartesian coordinate of the input local spherical coordinate.

        :param radius: spherical radius of point.
        :param azimuth: spherical azimuth angle (in radians) of point, measured from x axis around z axis.
        :param polar: spherical polar angle (in radians) of point, measured from z axis.
        :return: cartesian coordinate with respect to world frame.
        """
        return self.toWorld(Frame.fromSpherical(radius, azimuth, polar))

    def translate(self, vector):
        """Returns this frame translated by the input vector."""
        return Frame(self._u, self._v, self._w, self._origin + vector)

    def rotateInPlace(self, around, by):
        """Returns this frame rotated about its origin, anticlockwise around the input vector, by given amount (in radians)."""
        u = self._u.rotate(around, by)
        v = self._v.rotate(around, by)
        w = self._w.rotate(around, by)
        return Frame(u, v, w, self._origin)

    def rotate(self, aboutPoint, aroundAxis, by):
        """Returns this frame rotated about input point and vector, by given amount (in radians)"""
        u = self._u.rotate(aroundAxis, by)
        v = self._v.rotate(aroundAxis, by)
        w = self._w.rotate(aroundAxis, by)
        r = (self.origin - aboutPoint).rotate(aroundAxis, by)
        origin = aboutPoint + r
        return Frame(u, v, w, origin)

    def __str__(self):
        return 'u={0}, v={1}, w={2}, origin={3}'.format(self._u, self._v, self._w, self._origin)

    def __repr__(self):
        return 'Frame: u={0}, v={1}, w={2}, origin={3}.'.format(self._u, self._v, self._w, self._origin)

    @property
    def u(self):
        return self._u

    @property
    def v(self):
        return self._v

    @property
    def w(self):
        return self._w

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, newOrigin):
        self._origin = newOrigin

    @property
    def axes(self):
        return (self._u, self._v, self._w)

    @axes.setter
    def axes(self, newAxes):
        self._u, self._v, self._w = newAxes
        top = (self._u[0], self._v[0], self._w[0])
        middle = (self._u[1], self._v[1], self._w[1])
        bottom = (self._u[2], self._v[2], self._w[2])
        self._matrix = Mat3(top, middle, bottom)
        self._matrixInverse = self.matrix.transpose

    @property
    def matrix(self):
        return self._matrix

    @property
    def matrixInverse(self):
        return self._matrixInverse