import os
import math
from vec3 import Vec3
import numpy as np
from PIL import Image
from frame import Frame
from primitives import Spheroid


class Landscape:
    def __init__(self, Req, Rpol, Req_y=None):
        self._Req = Req
        if Req_y is None:
            self.Req_y = Req
        else:
            self.Req_y = Req_y
        self._Rpol = Rpol
        self._datum = Spheroid(Frame.universal(), (Req, self.Req_y, Rpol))

    @property
    def averageRadius(self):
        return self._datum.meanRadius

    @property
    def northPole(self):
        return self._datum.frame.w

    def localRadius(self, latitude):
        polar = (0.5 * math.pi) - latitude
        return self._datum.localRadius(0, polar)

    def sphericalToCart(self, latitude, longitude, altitude, returnAsLocalFrame=False):
        radius = self.localRadius(latitude) + altitude
        polar = (0.5 * math.pi) - latitude
        point = self._datum.frame.fromSpherical(radius, longitude, polar)
        if returnAsLocalFrame:
            w = (point - self._datum.position).norm
            u = w.cross(Vec3.k()).cross(w).norm
            frame = Frame.fromUW(u, w, origin=point)
            return frame
        return point

    def cartToSpherical(self, point):
        """Returns the result of converting the input cartesian point (in landscape frame) to latitude and longitude."""
        _, longitude, polar = self._datum.frame.toSpherical(point)
        latitude = (0.5 * math.pi) - polar
        return longitude, latitude

    def writeSurfaceMesh(self, resourcesName, pixelBox=None):
        """Converts DTM image to mesh of 3D points in space, and writes to file."""
        params = self.readResourceParameters(resourcesName)
        bounds = [params['minLong'], params['maxLat'], params['maxLong'], params['minLat']]
        minH = params['minAlt'] * 1000
        maxH = params['maxAlt'] * 1000
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'DTM_img.png')
        DTM = Image.open(path)
        DTM = DTM.convert('RGBA')
        pixels = DTM.load()
        DTM.close()
        width, height = DTM.size
        if pixelBox is not None:
            if not (pixelBox[0] < 0 or pixelBox[1] < 0 or pixelBox[2] > width or pixelBox[3] > height):
                minLong = ((pixelBox[0] / width) * (bounds[2] - bounds[0])) + bounds[0]
                maxLong = ((pixelBox[2] / width) * (bounds[2] - bounds[0])) + bounds[0]
                minLat = ((pixelBox[3] / height) * (bounds[1] - bounds[3])) + bounds[3]
                maxLat = ((pixelBox[1] / height) * (bounds[1] - bounds[3])) + bounds[3]
                longlatbox = (minLong, maxLat, maxLong, minLat)
            else:
                pixelBox = None
        if pixelBox is None:
            pixelBox = (0, 0, width, height)
            longlatbox = bounds
        longRange = abs(longlatbox[2] - longlatbox[0])
        latRange = abs(longlatbox[1] - longlatbox[3])
        shape = (pixelBox[3] - pixelBox[1], pixelBox[2] - pixelBox[0])
        mesh = []
        dH = maxH - minH
        for r in range(shape[0]):
            meshRow = []
            print(str(r) + '/' + str(shape[0]))
            row = r + pixelBox[1]
            lat = longlatbox[1] - ((r / shape[0]) * latRange)
            for c in range(shape[1]):
                col = c + pixelBox[0]
                long = longlatbox[0] + ((c / shape[1]) * longRange)
                pix = pixels[col, row]
                hAnom = (pix[0] / 255 * dH) + minH
                x, y, z = self.sphericalToCart(math.radians(lat), math.radians(long), hAnom)
                meshRow += [[x, y, z]]
            mesh += [meshRow]
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'surface_mesh')
        print('Writing mesh to file.')
        np.save(path, mesh)
        print('Finished writing mesh to file.')
        return None

    def generateMeshFromDTM(self, resourcesName, pixelBounds=None):
        """Converts DTM image to mesh of 3D points in space, and writes to file."""
        params = self.readResourceParameters(resourcesName)
        bounds = [params['minLong'], params['maxLat'], params['maxLong'], params['minLat']]
        leftLong = math.radians(params['minLong'])
        rightLong = math.radians(params['maxLong'])
        topLat = math.radians(params['maxLat'])
        bottomLat = math.radians(params['minLat'])
        extent = (leftLong, topLat, rightLong, bottomLat)
        minElevation = params['minAlt'] * 1000
        maxElevation = params['maxAlt'] * 1000
        elevations = np.linspace(minElevation, maxElevation, 256)
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'DTM_img.png')
        dtm = Image.open(path)
        dtm = dtm.convert('RGB')
        columns, rows = dtm.size
        longitudes = np.delete(np.linspace(leftLong, rightLong, columns + 1), columns)
        latitudes = np.delete(np.linspace(topLat, bottomLat, rows + 1), rows)
        if pixelBounds is None:
            pixelBounds = (0, 0, columns, rows)
        leftColumn = max(pixelBounds[0], 0)
        rightColumn = min(pixelBounds[2], columns)
        topRow = max(pixelBounds[1], 0)
        bottomRow = min(pixelBounds[3], rows)
        meshShape = (rightColumn - leftColumn, bottomRow - topRow)
        mesh = []
        for row in range(topRow, bottomRow):
            print('Calculating mesh nodes: {0}/{1} rows completed.'.format(row - topRow, meshShape[1]))
            meshRow = []
            latitude = latitudes[row]
            for column in range(leftColumn, rightColumn):
                longitude = longitudes[column]
                dtmPixelValue = int(dtm.getpixel((column, row))[0])
                elevation = elevations[dtmPixelValue]
                x, y, z = self.sphericalToCart(latitude, longitude, elevation)
                meshRow += [[x, y, z]]
            mesh += [meshRow]
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'surface_mesh')
        print('Writing mesh to file.')
        np.save(path, mesh)
        print('Finished writing mesh to file.')
        return None

    def generateTriangularMesh(self, resourcesName):
        params = self.readResourceParameters(resourcesName)
        bounds = [params['minLong'], params['maxLat'], params['maxLong'], params['minLat']]
        leftLong = math.radians(params['minLong'])
        rightLong = math.radians(params['maxLong'])
        topLat = math.radians(params['maxLat'])
        bottomLat = math.radians(params['minLat'])
        extent = (leftLong, topLat, rightLong, bottomLat)
        minElevation = params['minAlt'] * 1000
        maxElevation = params['maxAlt'] * 1000
        elevations = np.linspace(minElevation, maxElevation, 256)
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'DTM_img.png')
        dtm = Image.open(path)
        dtm = dtm.convert('RGB')
        columns, rows = dtm.size
        longPerPix = (rightLong - leftLong) / columns
        latPerPix = (topLat - bottomLat) / rows
        longitudes = np.delete(np.linspace(leftLong, rightLong, columns + 1), columns) + (0.5 * longPerPix)
        latitudes = np.delete(np.linspace(topLat, bottomLat, rows + 1), rows) + (0.5 * latPerPix)
        nodes = np.empty((3, rows, columns))
        for column, longitude in enumerate(longitudes):
            print('Generating mesh nodes: {0}/{1} columns completed...'.format(column, columns))
            for row, latitude in enumerate(latitudes):
                dtmPixelValue = int(dtm.getpixel((column, row))[0])
                elevation = elevations[dtmPixelValue]
                x, y, z = self.sphericalToCart(latitude, longitude, elevation)
                nodes[0, row, column] = x
                nodes[1, row, column] = y
                nodes[2, row, column] = z
        print('Mesh nodes generated, writing mesh to file...')
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'triangular_mesh')
        np.save(path, nodes)
        print('Mesh successfully generated and written to file.')
        return None

    def buildMeshFromDTM(self, resourcesName):
        "THIS FUNCTION IS THE MOST UP TO DATE VERSION"
        print('Generating mesh for resource {0}...'.format(resourcesName))
        params = self.readResourceParameters(resourcesName)
        minLong = math.radians(params['minLong'])
        maxLong = math.radians(params['maxLong'])
        minLat = math.radians(params['minLat'])
        maxLat = math.radians(params['maxLat'])
        minElevation = params['minAlt'] * 1000
        maxElevation = params['maxAlt'] * 1000
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'DTM_img.png')
        #dtm = np.array(Image.open(path).convert('RGB'))[:, :, 0]
        dtm = np.array(Image.open(path).convert('RGBA'))
        dtmNoData = dtm[:, :, 3] == 0
        dtmVals = dtm[:, :, 0]
        rows, columns = dtmVals.shape[:2]
        longPerColumn = (maxLong - minLong) / columns
        latPerRow = (maxLat - minLat) / rows
        longs = np.linspace(minLong, maxLong, columns + 1)[:-1] + 0.5 * longPerColumn
        lats = np.linspace(maxLat, minLat, rows + 1)[:-1] + 0.5 * latPerRow
        longs, lats = np.meshgrid(longs, lats)
        elevations = minElevation + (maxElevation - minElevation) * dtmVals / 255
        elevations[dtmNoData] = np.nan
        x, y, z = self.sphericalToCart(lats, longs, elevations)
        mesh = np.empty(dtmVals.shape + (3,))
        mesh[:, :, 0] = x
        mesh[:, :, 1] = y
        mesh[:, :, 2] = z
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'regular_triangular_mesh')
        np.save(path, mesh)
        print('Mesh successfully generated and written to file for resource {0}.'.format(resourcesName))
        return None

    def writeQuadtree(self, resourcesName):
        '''Writes spherical quadtree structure to file.'''
        params = self.readResourceParameters(resourcesName)
        minAlt = params['minAlt']
        maxAlt = params['maxAlt']
        dH = maxAlt - minAlt
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'DTM_img.png')
        DTM = Image.open(path)
        DTM = DTM.convert('RGBA')
        pixels = DTM.load()
        DTM.close()
        width, height = DTM.size
        def quarterRegions(regions):
            quadrants = []
            for region in regions:
                left, top, right, bottom = region
                midCol = int((left + right) * 0.5)
                midRow = int((top + bottom) * 0.5)
                q1 = (left, top, midCol, midRow)
                q2 = (midCol, top, right, midRow)
                q3 = (left, midRow, midCol, bottom)
                q4 = (midCol, midRow, right, bottom)
                quadrants += [q1, q2, q3, q4]
            return quadrants
        def findBoundingHeights(region):
            minDH, maxDH = 9e10, -9e10
            left, top, right, bottom = region
            for i in range(right - left):
                i = left + i
                for j in range(bottom - top):
                    j = top + j
                    pix = pixels[i, j]
                    hAnom = (pix[0] / 255 * dH) + minAlt
                    if minDH > hAnom:
                        minDH = hAnom
                    if maxDH < hAnom:
                        maxDH = hAnom
            return (minDH, maxDH)
        depth = int(math.log2(min(width, height)))
        quadrants = [(0, 0, width - 1, height - 1)]
        for m in range(depth):
            print(str(m + 1) + '/' + str(depth))
            quadrants = quarterRegions(quadrants)
            #data = np.empty((len(quadrants), 2))
            data = []
            for n, quadrant in enumerate(quadrants):
                minDH, maxDH = findBoundingHeights(quadrant)
                #data[n] = [minDH, maxDH]
                data += [minDH, maxDH]
            path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'quadtree')
            path = os.path.abspath(path)
            if not os.path.exists(path):
                os.makedirs(path)
            path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'quadtree', 'layer_' + str(m))
            np.save(path, data)
        return None

    def readResourceParameters(self, resourcesName):
        '''Returns dictionary of parameters associated with resource data.'''
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'params.txt')
        file = open(path)
        text = file.readlines()
        minLong = float(text[1])
        maxLat = float(text[2])
        maxLong = float(text[3])
        minLat = float(text[4])
        minAlt = float(text[6])
        maxAlt = float(text[7])
        flatAlt = float(text[9])
        data = {'minLong':minLong, 'maxLong':maxLong, 'minLat':minLat, 'maxLat':maxLat, 'minAlt':minAlt, 'maxAlt':maxAlt, 'flatAlt':flatAlt}
        return data