import math
from vec3 import Vec3
from primitives import Spheroid
from scene_objects import SceneObject, EmissionSpheroid
from spectra import PhotonRateSDC
#from geometries import Spheroid
#from shapes import Shape
from frame import Frame
import numpy as np
import scipy.integrate as integrate
import matplotlib.pyplot as plt

class Star(EmissionSpheroid):
    def __init__(self, position, temperature, radius):
        self._spectrum = PhotonRateSDC.blackbody(temperature)
        self._surfaceArea = 4 * 3.14159 * radius ** 2
        self._totalOutputFlux = 0.25 * self._surfaceArea * self._spectrum  # photons s^-1 sr^-1 nm^-1
        super().__init__(Frame.principal(position), (radius, radius, radius), self._spectrum)

    @classmethod
    def theSun(cls, position, radius):
        # Note: the 400-800nm range contributes 0.465 of the Sun's total power output
        return cls(position, 5778, radius)

    @property
    def surfaceArea(self):
        return self._surfaceArea

    @property
    def totalFlux(self):
        return self._totalOutputFlux  # photons s^-1 sr^-1 nm^-1

class Light(SceneObject):
    # Speed of light in SI
    _c = 3e8
    # Boltzmann constant in SI
    _k = 1.381e-23
    # Planck constant in SI
    _h = 6.626e-34

    def __init__(self, position, temperature, radius, minimumWavelength, maximumWavelength):
        frame = Frame(Vec3((1, 0, 0)), Vec3((0, 1, 0)), Vec3((0, 0, 1)), position)
        super().__init__()
        self._primitive = Spheroid(frame, (radius, radius, radius))
        self._temperature = temperature
        self._surfaceArea = 4 * np.pi * (radius ** 2)
        self._minimumWavelength = minimumWavelength
        self._maximumWavelength = maximumWavelength
        self._averageWavelength = (minimumWavelength + maximumWavelength) * 0.5
        self._photonFlux = self._radiantPower(minimumWavelength, maximumWavelength) * self._averageWavelength / (Light._h * Light._c)
        self._medium = 'light'

    def _spectralRadiance(self, wavelength):
        """Returns light's spectral radiance in W/[m^2 sr m], according to Planck's law."""
        x = wavelength
        a = 2.0 * Light._h * Light._c ** 2
        b = (Light._h * Light._c) / (x * Light._k * self._temperature)
        radiance = a / ((x ** 5) * (np.exp(b) - 1))
        return radiance

    def _spectralExitance(self, wavelength):
        """Returns body's emitted radiant flux per unit surface area per unit wavelength in W/[m^2 sr m]."""
        return self._spectralRadiance(wavelength) * np.pi

    def _radiance(self, minimumWavelength, maximumWavelength):
        """Returns body's radiance between two wavelengths, in W/[m^2 sr].

        Radiance is calculated by integration of Planck's law between the two input wavelengths.
        """
        integrand = self._spectralRadiance
        return integrate.quad(integrand, minimumWavelength, maximumWavelength)[0]

    def _radiantExitance(self, minimumWavelength, maximumWavelength):
        """Returns body's emitted radiant flux per unit surface area, between two wavelengths, in W/[m^2 sr]"""
        return self._radiance(minimumWavelength, maximumWavelength) * np.pi

    def _radiantPower(self, minimumWavelength, maximumWavelength):
        """Returns the total power radiated from the light, between two wavelengths, in W/sr"""
        return self._radiantExitance(minimumWavelength, maximumWavelength) * self._surfaceArea * 0.07958

    @property
    def photonFlux(self):
        """The total outward photon flux in 1/[s sr]"""
        return self._photonFlux

    @property
    def photonFluxPerArea(self):
        """The total outward photon flux per surface area of the light in 1/[s m^2 sr]"""
        return self._photonFlux / self._surfaceArea

#l = Light(Vec3.zeroVector(), 5778, 695e6, 1e-9, 1e3)
#print(l._radiantPower(600e-9, 1e3))
#print(l._radiantPower(600e-9, 700e-9))

c = 3e8
k = 1.381e-23
h = 6.626e-34
sig = 5.670e-8

def W(wavelength, T):
    return 2 * h * (c ** 2) / ((wavelength ** 5) * (np.exp(h * c / (wavelength * k * T)) - 1))

def SB(T):
    return sig * T ** 4


