import os
from vec3 import Vec3
from boundingBoxes import AABoundingBox
from frame import Frame
import math
import numpy as np
from PIL import Image

class Primitive:
    def __init__(self, frame):
        self._frame = frame# must be in world coordinates
        self._aabb = AABoundingBox.infinite()
        self._bulkRadius = 0
        self._inverseBulkRadius = math.inf
        self._description = 'Primitive[base class]'

    def contains(self, point):
        """Empty function to be overidden by subclasses, otherwise defaulting to False.
        
        Returns whether or not the geometry contains the input point.
        """
        return False

    def touchesPlane(self, plane):
        """Empty function to be overidden by subclasses, otherwise defaulting to False.
        
        Returns whether or not the geometry and input plane touch or intersect.
        """
        return False

    def _intersect(self, ray):
        """Private empty function to be overidden by subclasses."""
        print('Function has not been overidden by subclass. Primitive subclass must implement its own _intersect function.')
        return []

    def intersect(self, ray, bias=0):
        """Returns the intersections of the input ray with the primitive.
        
        ray must be in world frame coordinates.
        """
        depths = self._intersect(ray)
        hits = []
        if len(depths) == 1:
            if self.contains(ray.origin):
                # ray originates inside the object
                hit = dict()
                hit['depth'] = 0
                hit['position'] = ray.origin
                hit['type'] = 'internal'
                hits = [hit]
        for count, depth in enumerate(depths):
            hit = dict()
            position = ray.origin + depth * ray.direction
            if bias:
                position = position + bias * self.surfaceNormal(position)
                depth = (position - ray.origin).length
            hit['depth'] = depth
            hit['position'] = position
            if len(depths) == 1:
                if hits:
                    hit['type'] = 'exit'
                else:
                    hit['type'] = 'single intersection'
            else:
                if count == 0:
                    hit['type'] = 'entry'
                elif count == 1:
                    hit['type'] = 'exit'
            hits += [hit]
        return hits

    def intersectBool(self, ray):
        """Returns True/False result of whether input ray intersects the primitive."""
        return bool(self._intersect(ray))

    def surfaceNormal(self, point):
        """Returns the primitive's surface normal at the input point."""
        print('Function has not been overidden by subclass. Primitive subclass must implement its own surfaceNormal function.')
        return Vec3.zeroVector()

    def __str__(self):
        return '{0} at {1}.'.format(self._description, self.position)

    def __repr__(self):
        return '{0}(frame: {1}) id={2}.'.format(self._description, self._frame, id(self))

    @property
    def frame(self):
        return self._frame

    @property
    def position(self):
        return self._frame.origin

    @property
    def boundingBox(self):
        return self._aabb

    @property
    def bulkRadius(self):
        return self._bulkRadius

    @property
    def inverseBulkRadius(self):
        return self._inverseBulkRadius

class Plane(Primitive):
    def __init__(self, frame):
        self._normal = frame.w
        super().__init__(frame)
        self._description = 'Plane'

    @classmethod
    def containingPoints(cls, point1, point2, point3):
        """Class method which returns a Plane containing all three input points.
        
        This provides an alternative to initialising a Plane from a frame.
        Subclasses of Plane must override this class method with their own versions.
        """
        u = (point2 - point1).norm
        w = u.cross(point3 - point1).norm
        v = w.cross(u)
        frame = Frame(u, v, w, point1)
        return cls(frame)

    def _intersect(self, ray):
        """Return the range to the intersection (if any) of the input ray and the surface.
        
        ray must be in world coordinates.
        """
        denom = self._normal.dot(ray.direction)
        if abs(denom) > 1e-6:
            t = (self._frame.origin - ray.origin).dot(self._normal) / denom
            if t >= 0:
                return [t]
            #https://www.scratchapixel.com/lessons/3d-basic-rendering/minimal-ray-tracer-rendering-simple-shapes/ray-plane-and-ray-disk-intersection
        return []

    def surfaceNormal(self, point):
        return self._normal

    @property
    def normal(self):
        return self._normal

class Circle(Plane):
    def __init__(self, frame, radius):
        self._radius = radius
        super().__init__(frame)
        self._description = 'Circle'

    @classmethod
    def containingPoints(cls, centrePoint, point2, point3, radius):
        u = (point2 - centrePoint).norm
        w = u.cross(point3 - centrePoint).norm
        v = w.cross(u)
        frame = Frame(u, v, w, centrePoint)
        return cls(frame, radius)

    def _intersect(self, ray):
        """ray must be in world coordinates."""
        intersection = super()._intersect(ray)
        if not intersection:
            return []
        point = ray.origin + intersection[0] * ray.direction
        return intersection if (point - self.frame.origin).length <= self._radius else []

    def intersectRoots(self, ray):
        """Performs maximally efficient ray intersection test, and returns only the roots of the intersection.

        ray must be in world coordinates.
        """
        return self._intersect(ray)

class _InfiniteCylinder(Primitive):
    def __init__(self, frame, radius):
        self._radius = radius
        self._radiusSq = radius ** 2
        super().__init__(frame)
        self._description = '_InfiniteCylinder'

    def _intersect(self, ray):
        """ray must already be in this primitive's coordinate frame."""
        r = ray.origin
        d = ray.direction
        A = d.x ** 2 + d.y ** 2
        B = 2 * (r.x * d.x + r.y * d.y)
        C = r.x ** 2 + r.y ** 2 - self._radiusSq
        determinant = ((B ** 2) - (4 * A * C))
        if determinant < 0:
            return []
        elif determinant == 0:
            # ray has a single root with the surface
            root = -B / (2 * A)
            # return root if it's positive:
            return [root] if root >= 0 else []
        else:
            # ray has two roots with surface
            roots = [(-B + (determinant ** 0.5)) / (2 * A), (-B - (determinant ** 0.5)) / (2 * A)]
            # keep only positive roots:
            roots = [root for root in roots if root >= 0]
            # order roots in increasing depth
            roots.sort()
            return roots

    def contains(self, point):
        """point must already be in this primitive's coordinate frame."""
        return point.x ** 2 + point.y ** 2 <= self._radiusSq

class _OpenCylinder(_InfiniteCylinder):
    def __init__(self, frame, radius, length):
        self._length = length
        super().__init__(frame, radius)
        v1 = Vec3((-radius, -radius, 0))
        v2 = Vec3((radius, -radius, 0))
        v3 = Vec3((-radius, radius, 0))
        v4 = Vec3((radius, radius, 0))
        v5 = Vec3((-radius, -radius, length))
        v6 = Vec3((radius, -radius, length))
        v7 = Vec3((-radius, radius, length))
        v8 = Vec3((radius, radius, length))
        self._aabb = AABoundingBox.fromPoints([v1, v2, v3, v4, v5, v6, v7, v8], fromFrame=frame)
        self._description = '_OpenCylinder'

    def _intersect(self, ray):
        """ray must already be in this primitive's coordinate frame."""
        roots = super()._intersect(ray)
        # only return roots which sit within the cylinder's length limits:
        def clipRoot(root):
            z = ray.origin.z + root * ray.direction.z
            return 0 <= z <= self._length
        return [root for root in roots if clipRoot(root)]

    def contains(self, point):
        """point must already be in this primitive's coordinate frame."""
        return 0 <= point.z <= self._length and super().contains(point)

class Cylinder(_OpenCylinder):
    def __init__(self, frame, radius, length):
        self._nearEndCap = Circle(frame, radius)
        self._farEndCap = Circle(frame.translate(frame.w * length), radius)
        super().__init__(frame, radius, length)
        self._description = 'Cylinder'

    def _intersect(self, ray):
        """input ray must be in world coordinates."""
        transformedRay = ray.toFrame(self._frame)
        roots = super()._intersect(transformedRay)
        if len(roots) == 2:
            # Geometrically impossible for ray to intersect more than twice with closed cylinder
            return roots
        roots += self._nearEndCap.intersectRoots(ray)
        roots += self._farEndCap.intersectRoots(ray)
        roots.sort()
        return roots[:2]

class RoundedCylinder(_OpenCylinder):
    def __init__(self, frame, radius, length):
        self._nearEndCap = Spheroid(frame, (radius, radius, radius))
        self._farEndCap = Spheroid(frame.translate(frame.w * length), (radius, radius, radius))
        super().__init__(frame, radius, length)
        v1 = Vec3((-radius, -radius, -radius))
        v2 = Vec3((radius, -radius, -radius))
        v3 = Vec3((-radius, radius, -radius))
        v4 = Vec3((radius, radius, -radius))
        v5 = Vec3((-radius, -radius, length + radius))
        v6 = Vec3((radius, -radius, length + radius))
        v7 = Vec3((-radius, radius, length + radius))
        v8 = Vec3((radius, radius, length + radius))
        self._aabb = AABoundingBox.fromPoints([v1, v2, v3, v4, v5, v6, v7, v8], fromFrame=frame)
        self._description = 'RoundedCylinder'

    def _intersect(self, ray):
        """input ray must be in world coordinates."""
        transformedRay = ray.toFrame(self._frame)
        cylinderRoots = super()._intersect(transformedRay)
        nearCapRoots = self._nearEndCap.intersectRoots(ray)
        farCapRoots = self._farEndCap.intersectRoots(ray)
        capRoots = nearCapRoots + farCapRoots
        def clipRoot(root):
            z = transformedRay.origin.z + root * transformedRay.direction.z
            return not 0 <= z <= self._length
        capRoots = [root for root in capRoots if clipRoot(root)]
        roots = cylinderRoots + capRoots
        roots.sort()
        return roots

    def contains(self, point):
        return super().contains(point) or self._nearEndCap.contains(point) or self._farEndCap.contains(point)

class Spheroid(Primitive):
    def __init__(self, frame, radii):
        self._radii = radii
        self._r1, self._r2, self._r3 = radii
        self._inverseSquareRadii = (1 / (self._r1 ** 2), 1 / (self._r2 ** 2), 1 / (self._r3 ** 2))
        self._meanRadius = math.sqrt(((self._r1 ** 2) + (self._r2 ** 2) + (self._r3 ** 2)) / 3)
        super().__init__(frame)
        # calculate spheroid's axis aligned bounding box by calculating
        # the vertices of its minimum bounding box:
        v1 = Vec3((self._r1, self._r2, self._r3))
        v2 = Vec3((self._r1, self._r2, -self._r3))
        v3 = Vec3((self._r1, -self._r2, self._r3))
        v4 = Vec3((self._r1, -self._r2, -self._r3))
        v5 = Vec3((-self._r1, self._r2, self._r3))
        v6 = Vec3((-self._r1, self._r2, -self._r3))
        v7 = Vec3((-self._r1, -self._r2, self._r3))
        v8 = Vec3((-self._r1, -self._r2, -self._r3))
        self._aabb = AABoundingBox.fromPoints([v1, v2, v3, v4, v5, v6, v7, v8], fromFrame=frame)
        self._bulkRadius = min(self._radii)
        self._inverseBulkRadius = 1 / self._bulkRadius
        self._description = 'Spheroid'

    def contains(self, point):
        centre = self.position
        RInvSq = self._inverseSquareRadii
        A = ((point.x - centre.x) ** 2) * RInvSq[0]
        B = ((point.y - centre.y) ** 2) * RInvSq[1]
        C = ((point.z - centre.z) ** 2) * RInvSq[2]
        return A + B + C < 1

    def touchesPlane(self, plane):
        """Returns whether or not the geometry and input plane touch or intersect."""
        return abs(plane.frame.fromWorld(self.position).z) < self._bulkRadius

    def _intersect(self, ray):
        """Return the ranges to any intersections of the input ray with the surface.
        
        ray must be in world coordinates
        """
        localRay = ray.toFrame(self._frame)
        RInvSq = self._inverseSquareRadii
        o = localRay.origin
        d = localRay.direction
        A = ((d.x ** 2) * RInvSq[0]) + ((d.y ** 2) * RInvSq[1]) + ((d.z ** 2) * RInvSq[2])
        B1 = o.x * d.x * RInvSq[0]
        B2 = o.y * d.y * RInvSq[1]
        B3 = o.z * d.z * RInvSq[2]
        B = 2 * (B1 + B2 + B3)
        C1 = (o.x ** 2) * RInvSq[0]
        C2 = (o.y ** 2) * RInvSq[1]
        C3 = (o.z ** 2) * RInvSq[2]
        C = C1 + C2 + C3 - 1
        determinant = ((B ** 2) - (4 * A * C))
        if (type(ray.origin.x) is not np.ndarray) and (type(ray.direction.x) is not np.ndarray):
            if determinant <= 0:
                return []
            roots = []
            root1 = (-B + (determinant ** 0.5)) / (2 * A)
            root2 = (-B - (determinant ** 0.5)) / (2 * A)
            if root1 > 0:
                roots = [root1]
            if root2 > 0:
                roots += [root2]
            roots.sort()
            return roots
        else:
            noIntersection = determinant <= 0
            determinant[noIntersection] = np.nan
            root1 = (-B + (determinant ** 0.5)) / (2 * A)
            root2 = (-B - (determinant ** 0.5)) / (2 * A)
            root1[root1 < 0] = np.nan
            root2[root2 < 0] = np.nan
            rootsUnordered = np.array((root1, root2))
            roots = [np.nanmin(rootsUnordered, axis=0), np.nanmax(rootsUnordered, axis=0)]
            return roots

    def intersectRoots(self, ray):
        """Performs maximally efficient ray intersection test, and returns only the roots of the intersection.

                ray must be in world coordinates.
                """
        return self._intersect(ray)

    def surfaceNormal(self, point):
        """Returns the primitive's surface normal at the input point.
        
        point must be in world coordinates.
        """
        Rx, Ry, Rz = self._inverseSquareRadii
        point = self._frame.fromWorld(point)
        return self._frame.toWorld(Vec3((point.x * Rx, point.y * Ry, point.z * Rz)).norm)

    def localRadius(self, azimuth, polar):
        """Returns the radius of the spheroid's surface at the input azimuth and polar coordinates."""
        try:
            csa = math.cos(azimuth) ** 2
            ssa = math.sin(azimuth) ** 2
        except TypeError:
            csa = np.cos(azimuth) ** 2
            ssa = np.sin(azimuth) ** 2
        try:
            csp = math.cos(polar) ** 2
            ssp = math.sin(polar) ** 2
        except TypeError:
            csp = np.cos(polar) ** 2
            ssp = np.sin(polar) ** 2
        rr = self._inverseSquareRadii
        denom = (csa * ssp * rr[0]) + (ssa * ssp * rr[1]) + (csp * rr[2])
        return (1 / denom) ** 0.5

    @property
    def meanRadius(self):
        return self._meanRadius

class _DoubleInfiniteCone(Primitive):
    def __init__(self, frame, angle):
        """A double cone of infinite length.

        :param frame: The cone's frame. Its origin is the cone's vertex, and its w axis the cone's long axis.
        :param angle: The cone's half angle in radians (the angle from central axis to cone's surface).
        """
        self._angle = angle
        self._tanSquaredAngle = math.tan(self._angle) ** 2
        super().__init__(frame)
        self._description = '_DoubleInfiniteCone'

    def _intersect(self, ray):
        """ray must already be in this primitive's coordinate frame."""
        #transformedRay = ray.toFrame(self._frame)
        #r = transformedRay.origin
        #d = transformedRay.direction
        r = ray.origin
        d = ray.direction
        A = (d.x ** 2) + (d.y ** 2) - ((d.z ** 2) * self._tanSquaredAngle)
        B = 2 * (r.x * d.x + r.y * d.y - r.z * d.z * self._tanSquaredAngle)
        C = (r.x ** 2) + (r.y ** 2) - ((r.z ** 2) * self._tanSquaredAngle)
        determinant = ((B ** 2) - (4 * A * C))
        if type(r) is np.ndarray or type(d) is np.ndarray:
            root1 = np.full_like(d, np.nan)
            root2 = np.full_like(d, np.nan)
            root1[determinant > 0] = (-B + (determinant ** 0.5)) / (2 * A)
            root2[determinant > 0] = (-B - (determinant ** 0.5)) / (2 * A)
            root1[root1 < 0] = np.nan
            root2[root2 < 0] = np.nan
            return root1, root2
        else:
            if determinant < 0:
                return []
            elif determinant == 0:
                # ray has a single root with the surface
                root = -B / (2 * A)
                # return root if it's positive:
                return [root] if root >= 0 else []
            else:
                # ray has two roots with surface
                roots = [(-B + (determinant ** 0.5)) / (2 * A), (-B - (determinant ** 0.5)) / (2 * A)]
                # keep only positive roots:
                roots = [root for root in roots if root >= 0]
                # order roots in increasing depth
                roots.sort()
                return roots

    def contains(self, point):
        """point must already be in this primitive's coordinate frame."""
        if point.length == 0:
            return True
        return point.z / point.length < math.cos(self._angle)

class _SingleInfiniteCone(_DoubleInfiniteCone):
    def __init__(self, frame, angle):
        """A single cone of infinite length.

        :param frame: The cone's frame. Its origin is the cone's vertex, and its w axis the cone's long axis.
        :param angle: The cone's half angle in radians (the angle from central axis to cone's surface).
        """
        super().__init__(frame, angle)
        self._description = '_SingleInfiniteCone'

    def _intersect(self, ray):
        """ray must already be in this primitive's coordinate frame."""
        #ray = ray.toFrame(self._frame)
        roots = super()._intersect(ray)
        # only return roots which correspond to the positive z cone:
        def clipRoot(root):
            z = ray.origin.z + root * ray.direction.z
            return 0 <= z
        return [root for root in roots if clipRoot(root)]

    def contains(self, point):
        """point must already be in this primitive's coordinate frame."""
        return point.z > 0 and super().contains(point)

class _SingleCone(_DoubleInfiniteCone):
    def __init__(self, frame, angle, maxLength, minLength):
        """A single cone with finite length.

        :param frame: The cone's frame. Its origin is the cone's vertex, and its w axis the cone's long axis.
        :param angle: The cone's half angle in radians (the angle from central axis to cone's surface).
        :param maxLength: The length the cone extends along its central axis from its frame's origin.
        :param minLength: The length at which the cone begins along its central axis from its frame's origin.
        """
        self._maxLength = maxLength
        self._minLength = minLength
        super().__init__(frame, angle)
        self._description = '_SingleCone'

    def _intersect(self, ray):
        """ray must already be in this primitive's coordinate frame."""
        #ray = ray.toFrame(self._frame)
        roots = super()._intersect(ray)
        # only return roots which conform to cone's length extents:
        def clipRoot(root):
            z = ray.origin.z + root * ray.direction.z
            return self._minLength <= z < self._maxLength
        return [root for root in roots if clipRoot(root)]

    def contains(self, point):
        """point must already be in this primitive's coordinate frame."""
        return (self._minLength < point.z < self._maxLength) and super().contains(point)

class Cone(_SingleCone):
    def __init__(self, frame, angle, length, minLength=0):
        """A single fully closed cone.

        :param frame: The cone's frame. Its origin is the cone's vertex, and its w axis the cone's long axis.
        :param angle: The cone's half angle in degrees (the angle from central axis to cone's surface).
        :param length: The length the cone extends along its central axis from its frame's origin.
        :param minLength: The length at which the cone begins along its central axis from its frame's origin.
        """
        angle = math.radians(angle)
        r = minLength * math.tan(angle)
        R = length * math.tan(angle)
        self._farEndCap = Circle(frame.translate(frame.w * length), R)
        self._nearEndCap = None
        if minLength > 0:
            self._nearEndCap = Circle(frame.translate(frame.w * minLength), r)
        self._lengthInv = 1 / length
        self._centre = frame.origin + 0.5 * length * frame.w
        self._centreOpeningAngle = math.atan(2 * R / length)
        super().__init__(frame, angle, length, minLength)
        # calculate cone's axis aligned bounding box:
        v1 = Vec3((r, r, minLength))
        v2 = Vec3((r, -r, minLength))
        v3 = Vec3((-r, r, minLength))
        v4 = Vec3((-r, -r, minLength))
        v5 = Vec3((R, R, length))
        v6 = Vec3((R, -R, length))
        v7 = Vec3((-R, R, length))
        v8 = Vec3((-R, -R, length))
        self._aabb = AABoundingBox.fromPoints([v1, v2, v3, v4, v5, v6, v7, v8], fromFrame=frame)
        self._description = 'Cone'

    def _intersect(self, ray):
        """input ray must be in world coordinates."""
        transformedRay = ray.toFrame(self._frame)
        roots = super()._intersect(transformedRay)
        if len(roots) == 2:
            # Geometrically impossible for ray to intersect more than two times with cone, so roots are already found
            return roots
        # check for intersections with end caps:
        roots += self._farEndCap.intersectRoots(ray)
        if self._nearEndCap is not None:
            roots += self._nearEndCap.intersectRoots(ray)
        # order roots in increasing depth
        roots.sort()
        return roots[:2]

    def contains(self, point):
        """point must be in world coordinates."""
        point = self.frame.fromWorld(point)
        return super().contains(point)

    def surfaceNormal(self, point):
        r = point - self._centre
        if self.frame.w.angleWith(r) <= self._centreOpeningAngle:
            # return the normal of the end cap surface
            return self.frame.w
        rotationAxis = self.frame.w.cross(r)
        return self.frame.w.rotate(rotationAxis, self._angle + 1.571)

    @property
    def lengthInv(self):
        return self._lengthInv

class RoundedCone(_SingleCone):
    def __init__(self, frame, angle, length, radius):
        angle = math.radians(angle)
        sphereZ = radius / math.sin(angle)
        self._tipCutoffZ = sphereZ * math.cos(angle) ** 2
        R = length * math.tan(angle)
        frame = frame.translate(frame.w * -sphereZ)
        self._nearEndCap = Spheroid(frame.translate(frame.w * sphereZ), [radius, radius, radius])
        self._farEndCap = Circle(frame.translate(frame.w * length), R)
        super().__init__(frame, angle, length, self._tipCutoffZ)
        self._description = 'RoundedCone'

    def _intersect(self, ray):
        """input ray must be in world coordinates."""
        transformedRay = ray.toFrame(self._frame)
        coneRoots = super()._intersect(transformedRay)
        nearCapRoots = self._nearEndCap.intersectRoots(ray)
        farCapRoots = self._farEndCap.intersectRoots(ray)
        def clipRoot(root):
            z = transformedRay.origin.z + root * transformedRay.direction.z
            return z <= self._tipCutoffZ
        nearCapRoots = [root for root in nearCapRoots if clipRoot(root)]
        roots = coneRoots + nearCapRoots + farCapRoots
        roots.sort()
        return roots

    def contains(self, point):
        pass
