from vec3 import Vec3, Vec4
from primitives import Plane
from frame import Frame
from spectra import PhotonRateSDC

class Ray:
    def __init__(self, origin, direction, coordinateFrame=Frame.universal()):
        self._origin = origin
        self._direction = direction.norm
        self._coordinateFrame = coordinateFrame

    def toFrame(self, frame):
        """Returns the result of changing this ray's reference frame to the input frame."""
        newOrigin = self._coordinateFrame.toFrame(self._origin, frame)
        newDir = Frame.changeBasis(self.direction, self._coordinateFrame, frame)
        return Ray(newOrigin, newDir, coordinateFrame=frame)

    def point(self, atLength):
        return self._origin + atLength * self._direction

    @property
    def origin(self):
        return self._origin

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, newDirection):
        self._direction = newDirection.norm

    def __str__(self):
        return '{0}, {1}'.format(self.origin, self.direction)

    def __repr__(self):
        return 'Ray({0})'.format(self)

class LightRay_old(Ray):
    def __init__(self, origin, direction, coordinateFrame=Frame.universal()):
        super().__init__(origin, direction, coordinateFrame=coordinateFrame)
        # light's intensity in photons / [m^2 s sr] at ray's origin
        self._intensity = 0
        # store all the light ray's components of polarisation as a list of tuples
        # each tuple containing axis of polarisation, and proportion of total intensity:
        self._polarisationStates = []
        self._polarisationAxis = None
        self._polarisationPlaneNorm = None
        self.polarisationDegree = 0

    @classmethod
    def allZero(cls):
        return cls(Vec3.zeroVector(), Vec3.zeroVector())

    def scaleIntensity(self, factor):
        """Scales this light ray's intensity by the input factor"""
        self._intensity = abs(self._intensity * factor)

    @property
    def intensity(self):
        return self._intensity

    @intensity.setter
    def intensity(self, newIntensity):
        self._intensity = newIntensity

    @property
    def polarisationAxis(self):
        return self._polarisationAxis

    @polarisationAxis.setter
    def polarisationAxis(self, newAxis):
        self._polarisationPlaneNorm = newAxis.cross(self.direction).norm
        self._polarisationAxis = self.direction.cross(self._polarisationPlaneNorm).norm

    @property
    def polarisationPlaneNorm(self):
        return self._polarisationPlaneNorm

    def __add__(self, other):
        # it is only meaningful to add light rays if they have the same direction vector
        newRay = LightRay(self._origin, self._direction, coordinateFrame=self._coordinateFrame)
        newRay.intensity = self._intensity + other.intensity
        return newRay

    def __radd__(self, other):
        # light rays should only be added if they have the same direction vector
        newRay = LightRay(self._origin, self._direction, coordinateFrame=self._coordinateFrame)
        newRay.intensity = self._intensity + other.intensity
        return newRay

class LightRay(Ray):
    def __init__(self, origin, direction, coordinateFrame=Frame.universal()):
        # store light's intensity in photons / [m^2 s sr] at ray's origin
        ###self._intensity = 0
        self._intensity = PhotonRateSDC.zero()
        self._polarisedIntensity = PhotonRateSDC.zero()
        # use a plane to describe polarisation axis,
        # the plane's u axis is the ray's direction of propagation, its v axis the polarisation axis,
        # LightRay initiates with zero polarisation, so the plane is initialised with arbitrary v axis:
        self._polarisationPlane = Plane(Frame.withU(direction, origin=origin))
        super().__init__(origin, direction, coordinateFrame=coordinateFrame)

    @classmethod
    def zero(cls):
        return cls(Vec3.zeroVector(), Vec3.zeroVector())

    def mapTo(self, newOrigin, newDirection):
        """Returns the result of mapping this light ray to a new origin and direction, keeping any polarisation within the same plane."""
        newRay = LightRay(newOrigin, newDirection)
        newRay.intensity = self._intensity
        if self.polarisation == 0:
            return newRay
        newRay.polarisation = self.polarisation
        newRay.polarisationAxis = self.polarisationAxis.cross(newDirection).cross(newDirection)
        return newRay

    def changeDirection(self, vertexPosition, newDirection):
        """Returns the result of changing this light ray's direction, keeping any polarisation within the same plane."""
        redirectedRay = LightRay(vertexPosition, newDirection)
        redirectedRay.intensity = self._intensity
        if self.isPolarised:
            redirectedRay.polarisedIntensity = self.polarisedIntensity
            redirectedRay.polarisationAxis = self.polarisationAxis.cross(newDirection).cross(newDirection)
        return redirectedRay

    @property
    def intensity(self):
        return self._intensity

    @intensity.setter
    def intensity(self, newIntensity):
        self._intensity = newIntensity

    @property
    def polarisedIntensity(self):
        return self._polarisedIntensity

    @polarisedIntensity.setter
    def polarisedIntensity(self, newPolarisedIntensity):
        self._polarisedIntensity = newPolarisedIntensity

    @property
    def polarisationAxis(self):
        return self._polarisationPlane.frame.v

    @polarisationAxis.setter
    def polarisationAxis(self, newAxis):
        u = self._direction
        v = newAxis
        frame = Frame.fromUV(u, v, origin=self._origin)
        self._polarisationPlane = Plane(frame)

    @property
    def isPolarised(self):
        return not self.polarisedIntensity.hasZeroPhotons