from primitives import Plane
from frame import Frame
from ray import LightRay
from spectra import PhotonRateSDC
import math

class RayTracer:
    def __init__(self):
        pass

    def traceRay(self, ray, scene, polarisation=False):
        """Traces a ray into a scene and returns the light conditions at the ray's origin."""
        # works on the assumption that the footprint of the pixel viewing this ray always sees uniform conditions
        # and thus the brightness it sees due to objects is independant of proximity to said objects
        # this assumption holds true as long as variations in the scene are larger than pixel scale
        # and holds true even when they aren't if multiple disperse rays per pixel are averaged
        interactions = scene.getRayTraversal(ray)
        return self._calculateSurfaceBrightness(interactions, ray, scene, polarisation=polarisation)  # should return a lightRay with a spectral intensity

    def _calculateSurfaceBrightness(self, interactions, ray, scene, polarisation=False):
        """Returns the surface brightness seen by the input ray due to the input interactions with the input scene.

        :param interactions:
        :param ray:
        :param scene:
        :param polarisation:
        :return:
        """
        if not interactions:
            return LightRay.zero()
        lightSource = scene.lightSource
        rayPropagationDirection = ray.direction.negated
        lightRay = LightRay(interactions[-1]['startPoint'], rayPropagationDirection)
        totalFlux = PhotonRateSDC.zero()  # totalFlux = 0
        if polarisation:
            totalPolarisedFlux = PhotonRateSDC.zero()
            scatterPlane = Plane(Frame.withW(ray.direction.cross(ray.origin - lightSource.position)))
            polarisationAxis = scatterPlane.normal
        # iterate through the interactions in farthest-nearest order, first checking if an opaque surface is seen:
        finalInteraction = interactions[-1]
        if finalInteraction['object'].opaque:
            # ray terminates on an opaque object's surface
            # calculate the flux that this surface reflects along the ray:
            pointSeen = finalInteraction['startPoint']
            objectSeen = finalInteraction['object']
            incidentFlux, dirOfIncidence = scene.fluxFromLight(pointSeen, lightSource)  # photons s^-1 m^-2 nm^-1
            if incidentFlux is not None:
                reflectedFlux = objectSeen.reflectNEW(incidentFlux, pointSeen, dirOfIncidence)  # photons s^-1 m^-2 sr^-1 nm^-1
                totalFlux = reflectedFlux
            # remove this interaction from the list so that its flux is not double counted:
            interactions = interactions[:-1]
        for interaction in reversed(interactions):
            pointSeen = interaction['startPoint']
            objectSeen = interaction['object']
            incidentFlux, dirOfIncidence = scene.fluxFromLight(pointSeen, lightSource, includeShadows=False)  # photons s^-1 m^-2 nm^-1
            phaseAngle = ray.direction.angleWith(dirOfIncidence)
            """transmittance = objectSeen.pathTransmittance(interaction['length'], pointSeen, interaction['endPoint'])
            phaseFactor, dop = 0.3, 0.3
            #phaseFactor = objectSeen.phaseFunction(phaseAngle)
            phaseFactor = 0.06 * (1 + (math.cos(phaseAngle) ** 2))
            dop = 0
            scatteredFlux = phaseFactor * (1 - transmittance) * incidentFlux
            attenuatedFlux = transmittance * totalFlux
            totalFlux = attenuatedFlux + scatteredFlux"""
            """transmittance, scatterance = objectSeen.transmit(pointSeen, interaction['endPoint'], interaction['length'], phaseAngle)
            totalFlux = transmittance * totalFlux
            totalFlux += scatterance * incidentFlux"""
            totalFlux = objectSeen.transmit(pointSeen, interaction['endPoint'], interaction['length'], totalFlux, incidentFlux, phaseAngle)
            if polarisation:
                attenuatedPolarisedFlux = transmittance * totalPolarisedFlux
                totalPolarisedFlux = attenuatedPolarisedFlux + dop * scatteredFlux
        lightRay.intensity = totalFlux
        if polarisation:
            lightRay.polarisedIntensity = totalPolarisedFlux
            lightRay.polarisationAxis = polarisationAxis
        return lightRay  # should return a light ray with spectral intensity
