import os
import copy
import math
import numpy as np
from vec3 import Vec3
from ray import Ray
from raytracer import RayTracer
from PIL import Image, ImageDraw, ImageFilter
from operator import itemgetter
from frame import Frame
import h5py
import time
import itertools

class Renderer:
    """Class for rendering and saving a camera's images of 3D scenes."""

    def __init__(self, viewpoint):
        self.viewpoint = viewpoint
        self.camera = viewpoint.pose.camera

    def renderRays(self, raysArray, scene, productFilename, filters=()):
        # TODO: optimise, neaten up etc...
        rayTracer = RayTracer()
        # the shape of the image being rendered:
        renderingShape = np.shape(raysArray)
        # initialise arrays for storing each ray's radiance information:
        absoluteRadianceSpectra = np.zeros(renderingShape + (401,))
        polarisedRadianceSpectra = np.zeros_like(absoluteRadianceSpectra)
        anglesOfPolarisation = np.empty(renderingShape)
        # calculate a polarimetry mask, which defines for which rays polarimetry need and needn't be calculated:
        polarimetryMask = np.full_like(anglesOfPolarisation, False)
        for opticalFilter in filters:
            if opticalFilter.polarising:
                mask = opticalFilter.mask
                polarimetryMask += mask
        polarimetryMask = polarimetryMask.tolist()
        # perform the actual casting of the rays into the scene, calculating the radiance seen by each ray:
        print("Casting individual rays...0%", end="")
        for r, row in enumerate(raysArray):
            print("\rCasting individual rays...{0}%".format(int(100 * r / len(raysArray))), end="")
            for c, ray in enumerate(row):
                if ray is None:
                    continue
                # from the polarimetry mask, get whether this ray requires polarisation calculations:
                polarisationSensitive = polarimetryMask[r][c]
                # use ray tracing to calculate the light seen by this ray:
                lightRay = rayTracer.traceRay(ray, scene, polarisation=polarisationSensitive)
                # add this light ray's spectrum to the array of spectra:
                absoluteRadianceSpectra[r, c] = lightRay.intensity.curve
                if polarisationSensitive:
                    # add this light ray's polarised spectrum to the array of polarised spectra:
                    polarisedRadianceSpectra[r, c] = lightRay.polarisedIntensity.curve
                    # for the calculation of polarisation angle, redirect the light as the optics would:
                    lightRay.changeDirection(vertexPosition=self.camera.position, newDirection=self.camera.frame.u)
                    # calculate the light's angle of polarisation relative to the reference axis, and add to the array:
                    referenceAxis = self.camera._sensor._frame.v.negated
                    angleOfPolarisation = referenceAxis.anticlockAngleWith(lightRay.polarisationAxis, lightRay.direction)
                    anglesOfPolarisation[r, c] = angleOfPolarisation
        # every individual ray has been cast and the radiances arriving at the camera have been calculated.
        # simulate the transmission of the light through the camera, and its measurement by the detector:
        print("\rCast individual rays...100%.\nPerforming array-wide calculations.")
        # calculate the fluxes passing into the camera and onto each pixel:
        absoluteFluxSpectra = self.camera.captureFlux(absoluteRadianceSpectra)
        polarisedFluxSpectra = self.camera.captureFlux(polarisedRadianceSpectra)
        # if any filters are present, simulate their effects on the flux:
        for opticalFilter in filters:
            # retrieve the filter's mask, describing which pixels it does and doesn't act on:
            filterMask = opticalFilter.mask
            # retrieve the filter's spectral transmission curve:
            transmission = opticalFilter.transmission.curve
            # simulate the spectral attenuation of the flux by the filter:
            absoluteFluxSpectra[filterMask] = transmission * absoluteFluxSpectra[filterMask]
            # if the filter is also polarising, additional calculations are required:
            if opticalFilter.polarising:
                # simulate the spectral attenuation of the polarised flux by the filter:
                polarisedFluxSpectra[filterMask] = transmission * polarisedFluxSpectra[filterMask]
                # calculate the unpolarised portion of the flux:
                unpolarisedFluxSpectra = absoluteFluxSpectra[filterMask] - polarisedFluxSpectra[filterMask]
                # calculate the separations between the angles of polarisation and the filter's polarising axis:
                filterAng = opticalFilter.polarisingAngle
                separations = anglesOfPolarisation[filterMask] - filterAng
                # simulate the polarimetric transmission of the polarised flux, using Malus' law:
                polarisedTrans = np.cos(separations) ** 2
                polarisedFluxSpectra[filterMask] = polarisedTrans[..., None] * polarisedFluxSpectra[filterMask]
                # simulate the polarimetric transmission of the unpolarised flux:
                unpolarisedFluxSpectra = 0.5 * unpolarisedFluxSpectra
                # total transmitted flux is the sum of polarised and unpolarised fluxes:
                absoluteFluxSpectra[filterMask] = unpolarisedFluxSpectra + polarisedFluxSpectra[filterMask]
        # convert flux spectra to pixel digital values:
        # TODO: remove hard-coding of exposure time
        for n in range(1):
            image = self.camera.sensor.integrateFlux(absoluteFluxSpectra, 4 * 1e-3)
            #path = '../data/products/{0}{1}'.format(productFilename, n)
            path = '../data/products/{0}'.format(productFilename)
            np.save(path, image)
        #image = self.camera.sensor.integrateFlux(absoluteFluxSpectra, 4 * 1e-3)
        #path = '../data/products/{0}'.format(productFilename)
        #np.save(path, image)

    def _debug_silhouette(self, raysArray, scene, productFilename):
        rayTracer = RayTracer()
        product = np.zeros_like(raysArray)
        print("Casting individual rays...0%", end="")
        for r, row in enumerate(raysArray):
            print("\rCasting individual rays...{0}%".format(int(100 * r / len(raysArray))), end="")
            for c, ray in enumerate(row):
                traversal = scene.getRayTraversal(ray)
                if len(traversal) > 0:
                    product[r, c] = 255
        print("\rCasted individual rays...100%.")
        img = Image.fromarray(product.astype(int)).convert('RGB')
        img.save('../data/products/{0}.png'.format(productFilename))

    def getRaysSingleFrame(self):
        resolution = self.camera.resolution
        xfracts = tuple(np.linspace(0, 1, resolution[0] + 1)[:-1])
        yfracts = tuple(np.linspace(0, 1, resolution[1] + 1)[:-1])
        raysArray = []
        print("Generating rays to cast...0%", end="")
        for yfract in yfracts:
            print("\rGenerating rays to cast...{0}%".format(int(yfract * 100)), end="")
            raysArrayRow = []
            for xfract in xfracts:
                rayDir = self.camera.castRayOut(xfract, yfract)
                if rayDir is not None:
                    ray = Ray(self.camera.position, rayDir)
                    raysArrayRow += [ray]
                else:
                    raysArrayRow += [None]
            raysArray += [raysArrayRow]
        print("\rGenerated rays to cast 100%.")
        return raysArray

    def getRaysPushbroom(self, angIncrement, angScan, xfract=0.5, minYfract=0, maxYfract=1):
        rows = self.camera.resolution[1]
        columns = int(angScan / angIncrement)
        resolution = (columns, rows)
        yfracts = tuple(np.arange(int(minYfract * rows), int(maxYfract * rows)) / rows)
        raysArray = np.empty((rows, columns)).tolist()
        print("Generating rays to cast...0%", end="")
        for scanN in range(columns):
            print("\rGenerating rays to cast...{0}%".format(int(100 * scanN / columns)), end="")
            for row, yfract in enumerate(yfracts):
                rayDir = self.camera.castRayOut(xfract, yfract)
                if rayDir is not None:
                    ray = Ray(self.camera.position, rayDir)
                    raysArray[row][scanN] = ray
                else:
                    raysArray[row][scanN] = None
            self.viewpoint.pose.azimuth += -angIncrement
        print("\rGenerated rays to cast 100%.")
        return raysArray

    def renderScene_Frame(self, scene, productFilename):
        resolution = self.camera.resolution
        product = Image.new('RGBA', resolution, color=(0, 0, 0, 255))
        for x in range(resolution[0]):
            print(str(x) + '/' + str(resolution[0]))
            xfract = x / resolution[0]
            for y in range(resolution[1]):
                yfract = y / resolution[1]
                product.putpixel((x, y), self._renderPixel(scene, xfract, yfract))
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def renderScene_Frame_numpy(self, scene, productFilename, exposureTime=0.001):
        resolution = self.camera.resolution
        product = Image.new('RGBA', resolution, color=(0, 0, 0, 0))
        xfract, yfract = self.camera.rayFracts_numpy()
        ray = self.camera.castRaysOut_numpy(xfract, yfract)
        rayTracer = RayTracer()
        lightRays = rayTracer.traceRay_numpy(ray, scene)
        for y, row in enumerate(lightRays):
            for x, lightRay in enumerate(row):
                l = int(self.camera.digitiseLight(lightRay, exposureTime, x / resolution[0], y / resolution[1]))
                #print(l)
        product.putpixel((x, y), (l, l, l))
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def renderScene_Pushbroom(self, scene, xfract, angularResolution, angularRegard, productFilename):
        self.viewpoint.pose.azimuth += math.pi
        resolution = (int(angularRegard / angularResolution), self.camera.resolution[1])
        product = Image.new('RGBA', resolution, color=(0, 0, 0, 255))
        for x in range(resolution[0]):
            print(str(x) + '/' + str(resolution[0]))
            for y in range(resolution[1]):
                yfract = y / resolution[1]
                product.putpixel((x, y), self._renderPixel(scene, xfract, yfract))
                self.viewpoint.pose.azimuth += angularResolution / resolution[1]
                #self.viewpoint.position = self.viewpoint.position# + Vec3((0, 0, 0.17 * 7e3 / resolution[1]))
            #self.viewpoint.pose.azimuth += angularResolution
            #self.viewpoint.position = self.viewpoint.position# + Vec3((0, 0, -15e3 / resolution[0]))
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        self.viewpoint.pose.azimuth += math.pi
        return None

    def renderScene_Pushframe(self, scene, minXfract, maxXfract, angPerFramelet, numOfFramelets, productFilename):
        self.viewpoint.pose.azimuth += math.pi
        pixPerFramelet = int((maxXfract - minXfract) * self.camera.resolution[0])
        xfracts = []
        for n in range(pixPerFramelet):
            xfracts += [minXfract + (n / self.camera.resolution[0])]
        yRes = self.camera.resolution[1]
        xRes = pixPerFramelet * numOfFramelets
        product = Image.new('RGBA', (xRes, yRes), color=(0, 0, 0, 0))
        angIteration = angPerFramelet / self.camera.resolution[1]
        for N in range(numOfFramelets):
            print('{0}/{1} framelets.'.format(N, numOfFramelets))
            columnNumbers = []
            for n in range(pixPerFramelet):
                columnNumbers += [(N * pixPerFramelet) + n]
            for y in range(yRes):
                yfract = y / yRes
                for c, xfract in zip(columnNumbers, xfracts):
                    product.putpixel((c, y), self._renderPixel(scene, xfract, yfract))
                self.viewpoint.pose.azimuth += -angIteration
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def _renderPixel(self, scene, xfract, yfract, exposureTime=0.001):
        exposureTime = 282e-3
        rayTracer = RayTracer()
        rayVector = self.camera.castRayOut(xfract, yfract)
        if rayVector is None:
            return (255, 0, 0)
        ray = Ray(self.camera.position, rayVector)
        lightRay = rayTracer.traceRay(ray, scene)
        l = int(self.camera.digitiseLight(lightRay, exposureTime, xfract, yfract))
        return (l, l, l)

    def castFrame_spheroid(self, resourcesName, productFilename):
        """Renders frame image, by ray casting, assuming simple spheroid surface with no elevation data"""
        params = self.viewpoint.landscape.readResourceParameters(resourcesName)
        bounds = [params['minLong'], params['minLat'], params['maxLong'], params['maxLat']]
        deltaH = params['flatAlt']
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_fullRes_img.png')
        try:
            im = Image.open(path).convert('RGBA')
        except:
            path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
            im = Image.open(path).convert('RGBA')
        width, height = im.size
        resolution = self.camera.resolution
        # create image to render to
        product = Image.new('RGBA', resolution)
        # cast ray and calculate intersection for all pixels and scan lines
        for x in range(resolution[0]):
            print(str(x) + '/' + str(resolution[0]))
            xfract = x / resolution[0]
            for y in range(resolution[1]):
                yfract = y / resolution[1]
                rayVector = self.camera.castRayOut(xfract, yfract)
                seenPoint = self.seenPoint(rayVector, deltaH=deltaH)
                if seenPoint is not None:
                    lat, long = seenPoint
                    equiFracts = self.equirectFracts(lat, long, bounds=None)
                    if equiFracts is not None:
                        pixel = im.getpixel((int(equiFracts[0] * width), int(equiFracts[1] * height)))
                        product.putpixel((x, y), pixel)
        im.close()
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def castPB_spheroid(self, resourcesName, productFilename):
        """Renders push-broom image, by ray casting, assuming simple spheroid surface with no elevation data"""
        params = self.viewpoint.landscape.readResourceParameters(resourcesName)
        bounds = [params['minLong'], params['minLat'], params['maxLong'], params['maxLat']]
        deltaH = params['flatAlt']
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_fullRes_img.png')
        try:
            im = Image.open(path).convert('RGBA')
        except:
            path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
            im = Image.open(path).convert('RGBA')
        width, height = im.size
        prodWidth = int(self.camera.resolution[1] / self.camera.FOV[1] * 360)
        resolution = (3000, self.camera.resolution[1])
        # create image to render to
        product = Image.new('RGBA', resolution)
        # cast ray and calculate intersection for all pixels and scan lines
        xfract = 0.5
        for i in range(resolution[0]):
            print(str(i) + '/' + str(resolution[0]))
            for y in range(resolution[1]):
                yfract = y / resolution[1]
                rayVector = self.camera.castRayOut(xfract, yfract)
                seenPoint = self.seenPoint(rayVector, deltaH=deltaH)
                if seenPoint is not None:
                    lat, long = seenPoint
                    equiFracts = self.equirectFracts(lat, long, bounds=None)
                    if equiFracts is not None:
                        pixel = im.getpixel((int(equiFracts[0] * width), int(equiFracts[1] * height)))
                        product.putpixel((i, y), pixel)
            # rotate camera
            self.viewpoint.pose.azimuth += 360 / resolution[0]
        im.close()
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def rasterMesh(self, resourcesName, productFilename=None, mappingFunction=None, antiwrapThresh=0.0, smooth=None):
        # the function by which points in 3d space are mapped to the camera's detector can be specified, otherwise
        # use the camera's default mapping function:
        if mappingFunction is None:
            mappingFunction = self.camera.castRayIn
        # check if a mesh exists for the input resource and attempt to build one if not:
        meshPath = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'regular_triangular_mesh.npy')
        if not os.path.exists(meshPath):
            print('No mesh exists for resource {0}. Attempting to build mesh:'.format(resourcesName))
            self.viewpoint.landscape.buildMeshFromDTM(resourcesName)
        # load the mesh:
        print('Loading mesh from file...', end="")
        mesh = np.load(meshPath)
        print('\rLoaded mesh from file.')
        meshShape = mesh.shape[:2]
        # get boolean array of which mesh vertices contain no data (nan):
        noData = np.any(np.isnan(mesh), 2)
        # create Vec3 object of all (but only) the mesh vertices which contain data:
        vertices = Vec3((mesh[:, :, 0][~noData], mesh[:, :, 1][~noData], mesh[:, :, 2][~noData]))
        # create array to store distances to every vertex with data. This will be used later for ordering the rendering.
        # the array has shape meshShape but vertices with no data have value nan:
        vertexDistances = np.full(meshShape, np.nan, dtype=float)
        vertexDistances[~noData] = (self.camera.position - vertices).length
        if False:
            outOfRange = vertexDistances > 15000
            vertexDistances[outOfRange] = np.nan
            noData[outOfRange] = True
        vertices = Vec3((mesh[:, :, 0][~noData], mesh[:, :, 1][~noData], mesh[:, :, 2][~noData]))
        # map the mesh's vertices from 3d coordinates to detector coordinates:
        print('Mapping mesh to image coordinates...', end="")
        # use the mapping function to get the xfracts and yfracts of the mesh vertices with data.
        # where vertices have no data, xfracts and yfracts have value nan:
        xfracts = np.full(meshShape, np.nan, dtype=float)
        yfracts = np.full(meshShape, np.nan, dtype=float)
        xfracts[~noData], yfracts[~noData] = mappingFunction(vertices)
        # in addition to the mesh nodes with no data, vertices which cannot be seen (according to the mapping function)
        # will also now have value nan. Get a boolean array of which vertices are not nan and therefore need rendering:
        renderableVerts = ~(np.isnan(xfracts) + np.isnan(yfracts))
        if False:
            renderableVerts = (xfracts > 0.21) * (xfracts < 0.21 + 1/36)
        # convert xfracts and yfracts into integer pixel values of the vertices' positions in final image:
        resolution = self.camera.resolution
        if smooth is not None:
            scale = int(max(1, smooth))
            resolution = (resolution[0] * scale, resolution[1] * scale)
        imageCoords = np.full(meshShape + (2,), np.nan, dtype=int)
        imageCoords[:, :, 0][renderableVerts] = xfracts[renderableVerts] * resolution[0]
        imageCoords[:, :, 1][renderableVerts] = yfracts[renderableVerts] * resolution[1]
        print('\rMapped mesh to image coordinates.')
        # mesh vertices make up triangular facets, which are what need to be drawn to render image. Each set of four
        # vertices makes two facets (here called lower and upper). Make array of these facets describing image
        # coordinates of facet vertices, distance to each facet and colour of each facet:
        facetArrayShape = (meshShape[0] - 1, meshShape[1] - 1, 2, 11)
        facets = np.full(facetArrayShape, np.nan, dtype=float)
        facets[:, :, 0, 0:2] = imageCoords[:-1, :-1, :]  # lower x1, y1
        facets[:, :, 0, 2:4] = imageCoords[:-1, 1:, :]  # lower x2, y2
        facets[:, :, 0, 4:6] = imageCoords[1:, :-1, :]  # lower x3, y3
        facets[:, :, 1, 0:2] = imageCoords[1:, 1:, :]  # upper x1, y1
        facets[:, :, 1, 2:4] = imageCoords[:-1, 1:, :]  # upper x2, y2
        facets[:, :, 1, 4:6] = imageCoords[1:, :-1, :]  # upper x3, y3
        # let the defining distance of a facet be the minimum distance to any of its three vertices.
        # make an array containing the distance to each vertex of each facet:
        distArrayShape = (meshShape[0] - 1, meshShape[1] - 1, 2, 3)
        facetDists = np.full(distArrayShape, np.nan, dtype=float)
        facetDists[:, :, 0, 0] = vertexDistances[:-1, :-1]
        facetDists[:, :, 0, 1] = vertexDistances[:-1, 1:]
        facetDists[:, :, 0, 2] = vertexDistances[1:, :-1]
        facetDists[:, :, 1, 0] = vertexDistances[1:, 1:]
        facetDists[:, :, 1, 1] = vertexDistances[:-1, 1:]
        facetDists[:, :, 1, 2] = vertexDistances[1:, :-1]
        # reduce the array to a single distance per facet, by taking the minimum distance for each facet:
        facetDists = np.nanmin(facetDists, 3)
        # put the facets' distances into the facets array (negating them such that farthest are later rendered first):
        facets[:, :, :, 6] = -facetDists[:, :, :]
        # load ori image in order to get colour of each facet:
        oriPath = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
        ori = np.array(Image.open(oriPath).convert('RGBA'))
        # get facet colours and put into facets array:
        facets[:, :, 0, 7:] = ori[:-1, :-1]
        facets[:, :, 1, 7:] = ori[:-1, :-1]
        # the shape of the mesh is no longer important, so flatten facets array into a simple list:
        facets = facets.reshape(-1, 11)
        # any facets where at least one of the vertices is nan cannot be rendered, so remove these:
        nonRenderable = np.any(np.isnan(facets[:, :6]), 1)
        facets = facets[~nonRenderable]
        # sub-pixel facets (facets where all vertices have same image pixel coordinate) needn't go through polygon
        # drawing function. It's faster to render these pixels through numpy:
        subcolumn = (facets[:, 0] == facets[:, 2]) * (facets[:, 0] == facets[:, 4])
        subrow = (facets[:, 1] == facets[:, 3]) * (facets[:, 1] == facets[:, 5])
        subpixel = subcolumn * subrow
        facets = facets[~subpixel]
        if antiwrapThresh != 0.0:
            # check for any facets 'wrapping round' the image edges and causing streaks, and remove them:
            antiwrapThresh = antiwrapThresh * resolution[0]
            wraps = np.abs(facets[:, 0] - facets[:, 2]) > antiwrapThresh
            facets = facets[~wraps]
            wraps = np.abs(facets[:, 2] - facets[:, 4]) > antiwrapThresh
            facets = facets[~wraps]
            wraps = np.abs(facets[:, 4] - facets[:, 0]) > antiwrapThresh
            facets = facets[~wraps]
        # sort facets in order of distance, so farther facets are rendered before nearer facets:
        facets = facets[np.argsort(facets[:, 6])]
        # numpy arithmetic no longer needed so convert facets list to integer values and python list type:
        facets = facets.astype(int).tolist()
        # iterate through facets list and draw them to image:
        product = Image.new('RGBA', resolution)
        draw = ImageDraw.Draw(product)
        nFacets = len(facets)
        for n, facet in enumerate(facets):
            print("\rDrawing image: {0}%".format(int(100 * n / nFacets)), end="")
            draw.polygon((tuple(facet[:2]), tuple(facet[2:4]), tuple(facet[4:6])), fill=tuple(facet[7:]))
        print("\rImage fully rendered.")
        facets = None
        if smooth is not None:
            radius = max(1, int(0.5 * smooth))
            product = product.filter(ImageFilter.GaussianBlur(radius=radius))
            product = product.resize(self.camera.resolution)
        # save image:
        if productFilename is not None:
            path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
            product.save(path)
            product.close()
        return None

    def rasterFrame(self, resourcesName, productFilename):
        """Renders full 3D frame image, by rastering mesh elements to image"""
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
        vecData = self.vectoriseDTMData(resourcesName)
        visImage = Image.open(path).convert('RGBA')
        rows = len(vecData)
        columns = len(vecData[0])
        print(rows, columns)
        visWidth, visHeight = visImage.size
        prodWidth, prodHeight = self.camera.resolution
        drawBuffer = []
        for row in range(rows - 1):
            print("\rRendering: {0}%".format(100 * row / (rows - 1)), end = "")
            for column in range(columns - 1):
                v1 = vecData[row][column]
                v2 = vecData[row][column + 1]
                v3 = vecData[row + 1][column]
                v4 = vecData[row + 1][column + 1]
                imF1 = self.camera.castRayIn(v1, cropToImageBoundaries=False)
                imF2 = self.camera.castRayIn(v2, cropToImageBoundaries=False)
                imF3 = self.camera.castRayIn(v3, cropToImageBoundaries=False)
                imF4 = self.camera.castRayIn(v4, cropToImageBoundaries=False)
                #print(imF1, imF2, imF3, imF4)
                if imF1 is not None and imF2 is not None and imF3 is not None and imF4 is not None:
                    im1 = (int(imF1[0] * prodWidth), int(imF1[1] * prodHeight))
                    im2 = (int(imF2[0] * prodWidth), int(imF2[1] * prodHeight))
                    im3 = (int(imF3[0] * prodWidth), int(imF3[1] * prodHeight))
                    im4 = (int(imF4[0] * prodWidth), int(imF4[1] * prodHeight))
                    dist1 = (((v1 + v2 + v3) / 3) - self.camera.position).length
                    dist2 = (((v4 + v2 + v3) / 3) - self.camera.position).length
                    try:
                        colour1 = visImage.getpixel((column, row))
                        colour2 = visImage.getpixel((column, row))
                    except:
                        colour1 = (0, 255, 0)
                        colour2 = colour1
                    entry1 = (im1, im2, im3, -dist1, colour1)
                    entry2 = (im4, im2, im3, -dist2, colour2)
                    drawBuffer += [entry1, entry2]
        #print(drawBuffer)
        #execute draw buffer in order
        product = Image.new('RGBA', (prodWidth, prodHeight))
        draw = ImageDraw.Draw(product)
        drawBuffer = sorted(drawBuffer, key=itemgetter(3))
        for element in drawBuffer:
            try:
                draw.polygon((element[0], element[1], element[2]), fill=element[4])
            except:
                pass
        print("\rRendered 100%.")
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        visImage.close()
        vecData.clear()
        return None

    def rasterPB_DEPRECATED(self, angularRange, resourcesName, productFilename, border=0):
        """Renders full 3D push-broom image, by rastering mesh elements to image"""
        # Setup image cone
        principalPointLength = self.camera.focalLength * (math.sin(math.pi - math.radians(self.viewpoint.pose.tilt)) / math.sin(math.radians(self.viewpoint.pose.tilt)))
        coneAxis = self.viewpoint.w
        conePoint = self.camera.imagePlane.intercept(self.camera.position, coneAxis)
        cone = geom.Cone(conePoint, coneAxis, self.viewpoint.pose.tilt, minLength=principalPointLength - (0.5 * self.camera.sensorHeight), maxLength=principalPointLength + (0.5 * self.camera.sensorHeight))
        verticalOffset = coneAxis * 100
        projectionPlane = geom.Plane(conePoint + verticalOffset, conePoint + self.viewpoint.u + verticalOffset, conePoint + self.viewpoint.v + verticalOffset)
        rotationCentre = conePoint + verticalOffset
        # Open surface image and DTM
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
        vecData = self.vectoriseDTMData(resourcesName)
        visImage = Image.open(path).convert('RGBA')
        rows = len(vecData)
        columns = len(vecData[0])
        def helper(point):
            v = point
            c = cone.rayIntersect(v, self.camera.position - v)
            if c is None:
                return None
            projectedPoint = projectionPlane.intercept(c, coneAxis)
            arm = projectedPoint - rotationCentre
            angle = self.viewpoint.u.anticlockAngleWith(arm, coneAxis)
            if angle is None:
                return None
            angle = math.degrees(angle)
            #dist = self.camera.sensorHeight - (c - conePoint).mag
            dist = (0.5 * self.camera.sensorHeight) + (principalPointLength - (c - conePoint).mag)
            xFract = (1 - (angle / 360)) % 1
            yFract = dist / self.camera.sensorHeight
            return (xFract, yFract)
        # Calculate draws
        prodHeight = self.camera.resolution[1]
        #prodWidth = int(angularRange * (prodHeight / self.camera.FOV[1]))
        prodWidth = 2 * prodHeight
        drawBuffer = []
        lim = 0.01
        for r in range(rows - (2 * border) -1):
            print(str(r) + '/' + str(rows - (2 * border) - 1))
            row = border + r
            for c in range(columns - (border * 2) - 1):
                column = border + c
                v1 = vecData[row][column]
                v2 = vecData[row][column + 1]
                v3 = vecData[row + 1][column]
                v4 = vecData[row + 1][column + 1]
                imF1 = helper(v1)
                imF2 = helper(v2)
                imF3 = helper(v3)
                imF4 = helper(v4)
                if imF1 is not None and imF2 is not None and imF3 is not None and imF4 is not None:
                    condition1 = abs(imF1[0] - imF2[0]) < lim and abs(imF1[0] - imF3[0]) < lim and abs(imF2[0] - imF3[0]) < lim
                    condition2 = abs(imF4[0] - imF2[0]) < lim and abs(imF4[0] - imF3[0]) < lim and abs(imF2[0] - imF3[0]) < lim
                    if condition1 or condition2:
                        im2 = (int(imF2[0] * prodWidth), int(imF2[1] * prodHeight))
                        im3 = (int(imF3[0] * prodWidth), int(imF3[1] * prodHeight))
                        if condition1:
                            im1 = (int(imF1[0] * prodWidth), int(imF1[1] * prodHeight))
                            dist = (((v1 + v2 + v3) / 3) - self.camera.position).mag
                            colour = visImage.getpixel((column, row))
                            entry = (im1, im2, im3, -dist, colour)
                            drawBuffer += [entry]
                        if condition2:
                            im4 = (int(imF4[0] * prodWidth), int(imF4[1] * prodHeight))
                            dist = (((v4 + v2 + v3) / 3) - self.camera.position).mag
                            colour = visImage.getpixel((column, row))
                            entry = (im4, im2, im3, -dist, colour)
                            drawBuffer += [entry]
        #execute draw buffer in order
        product = Image.new('RGBA', (prodWidth, prodHeight))
        draw = ImageDraw.Draw(product)
        drawBuffer = sorted(drawBuffer, key=itemgetter(3))
        for element in drawBuffer:
            draw.polygon((element[0], element[1], element[2]), fill=element[4])
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        visImage.close()
        vecData.clear()
        return None

    def vectoriseDTMData(self, resourcesName, customPath=None, toC=False):
        """Loads 3D surface mesh file to 2D array of Vec3 objects, and returns it.
        
        customPath can be used to override conventional file and directory structure, and load from an explicit path.
        if toC is True, returned array comprises Vec3_C objects (cython version of Vec3).
        """
        if customPath is not None:
            path = customPath
        else:
            path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'surface_mesh.npy')
        print('Loading raw mesh from file.')
        data = np.load(path)
        typeDescription = 'Vec3_C' if toC else 'Vec3'
        print('Finished loading raw mesh from file. \nConverting mesh to ' + typeDescription + ' array.')
        columns = len(data[0])
        rows = len(data)
        arr = []
        for row in range(rows):
            print('Converting raw mesh to ' + typeDescription + ' mesh: ' + str(int((row / rows) * 100)) + '%')
            rowArr = []
            for column in range(columns):
                if toC:
                    rowArr += [Vec3(data[row][column][:3]).toC]
                else:
                    rowArr += [Vec3(data[row][column][:3])]
            arr += [rowArr]
        return arr

    def flattenDTMData(self, vecData, projectionRay, projectionPlane, toC=False):
        """Projects a mesh of Vec3 points onto a plane, and returns 2D array of projected points
        
        if toC is True, returned array comprises Vec3_C objects (cython version of Vec3)
        """
        print('Projecting 3D mesh to planar mesh...')
        data = vecData
        columns = len(data[0])
        rows = len(data)
        arr = []
        for row in range(rows):
            print('Projecting 3D mesh to planar mesh: ' + str(int((row / rows) * 100)) + '%')
            rowArr = []
            for column in range(columns):
                point3D = data[row][column]
                point2D = projectionPlane.intercept(point3D, projectionRay)
                if toC:
                    rowArr += [point2D.toC]
                else:
                    rowArr += [point2D]
            arr += [rowArr]
        print('Projection complete.')
        return arr

    def loadQuadtree(self, resourcesName):
        root = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'quadtree')
        quadtree = []
        depth = len([name for name in os.listdir(root) if os.path.isfile(os.path.join(root, name))])
        print('Loading quadtree...')
        for n in range(depth):
            print('Loading layer ' + str(n + 1) + ' of ' + str(depth))
            path = os.path.join(root, 'layer_' + str(depth - n - 1) + '.npy')
            layer = np.load(path).tolist()
            quadtree = [layer] + quadtree
        print('Finished loading quadtree.')
        return quadtree

    def latLongGrid(self, resourcesName):
        params = self.viewpoint.landscape.readResourceParameters(resourcesName)
        bounds = [params['minLong'], params['maxLat'], params['maxLong'], params['minLat']]
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'DTM_img.png')
        DTM = Image.open(path)
        width, height = DTM.size
        DTM.close()
        longRange = bounds[2] - bounds[0]
        latRange = bounds[3] - bounds[1]
        grid = np.empty((height, width, 2))
        for i in range(width):
            longitude = bounds[0] + ((i / width) * longRange)
            for j in range(height):
                latitude = bounds[1] + ((j / height) * latRange)
                grid[j, i] = [longitude, latitude]
        return grid

    def quadtreeIntersect(self, vecData, srchDepth, initSrchRgn, rayOrigin, rayVector, finalMeshSrch=True):
        """Returns the result of an intersection test of a single ray with a surface mesh, using a quadtree acceleration structure.
        
        Surface mesh is stored as 2D array of Vec3 objects (the mesh vertices).
        Ray intersection is tested on successively smaller regions of the mesh, using a dynamic quadtree.
        """
        def quadrantCheck(left, top, right, bottom):
            midCol = int((left + right) * 0.5)
            midRow = int((top + bottom) * 0.5)
            # Quadrant 1, triangle 1
            v1 = vecData[top][left]
            v2 = vecData[top][midCol]
            v3 = vecData[midRow][left]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (left, top, midCol, midRow)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (left, top, midCol, midRow)
            # Quadrant 1, triangle 2
            v1 = vecData[midRow][midCol]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (left, top, midCol, midRow)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (left, top, midCol, midRow)
            # Quadrant 2, triangle 1
            v3 = vecData[top][right]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (midCol, top, right, midRow)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (midCol, top, right, midRow)
            # Quadrant 2, triangle 2
            v2 = vecData[midRow][right]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (midCol, top, right, midRow)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (midCol, top, right, midRow)
            # Quadrant 3, triangle 1
            v2 = vecData[midRow][left]
            v3 = vecData[bottom][left]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (left, midRow, midCol, bottom)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (left, midRow, midCol, bottom)
            # Quadrant 3, triangle 2
            v2 = vecData[bottom][midCol]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (left, midRow, midCol, bottom)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (left, midRow, midCol, bottom)
            # Quadrant 4, triangle 1
            v3 = vecData[midRow][right]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (midCol, midRow, right, bottom)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (midCol, midRow, right, bottom)
            # Quadrant 4, triangle 2
            v1 = vecData[bottom][right]
            if geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3)) is not None:
                return (midCol, midRow, right, bottom)
            #elif geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3)) is not None:
             #   return (midCol, midRow, right, bottom)
            return None
        m = srchDepth
        region = initSrchRgn
        failed = False
        while m > -1:
            region = quadrantCheck(region[0], region[1], region[2], region[3])
            if region is None:
                #print('None')
                if m != srchDepth:
                    failed = True
                return None, failed
            m -= 1
        #iterate through individual mesh triangles within smallest region
        if finalMeshSrch:
            for j in range(region[3] - region[1]):
                j = region[1] + j
                for i in range(region[2] - region[0]):
                    i = region[0] + i
                    v1 = vecData[j][i]
                    v2 = vecData[j][i + 1]
                    v3 = vecData[j + 1][i]
                    check = geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3))
                    if check is None:
                        check = geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3))
                    if check is not None:
                        return (i, j), failed
                    v1 = vecData[j + 1][i + 1]
                    check = geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3))
                    if check is None:
                        check = geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3))
                    if check is not None:
                        return (i, j), failed
        else:
            return (region[0], region[1]), failed
        return None, failed

    def planeProjectionIntercept(self, vecData3D, flatVecData, srchDepth, initSrchRgn, rayOrigin, rayVector, flatRayOrigin, flatRayVector, finalMeshSrch=True):
        """Returns the result of an intersection test of a single ray with a surface mesh, using a planar projection of mesh and ray
        
        Surface mesh, stored as 2D array of Vec3 objects, is projected onto a plane. Ray is also projected onto the plane.
        A 2D quadtree search of the projected mesh is used to check for intersection of ray and mesh.
        """
        vecData = flatVecData
        lineOrigin = flatRayOrigin
        line = flatRayVector
        def lineHit(v1, v2):
            intersect = geom.vec3_vec3Intersection(v1, v2 - v1, lineOrigin, line)
            if intersect is not None:
                return geom.isPointBetweenPoints(intersect, v1, v2)
            return None
        def quadrantCheck(left, top, right, bottom):
            results = [False, False, False, False]
            midCol = int((left + right) * 0.5)
            midRow = int((top + bottom) * 0.5)
            
            v1 = vecData[top][midCol]
            v2 = vecData[midRow][midCol]
            if lineHit(v1, v2):
                results[0] = True
                results[1] = True
            v1 = vecData[bottom][midCol]
            if lineHit(v1, v2):
                results[2] = True
                results[3] = True
            v1 = vecData[midRow][left]
            if lineHit(v1, v2):
                results[0] = True
                results[2] = True
            v1 = vecData[midRow][right]
            if lineHit(v1, v2):
                results[1] = True
                results[3] = True
            if not results[3]:
                v2 = vecData[bottom][right]
                if lineHit(v1, v2):
                    results[3] = True
                else:
                    v1 = vecData[bottom][midCol]
                    if lineHit(v1, v2):
                        results[3] = True
            if not results[0]:
                v1 = vecData[top][left]
                v2 = vecData[midRow][left]
                if lineHit(v1, v2):
                    results[0] = True
                else:
                    v2 = vecData[top][midCol]
                    if lineHit(v1, v2):
                        results[0] = True
            if not results[1]:
                v1 = vecData[top][right]
                v2 = vecData[top][midCol]
                if lineHit(v1, v2):
                    results[1] = True
                else:
                    v2 = vecData[midRow][right]
                    if lineHit(v1, v2):
                        results[1] = True
            if not results[2]:
                v1 = vecData[bottom][left]
                v2 = vecData[midRow][left]
                if lineHit(v1, v2):
                    results[2] = True
                else:
                    v2 = vecData[bottom][midCol]
                    if lineHit(v1, v2):
                        results[2] = True
            hitRgns = []
            if results[0]:
                hitRgns += [(left, top, midCol, midRow)]
            if results[1]:
                hitRgns += [(midCol, top, right, midRow)]
            if results[2]:
                hitRgns += [(left, midRow, midCol, bottom)]
            if results[3]:
                hitRgns += [(midCol, midRow, right, bottom)]
            if len(hitRgns) == 0:
                return None
            return hitRgns
        def findSubRgns(regions):
            subRgns = []
            for rgn in regions:
                subRgnsToAdd = quadrantCheck(rgn[0], rgn[1], rgn[2], rgn[3])
                if subRgnsToAdd is not None:
                    subRgns += subRgnsToAdd
            if len(subRgns) == 0:
                return None
            return subRgns
        
        m = srchDepth
        regions = [initSrchRgn]
        failed = False
        while m > -1:
            regions = findSubRgns(regions)
            if regions is None:
                #print('None')
                if m != srchDepth:
                    failed = True
                return None, failed
            m -= 1
        #iterate through individual mesh triangles within smallest region
        for region in regions:
            if finalMeshSrch:
                for j in range(region[3] - region[1]):
                    j = region[1] + j
                    for i in range(region[2] - region[0]):
                        i = region[0] + i
                        v1 = vecData3D[j][i]
                        v2 = vecData3D[j][i + 1]
                        v3 = vecData3D[j + 1][i]
                        check = geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3))
                        #if check is None:
                         #   check = geom.rayTriangleIntersection2(rayOrigin, rayVector, (v1, v2, v3))
                        if check is not None:
                            return (i, j), failed
                        v1 = vecData3D[j + 1][i + 1]
                        check = geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3))
                        #if check is None:
                         #   check = geom.rayTriangleIntersection(rayOrigin, rayVector, (v1, v2, v3))
                        if check is not None:
                            return (i, j), failed
        return None, failed

    def sphericalQuadtreeIntersect(self, latLongGrid, quadtree, srchDepth, rayOrigin, rayVector):
        def getChildRegions(region):
            '''Returns the child regions of the input region.'''
            left, top, right, bottom = region
            midCol = int((left + right) * 0.5)
            midRow = int((top + bottom) * 0.5)
            q1 = (left, top, midCol, midRow)
            q2 = (midCol, top, right, midRow)
            q3 = (left, midRow, midCol, bottom)
            q4 = (midCol, midRow, right, bottom)
            return [q1, q2, q3, q4]
        def checkRgns(regions, quadtreeCoords, depth):
            layer = quadtree[depth]
            results = []
            childQuadtreeCoords = []
            for n, region in enumerate(regions):
                left, top, right, bottom = region
                west, north = latLongGrid[top][left]
                east, south = latLongGrid[bottom][right]
                quadCoord = quadtreeCoords[n]
                innerAlt = layer[quadCoord * 2]
                outerAlt = layer[quadCoord * 2 + 1]
                #innerAlt, outerAlt = layer[quadCoord]
                outerIntersection = self.viewpoint.landscape.datum.rayIntersect(rayOrigin, rayVector, deltaH=outerAlt, givePolarCoords=True)
                5/0
                regionIsHit = False
                if outerIntersection is not None:
                    lo, la, _, _, _ = outerIntersection
                    print(lo, la)
                    hitResult = self.equirectFracts(la, lo, bounds=(west, north, east, south))
                    if hitResult is not None:
                        regionIsHit = True
                if not regionIsHit:
                    innerIntersection = self.viewpoint.landscape.datum.rayIntersect(rayOrigin, rayVector, deltaH=innerAlt, givePolarCoords=True)
                    if innerIntersection is not None:
                        lo, la, _, _, _ = innerIntersection
                        hitResult = self.equirectFracts(la, lo, bounds=(west, north, east, south))
                        if hitResult is not None:
                            regionIsHit = True
                if regionIsHit:
                    # ray intersects this quadrant of this layer of the quadtree
                    # add this quadrant's children to the results, to be checked at the next level
                    results += getChildRegions(region)
                    baseChildCoord = 4 * quadCoord
                    childQuadtreeCoords += [baseChildCoord, baseChildCoord + 1, baseChildCoord + 2, baseChildCoord + 3]
            return results, childQuadtreeCoords
        rows, columns, _ = latLongGrid.shape
        regions = getChildRegions((0, 0, columns - 1, rows - 1))
        quadtreeCoords = [0, 1, 2, 3]
        depth = 0
        maxDepth = len(quadtree)
        while depth < maxDepth:
            regions, quadtreeCoords = checkRgns(regions, quadtreeCoords, depth)
            depth += 1
        return regions

    def castFrame(self, resourcesName, productFilename):
        """Renders a full-3D frame image and saves the result to file."""
        # define functions
        def enclosingRegion(coord, dimension, dataSize):
            """Returns bounds of a region of a mesh surrounding a single mesh coordinate."""
            i, j = coord
            d = dimension
            rgn = [i-d, j-d, i+d, j+d]
            if i < d:
                rgn[0] = 0
            if j < d:
                rgn[1] = 0
            if i+d >= dataSize[0]:
                rgn[2] = dataSize[0] - 1
            if j+d >= dataSize[1]:
                rgn[3] = dataSize[1] - 1
            return rgn
        def maxGridSearchDepth(width, height):
            """Returns the maximum possible depth of a grid's quadtree structure"""
            return int(math.log2(min(width, height))) + 1
        # create instance of Renderer_C class, which handles rendering in C
        rendererC = renderer_C.Renderer()
        # load neccessary resources: surface mesh and ORI
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
        vecData = self.vectoriseDTMData(resourcesName, toC=True)
        visImage = Image.open(path).convert('RGBA')
        meshWidth, meshHeight = visImage.size
        # setup neccessary resources: plane projected mesh
        # this will be used for plane-projection ray-mesh intersection queries
        v1 = vecData[0][0].toPy()
        v2 = vecData[0][len(vecData[0]) - 1].toPy()
        v3 = vecData[len(vecData) - 1][0].toPy()
        projRay = (v2 - v1).cross(v3 - v1) * 100
        projPlane = geom.Plane(v1 + projRay, v2 + projRay, v3 + projRay)
        planarVecData = self.flattenDTMData(vecData, projRay, projPlane, toC=True)
        # setup camera related variables
        resolution = self.camera.resolution
        rayOrigin = self.camera.position
        rayOriginC = rayOrigin.toC
        planarRayOrigin = projPlane.intercept(self.camera.position, projRay).toC
        # setup mesh search related variables
        fullSearchDepth = maxGridSearchDepth(meshWidth, meshHeight)
        fullSearchRegion = (0, 0, meshWidth - 1, meshHeight - 1)
        # create image to render to
        product = Image.new('RGBA', resolution)
        # cast rays for each pixel and render according to mesh intersection location
        for y in range(resolution[1]):
            yfract = y / resolution[1]
            print('Rendering image pixels: ' + str(int(yfract * 100)) + '%')
            # define variable 'localRgn' for storing a localised search region, based on previous ray traces from same image row
            localSearchRegion = None
            # initialise a variable to keep track of how many rays have been cast since the localSearchRegion was last updated
            distToLocale = 0
            for x in range(resolution[0]):
                xfract = x / resolution[0]
                # get this pixel's ray from camera
                rayVector = self.camera.castRayOut(xfract, yfract)
                # define search parameters
                searchRegion = fullSearchRegion
                searchDepth = fullSearchDepth
                # check if mesh search can be localised based on a previous casting result, and setup search parameters if so
                if localSearchRegion is not None:
                    searchRegion = localSearchRegion
                    searchDepth = maxGridSearchDepth(localSearchRegion[2]-localSearchRegion[0], localSearchRegion[3]-localSearchRegion[1])
                # perform direct quadtree search of mesh for ray intersection
                meshIntercept, traversalFailed = rendererC.castRayToMesh(vecData, searchDepth, searchDepth, searchRegion, rayOriginC, rayVector.toC, interpThreshold=0)
                if meshIntercept is None:
                    if traversalFailed:
                        # ray intersects mesh, but traversal of quadtree failed, giving false negative
                        # localise intercept using simple spheroid intersect
                        spheroidIntercept = self.seenPoint(rayVector)
                        if spheroidIntercept is not None:
                            if localSearchRegion is None:
                                distToLocale = 0
                                localSearchRegion = enclosingRegion(spheroidIntercept, 100, (meshWidth, meshHeight))
                        # project pixel's ray onto the projection plane, for use with plane-projection ray-mesh method
                        planarRay = projPlane.projectVectorOnto(rayOrigin, rayVector, projRay).toC
                        # if a localised searchRegion exists, search it for intercept with the plane-projection method
                        if localSearchRegion is not None:
                            interceptedRegions = rendererC.castRayToPlanarMesh(planarVecData, searchDepth-2, searchDepth-2, searchRegion, planarRayOrigin, planarRay)
                            if interceptedRegions is not None:
                                # check each candidate region for ray-mesh intercept using dynamic quadtree
                                localisedSearchDepth = maxGridSearchDepth(interceptedRegions[0][2]-interceptedRegions[0][0], interceptedRegions[0][3]-interceptedRegions[0][1])
                                for candidateRegion in interceptedRegions:
                                    meshIntercept, traversalFailed = rendererC.castRayToMesh(vecData, localisedSearchDepth, localisedSearchDepth, candidateRegion, rayOriginC, rayVector.toC)
                                    if meshIntercept is not None and not traversalFailed:
                                        # ray-mesh interception has been found within a candidate region
                                        # stop searching remaining candidate regions
                                        break
                if distToLocale == 10:
                    # localSearchRegion has not been updated for 10 rays, and is no longer valid
                    localSearchRegion = None
                if meshIntercept is not None:
                    # pixel's ray strikes mesh element, so retrieve colour of the element from ORI and render to product
                    localSearchRegion = enclosingRegion(meshIntercept, 20, (meshWidth, meshHeight))
                    distToLocale = 0
                    pixel = visImage.getpixel(meshIntercept)
                    product.putpixel((x, y), pixel)
        # clean up data and save image to file
        visImage.close()
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def castPB(self, xfract, dropPerScanLine, resourcesName, productFilename):
        """Renders a full-3D push-broom image and saves the result to file."""
        # define functions
        def enclosingRegion(coord, dimension, dataSize):
            """Returns bounds of a region of a mesh surrounding a single mesh coordinate."""
            i, j = coord
            d = dimension
            rgn = [i-d, j-d, i+d, j+d]
            if i < d:
                rgn[0] = 0
            if j < d:
                rgn[1] = 0
            if i+d >= dataSize[0]:
                rgn[2] = dataSize[0] - 1
            if j+d >= dataSize[1]:
                rgn[3] = dataSize[1] - 1
            return rgn
        def maxGridSearchDepth(width, height):
            """Returns the maximum possible depth of a grid's quadtree structure"""
            return int(math.log2(min(width, height))) + 1
        # Open surface image and DTM
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
        vecData = self.vectoriseDTMData(resourcesName, toC=True)
        visImage = Image.open(path).convert('RGBA')
        visWidth, visHeight = visImage.size
        # create instance of Renderer_C class, which handles rendering in C
        rendererC = renderer_C.Renderer()
        # load neccessary resources: surface mesh and ORI
        path = os.path.join('..', 'data', 'resources', 'processed', resourcesName, 'ORI_img.png')
        vecData = self.vectoriseDTMData(resourcesName, toC=True)
        visImage = Image.open(path).convert('RGBA')
        meshWidth, meshHeight = visImage.size
        # setup neccessary resources: plane projected mesh
        # this will be used for plane-projection ray-mesh intersection queries
        v1 = vecData[0][0].toPy()
        v2 = vecData[0][len(vecData[0]) - 1].toPy()
        v3 = vecData[len(vecData) - 1][0].toPy()
        projRay = (v2 - v1).cross(v3 - v1) * 100
        projPlane = geom.Plane(v1 + projRay, v2 + projRay, v3 + projRay)
        planarVecData = self.flattenDTMData(vecData, projRay, projPlane, toC=True)
        # setup camera related variables
        resolution = (self.camera.resolution[1] * 2, self.camera.resolution[1])
        rayOrigin = self.camera.position
        rayOriginC = rayOrigin.toC
        planarRayOrigin = projPlane.intercept(self.camera.position, projRay).toC
        # setup mesh search related variables
        fullSearchDepth = maxGridSearchDepth(meshWidth, meshHeight)
        fullSearchRegion = (0, 0, meshWidth - 1, meshHeight - 1)
        # create image to render to
        product = Image.new('RGBA', resolution)
        # cast rays for each pixel and render according to mesh intersection location
        for i in range(resolution[0]):
            print('Rendering image pixels: ' + str(int((i / resolution[0]) * 100)) + '%')
            # define variable 'localRgn' for storing a localised search region, based on previous ray traces from same image row
            localSearchRegion = None
            # initialise a variable to keep track of how many rays have been cast since the localSearchRegion was last updated
            distToLocale = 0
            for y in range(resolution[1]):
                yfract = y / resolution[1]
                # get this pixel's ray from camera
                rayVector = self.camera.castRayOut(xfract, yfract)
                # define search parameters
                searchRegion = fullSearchRegion
                searchDepth = fullSearchDepth
                # check if mesh search can be localised based on a previous casting result, and setup search parameters if so
                if localSearchRegion is not None:
                    searchRegion = localSearchRegion
                    searchDepth = maxGridSearchDepth(localSearchRegion[2]-localSearchRegion[0], localSearchRegion[3]-localSearchRegion[1])
                # perform direct quadtree search of mesh for ray intersection
                meshIntercept, traversalFailed = rendererC.castRayToMesh(vecData, searchDepth, searchDepth, searchRegion, rayOriginC, rayVector.toC, interpThreshold=0)
                if meshIntercept is None:
                    if traversalFailed:
                        # ray intersects mesh, but traversal of quadtree failed, giving false negative
                        # localise intercept using simple spheroid intersect
                        spheroidIntercept = self.seenPoint(rayVector)
                        if spheroidIntercept is not None:
                            if localSearchRegion is None:
                                distToLocale = 0
                                localSearchRegion = enclosingRegion(spheroidIntercept, 100, (meshWidth, meshHeight))
                        # project pixel's ray onto the projection plane, for use with plane-projection ray-mesh method
                        planarRay = projPlane.projectVectorOnto(rayOrigin, rayVector, projRay).toC
                        # if a localised searchRegion exists, search it for intercept with the plane-projection method
                        if localSearchRegion is not None:
                            interceptedRegions = rendererC.castRayToPlanarMesh(planarVecData, searchDepth-2, searchDepth-2, searchRegion, planarRayOrigin, planarRay)
                            if interceptedRegions is not None:
                                # check each candidate region for ray-mesh intercept using dynamic quadtree
                                localisedSearchDepth = maxGridSearchDepth(interceptedRegions[0][2]-interceptedRegions[0][0], interceptedRegions[0][3]-interceptedRegions[0][1])
                                for candidateRegion in interceptedRegions:
                                    meshIntercept, traversalFailed = rendererC.castRayToMesh(vecData, localisedSearchDepth, localisedSearchDepth, candidateRegion, rayOriginC, rayVector.toC)
                                    if meshIntercept is not None and not traversalFailed:
                                        # ray-mesh interception has been found within a candidate region
                                        # stop searching remaining candidate regions
                                        break
                if distToLocale == 20:
                    # localSearchRegion has not been updated for 10 rays, and is no longer valid
                    localSearchRegion = None
                if meshIntercept is not None:
                    # pixel's ray strikes mesh element, so retrieve colour of the element from ORI and render to product
                    localSearchRegion = enclosingRegion(meshIntercept, 40, (meshWidth, meshHeight))
                    distToLocale = 0
                    pixel = visImage.getpixel(meshIntercept)
                    product.putpixel((i, y), pixel)
            # rotate camera
            self.viewpoint.pose.azimuth += 360 / resolution[0]
        # clean up data and save image to file
        visImage.close()
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def phaseAngleMap(self, minAlt, maxAlt, resourcesName, productFilename, longitude=None, latitude=None):
        sun = self.viewpoint.landscape.latLongAlt2XYZ(-2, 5, 228e6)
        if longitude is None:
            longitude = self.viewpoint.longitude
        if latitude is None:
            latitude = self.viewpoint.latitude
        cameraTop = self.viewpoint.landscape.latLongAlt2XYZ(latitude, longitude, maxAlt)
        cameraBottom = self.viewpoint.landscape.latLongAlt2XYZ(latitude, longitude, minAlt)
        def vecToPy(vec3_C):
            return Vec3((vec3_C.x, vec3_C.y, vec3_C.z))
        vecData = self.vectoriseDTMData(resourcesName, toC=True)
        columns = len(vecData[0])
        rows = len(vecData)
        arr = np.empty((rows, columns, 2))
        masterMinAng = 360
        masterMaxAng = -360
        maxPhaseDiff = -720
        minPhaseDiff = 720
        for row in range(rows):
            print(str(row) + '/' + str(rows))
            for column in range(columns):
                minAng = 360
                maxAng = 0
                surfPoint = vecToPy(vecData[row][column])
                surfToSun = sun - surfPoint
                surfToCamTop = cameraTop - surfPoint
                surfToCamBottom = cameraBottom - surfPoint
                normal1 = surfToSun.cross(surfToCamTop)
                normal2 = surfToSun.cross(surfToCamBottom)
                ang1 = surfToSun.directionalAngleWith(surfToCamTop, normal1)
                ang2 = surfToSun.directionalAngleWith(surfToCamBottom, normal2)
                angTop = math.degrees(surfToSun.angleWith(surfToCamTop))
                angBottom = math.degrees(surfToSun.angleWith(surfToCamBottom))
                if ang1 * ang2 <= 0:
                    minAng = 0
                    maxAng = max(angTop, angBottom)
                else:
                    minAng = min(angTop, angBottom)
                    maxAng = max(angTop, angBottom)
                if maxPhaseDiff < maxAng - minAng:
                    maxPhaseDiff = maxAng - minAng
                if minPhaseDiff > maxAng - minAng:
                    minPhaseDiff = maxAng - minAng
                arr[row][column] = [minAng, maxAng]
        resolution = self.camera.resolution
        rayOrigin = self.camera.position.toC
        searchDepth = int(math.log2(min(rows, columns))) + 1
        startRegion = (0, 0, columns - 1, rows - 1)
        renderer = renderer_C.Renderer()
        product = Image.new('RGBA', resolution)
        for y in range(resolution[1]):
            print('y: ' + str(y) + '/' + str(resolution[1]))
            yfract = y / resolution[1]
            for x in range(resolution[0]):
                xfract = x / resolution[0]
                rayVector = self.camera.castRay(xfract, yfract)
                intercept, failed = renderer.castRayToMesh(vecData, searchDepth-4, searchDepth-4, startRegion, rayOrigin, rayVector.toC, interpThreshold=0)
                if intercept is not None and not failed:
                    minAng, maxAng = arr[intercept[1]][intercept[0]]
                    #n = int(((maxAng - minAng) / (masterMaxAng - masterMinAng)) * 255)
                    r = int((maxAng / maxPhaseDiff) * 152)
                    g = int((maxAng / maxPhaseDiff) * 38)
                    b = int((maxAng / maxPhaseDiff) * 62)
                    product.putpixel((x, y), (r, g, b, 255))
        print(minPhaseDiff, maxPhaseDiff)
        path = os.path.abspath(os.path.join('..', 'data', 'products', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

    def seenPoint(self, rayVector, deltaH=0):
        """Calculates lat long coordinates of the point on the spheroid surface seen by a given ray"""
        intersect = self.viewpoint.landscape.datum.rayIntersect(self.camera.position, rayVector, deltaH=deltaH, givePolarCoords=True)
        if intersect is not None:
            long, lat = intersect[:2]
            return (lat, long)
        return None

    def equirectFracts(self, targetLat, targetLong, bounds=None):
        """converts a lat long coordinate into a fractional coordinate for an equirectangular bounding box
        
        bounds should be of form [west, north, east, south]
        """
        long = targetLong % 360
        lat = targetLat
        if bounds is None:
            x = (targetLong % 360) / 360
            y = 0.5 - (targetLat / 180)
            return (x, y)
        west, north, east, south = bounds
        west = west % 360
        east = east % 360
        if east < west:
            longRange = east + (360 - west)
            if east <= long < west:
                return None
            if long >= west:
                x = (long - west) / longRange
            else:
                x = 1 - ((east - long) / longRange)
        else:
            longRange = east - west
            x = (long - west) / longRange
        y = 1 - ((lat - south) / (north - south))
        if 0 <= x < 1 and 0 <= y < 1:
            return (x, y)
        return None