import math
from vec3 import Vec3
from ray import Ray
from raytracer import RayTracer

class Scene:
    def __init__(self, sceneObjects, lightSource):
        self._sceneObjects = sceneObjects
        self._opaqueObjects = []
        for sceneObject in self._sceneObjects:
            if sceneObject.opaque:
                self._opaqueObjects += [sceneObject]
        self._lightSource = lightSource

    def getRayTraversal(self, ray: 'Ray') -> list:
        """Return an ordered list of all the interactions the input ray has with this scene.
        This function tests the input ray against all the objects within the scene, and returns a list of its
        interactions (if any) with these objects.

        :param ray: the ray to be tested.
        :return: interactions: ordered list of any interactions.
        """
        # track the depth at which the ray terminates, i.e. the depth of the first opaque surface.
        # if the ray hits no opaque surfaces, it never terminates, and the distance is set at infinity:
        terminalDepth = math.inf
        # populate a list of all the interactions the ray has with the scene, by iterating over all scene objects:
        interactions = []
        for sceneObject in self._sceneObjects:
            # if the ray interacts with a scene object (or an object group), add the interaction(s) to the list:
            _, interactionsToAdd = sceneObject.intersect(ray)
            interactions += interactionsToAdd
            # check if any of the new interactions are with an opaque surface and adjust terminalDepth accordingly:
            for interaction in interactionsToAdd:
                if interaction['object'].opaque:
                    terminalDepth = min(terminalDepth, interaction['depth'])
        # calculate ray's terminal point from the terminal depth:
        terminalPoint = ray.origin + terminalDepth * ray.direction
        # order the interactions in increasing depth:
        interactions.sort(key=lambda x: x['depth'])
        # truncate the list of interactions at the first opaque object:
        for n, interaction in enumerate(interactions):
            if interaction['object'].opaque:
                del interactions[n + 1:]
                break
            # if the interaction is with a translucent object which extends beyond the ray's terminal point, the
            # interaction's length and endPoint must be changed to match the terminal point:
            if interaction['depth'] + interaction['length'] > terminalDepth:
                interaction['length'] = terminalDepth - interaction['depth']
                interaction['endPoint'] = terminalPoint
        return interactions

    def fluxFromLight(self, point, light, includeShadows=True):
        """Returns the spectral photon rate received per unit area at the input point. Units: photons s^-1 m^-2 nm^-1.
        Also returns the direction vector of the incident light.

        :param point:
        :param light:
        :param includeShadows:
        :return:
        """
        # return as (flux spectrum, normalised direction) tuple
        if includeShadows:
            for sceneObject in self._opaqueObjects:
                if sceneObject.occludes(light.position, point):
                    return None, None
        lightPath = point - light.position
        incidentFlux = lightPath.length ** -2 * light.totalFlux  # photons s^-1 m^-2 nm^-1
        return incidentFlux, lightPath.norm

    def probe(self, ray):
        return RayTracer().traceRay(ray, self).intensity#.toSpectralRadiance

    @property
    def lightSource(self):
        return self._lightSource

    @property
    def opaqueObjects(self):
        return self._opaqueObjects