import os
import math
import numpy as np
from PIL import Image
from vec3 import Vec3
from frame import Frame
from primitives import Cylinder, RoundedCylinder, Spheroid, Cone, RoundedCone
from boundingBoxes import AABoundingBox
from ray import Ray, LightRay
from scene import Scene

class SceneObject:
    def __init__(self):
        self._primitive = None
        self._opaque = True
        self._description = 'SceneObject[base class]'

    def intersect(self, ray):
        """Function to be overridden by subclasses."""
        print('Function has not been overidden by subclass. SceneObject subclass must implement its own intersect function.')
        return [], []

    def intersect_numpy(self, ray):
        return []

    def occludes(self, viewpoint, target, bias=1e-6):
        """Returns whether or not this SceneObject blocks the input point target from view of input viewpoint.

        If not overridden by a subclass of SceneObject, this function returns False for all inputs
        """
        return False    

    def phaseFunction(self, phaseAngle):
        """Returns the fraction of light reflected, scattered or transmitted to an observation as a function of its phase angle.
        
        If not overridden by a subclass of SceneObject, this function returns 1 for all inputs.
        """
        return 1

    @property
    def description(self):
        return self._description

    @property
    def opaque(self):
        return self._opaque

    @property
    def position(self):
        return self._primitive.position

    @property
    def frame(self):
        return self._primitive.frame

    @property
    def boundingBox(self):
        return self._primitive.boundingBox

    def __str__(self):
        return '{0} at {1}'.format(self._description, self.position)

    def __repr__(self):
        return '{0}(frame: {1}) id={2}'.format(self._description, self.frame, id(self))

class ObjectGroup(SceneObject):
    def __init__(self, sceneObjects):
        self._sceneObjects = sceneObjects
        super().__init__()
        constituentBoxes = []
        for sceneObject in sceneObjects:
            constituentBoxes += [sceneObject.boundingBox]
        self._aabb = AABoundingBox.fromAABoundingBoxes(constituentBoxes)
        self._description = 'ObjectGroup'

    def intersect(self, ray):
        if not self._aabb.isHitBy(ray):
            return [], []
        hits, interactions = [], []
        for sceneObject in self._sceneObjects:
            objectHits, interaction = sceneObject.intersect(ray)
            hits += objectHits
            interactions += interaction
        return hits, interactions

    def occludes(self, viewpoint, target, bias=1e-6):
        # Need to optimise this to increase speed
        for sceneObject in self._sceneObjects:
            if sceneObject.occludes(viewpoint, target, bias=1e-6):
                return True
        return False

    @property
    def position(self):
        return self._aabb.centre

    @property
    def frame(self):
        return Frame.universal()

    @property
    def boundingBox(self):
        return self._aabb

    def __str__(self):
        return '{0} with {1} objects'.format(self._description, len(self._sceneObjects))

    def __repr__(self):
        return '{0}(object count: {1}) id={2}'.format(self._description, len(self._sceneObjects), id(self))

class OpaqueObject(SceneObject):
    def __init__(self, albedo=1, textureMap=None, normalMap=None):
        self._albedo = albedo
        self._textureMap = None
        self._textureSize = None
        if textureMap is not None:
            path = os.path.abspath(os.path.join('..', 'data', 'resources', 'processed', textureMap)) + '.png'
            self._textureMap = Image.open(path).convert('RGB')
            self._textureSize = self._textureMap.size
        self._normalMap = None
        self._normalMapSize = None
        if normalMap is not None:
            path = os.path.abspath(os.path.join('..', 'data', 'resources', 'processed', normalMap)) + '.png'
            self._normalMap = Image.open(path).convert('RGB')
            self._normalMapSize = self._normalMap.size
        super().__init__()
        self._description = 'OpaqueObject[base class]'

    def intersect(self, ray):
        hits = self._primitive.intersect(ray)
        if not hits:
            return [], []
        hit = hits[0]
        if hit['type'] == 'internal':
            hit = hits[1]
        hit['object'] = self
        #
        interaction = dict()
        interaction['object'] = self
        interaction['depth'] = hit['depth']
        interaction['startPoint'] = hit['position']
        return [hit], [interaction]

    def intersect_numpy(self, rays):
        hits = self._primitive.intersect_numpy(rays)
        # remove any internal hits
        hits[0]['depth'] = 1e20
        hits[0]['position'] = Vec3((1e20, 1e20, 1e20))
        hits[0]['type'] = 'noHit'
        hits[0]['object'] = None
        for hit in hits[1:]:
            hit['object'] = np.where(hit['type'] != 'noHit', self, None)
        return hits

    def textureCoordinate(self, point):
        """Function to be overidden by subclass.
        
        Returns the pixel coordinate of a texture image from a point in world space.
        """
        return (0, 0)

    def occludes(self, viewpoint, target, bias=1e-6):
        """Returns whether or not this SceneObject blocks the input point target from view of input viewpoint."""
        target = target + (bias * self._primitive.surfaceNormal(target))
        ray = Ray(target, viewpoint - target)
        return False
        return self._primitive.intersectBool(ray)

    def reflect(self, point, indcidentRay, reflectionDirection):
        # incidentRay is a vector whose magnitude is the incident intensity in photons / [m^2 s]
        # reflectionDirection is a unit vector describing direction of reflection for which intensity should be calculated
        surfaceNormal = self.surfaceNormal(point)
        # return the reflected ray, with an intensity given by Lambertian reflection
        reflectedRay = LightRay(point, reflectionDirection)
        reflectedRay.intensity = abs(indcidentRay.dot(surfaceNormal) * self._localAlbedo(point) * 0.3183)
        return reflectedRay

    def reflectNEW(self, incidentFlux, atPoint, directionOfIncidence):
        # lambertian reflection
        # incidentFlux PhotonRateSDC with units photons s^-1 m^-2 nm^-1
        # direction vectors are both normalised
        n = self.surfaceNormal(atPoint)
        reflectionFactor = 0.318 * abs(directionOfIncidence.dot(n)) * self._localAlbedo(atPoint)
        reflectedFlux = reflectionFactor * incidentFlux  # photons s^-1 m^-2 sr^-1 nm^-1
        return reflectedFlux

    def _localAlbedo(self, point):
        if self._textureMap is None:
            return self._albedo
        s = self._textureSize
        x, y = self.textureCoordinate(point)
        return self._textureMap.getpixel((int(x * s[0]), int(y * s[1])))[0] * self._albedo * 0.003922

    def surfaceNormal(self, point):
        if self._normalMap is None:
            return self._primitive.surfaceNormal(point)
        column, row = self.textureCoordinate(point)
        column = int(column * self._normalMapSize[0])
        row = int(row * self._normalMapSize[1])
        r, g, b = self._normalMap.getpixel((column, row))
        x = (r * 0.007844) - 1
        y = (g * 0.007844) - 1
        z = (b * 0.007844) - 1
        return self.frame.toWorld(Vec3((x, y, z)))

    def shadowRegion(self, light, extent=1e15):
        r = self._primitive.bulkRadius
        if r == 0:
            return None
        R = self.position - light.position
        #R = light.position - self.position
        angle = math.degrees(math.atan(r / R.length))
        cone = Cone(Frame.withW(R, light.position), angle, extent, minLength=R.length)
        return ShadowRegion(cone)

class OpaqueSpheroid(OpaqueObject):
    def __init__(self, frame, radii, albedo=1, textureMap=None, normalMap=None):
        super().__init__(albedo=albedo, textureMap=textureMap, normalMap=normalMap)
        self._primitive = Spheroid(frame, radii)
        self._description = 'OpaqueSphere'

    def textureCoordinate(self, point):
        """Returns the fractional pixel coordinate of a texture image from a point in world space."""
        _, azimuth, polar = self.frame.fromWorldToSpherical(point)
        return (azimuth * 0.15915, polar * 0.31830)

    def generateNormalMap(self, strength, bumpMapFilename, productFilename):
        """Generates and saves a normal map in this object's frame space."""
        path = os.path.abspath(os.path.join('..', 'data', 'resources', 'processed', bumpMapFilename)) + '.png'
        bm = Image.open(path).convert('RGB')
        size = bm.size
        invSize = (1 / size[0], 1 / size[1])
        # build a node mesh of displaced points from the bump map
        nodes = np.full((size[1], size[0]), Vec3.zeroVector())
        for row in range(size[1]):
            print('Calculating mesh nodes: {0}/{1} rows.'.format(row, size[1]))
            polar = row * invSize[1] * math.pi
            for column in range(size[0]):
                azimuth = column * invSize[0] * 2 * math.pi
                bump = ((bm.getpixel((column, row))[0] * 0.003922) - 0.5) * strength
                r = self._primitive.localRadius(azimuth, polar) * (1 + bump)
                nodes[row, column] = Frame.universal().polarCoordsInverse(r, azimuth, polar)
        # calculate normals from the node mesh
        product = Image.new('RGB', size)
        def convertVal(value):
            return int((value + 1) * 0.5 * 255)
        for row in range(size[1]):
            print('Generating normal map: {0}/{1} rows.'.format(row, size[1]))
            rn = ((row - 1) % size[1], (row + 1) % size[1])
            for column in range(size[0]):
                cn = ((column - 1) % size[0], (column + 1) % size[0])
                pCentre = nodes[row, column]
                p1 = nodes[rn[0], cn[0]]
                p2 = nodes[rn[1], cn[1]]
                p3 = nodes[rn[1], cn[0]]
                p4 = nodes[rn[0], cn[1]]
                n1 = ((p2 - p1).cross(pCentre).cross(p2 - p1)).norm
                n2 = ((p4 - p3).cross(pCentre).cross(p4 - p3)).norm
                n = (n1 + n2).norm
                red = convertVal(n.x)
                green = convertVal(n.y)
                blue = convertVal(n.z)
                product.putpixel((column, row), (red, green, blue))
        path = os.path.abspath(os.path.join('..', 'data', 'resources', 'processed', productFilename)) + '.png'
        product.save(path)
        product.close()
        return None

class OpaqueCone(OpaqueObject):
    def __init__(self, frame, angle, length, albedo=1):
        super().__init__(albedo=albedo)
        self._primitive = Cone(frame, angle, length)
        self._description = 'OpaqueCone'

class TranslucentObject(SceneObject):
    def __init__(self, attenuation=0.01, scatteringEfficiency=0.1, dop=0):
        self._attenuation = attenuation
        self._scatteringEfficiency = scatteringEfficiency
        self._dop = dop
        self._centreOfMass = None
        super().__init__()
        self._opaque = False
        self._description = 'TranslucentObject[base class]'

    def intersect(self, ray):
        hits = self._primitive.intersect(ray)
        if not hits:
            return [], []
        for hit in hits:
            hit['object'] = self
        #
        interaction = dict()
        interaction['object'] = self
        interaction['depth'] = hits[0]['depth']
        interaction['startPoint'] = hits[0]['position']
        try:
            interaction['endPoint'] = hits[1]['position']
        except:
            #print(hits)
            return [], []
        interaction['length'] = (interaction['endPoint'] - interaction['startPoint']).length
        return hits, [interaction]

    def intersect_numpy(self, rays):
        hits = self._primitive.intersect_numpy(rays)
        for hit in hits:
            hit['object'] = np.where(hit['type'] != 'noHit', self, None)
        return hits

    def pathTransmittance(self, pathLength, pathStart, pathEnd):
        """Returns the fraction of light that hits the medium when travelling a length through it."""
        # coefficient must be negative
        coefficient = -1e-5
        return math.exp(pathLength * self._attenuation * coefficient)

    def phaseFunction(self, phaseAngle):
        """Returns the fraction of inward intensity that is scattered at the input phase angle, and its degree of polarisation."""
        a = math.degrees(phaseAngle)
        h = 1.36 * self._dop
        m = 0.0085
        ai = 24
        al = 180
        dop = 0.01 * (h * (1 - math.exp(-m * a)) * (a - ai) * (a - al)) / ((1 - math.exp(-m * ai)) * (ai - al)) * math.sin((3.1416 - phaseAngle) * 0.5)
        return (1 * self._scatteringEfficiency, dop)

class TranslucentSpheroid(TranslucentObject):
    def __init__(self, frame, radii, attenuation=0.01, scatteringEfficiency=0.1, dop=0):
        super().__init__(attenuation=attenuation, scatteringEfficiency=scatteringEfficiency, dop=dop)
        self._primitive = Spheroid(frame, radii)
        self._description = 'TranslucentSphere'

    def pathTransmittance(self, pathLength, pathStart, pathEnd):
        """Returns the fraction of light that hits the medium when travelling a length through it."""
        # coefficient must be negative
        coefficient = -1e-5
        #scaler = 1 - (((0.5 * (pathStart + pathEnd)) - self._massCentre).length * self._primitive.inverseBulkRadius)
        return math.exp(pathLength * self._attenuation * coefficient)

    @property
    def massCentre(self):
        return self._massCentre

    @massCentre.setter
    def massCentre(self, newMassCentre):
        self._massCentre = newMassCentre

class TranslucentCone(TranslucentObject):
    def __init__(self, frame, angle, length, attenuation=0.01, scatteringEfficiency=0.1, dop=0):
        super().__init__(attenuation=attenuation, scatteringEfficiency=scatteringEfficiency, dop=dop)
        self._primitive = Cone(frame, angle, length)
        self._description = 'TranslucentCone'
        # constant needed for path cross section calculations:
        self._tanAng = math.tan(self._primitive._angle)

    def pathTransmittance(self, pathLength, pathStart, pathEnd):
        """Returns the fraction of light that hits the medium when travelling a length through it."""
        # coefficient must be negative
        coefficient = -3
        r = (pathStart - self.position).length
        radialScale = max(1 - ((2.5 * r * self._primitive.lengthInv) ** 2), 0)
        n = self._primitive.surfaceNormal(pathStart)
        tangentialScale = 1 - (n.angleWith(pathStart - pathEnd) * 0.6366)
        return math.exp(self._attenuation * coefficient * radialScale * tangentialScale)

    def phaseFunction(self, phaseAngle):
        """Returns the fraction of inward intensity that is scattered at the input phase angle, and its degree of polarisation."""
        # currently assume cones to be gas, so use Ohman formula:
        dop = self._dop * (math.sin(phaseAngle) ** 2) / (1 + self._dop * (math.cos(phaseAngle) ** 2))
        return (1 * self._scatteringEfficiency, dop)

class ShadowRegion(TranslucentObject):
    def __init__(self, primitive):
        super().__init__()
        self._primitive = primitive
        self._opaque = False
        self._description = 'ShadowRegion'

class EmissionObject(OpaqueObject):
    def __init__(self, flux, castsShadow=False, textureMap=None):
        self._flux = flux
        self._castsShadow = castsShadow
        super().__init__(textureMap=textureMap)
        self._description = 'EmissionObject[base class]'

    def reflect(self, point, incidentRay, reflectionDirection):
        reflectedRay = LightRay(point, reflectionDirection)
        reflectedRay.intensity = self._localFlux(point)
        return reflectedRay 

    def _localFlux(self, point):
        if self._textureMap is None:
            return self._flux
        s = self._textureSize
        x, y = self.textureCoordinate(point)
        return self._textureMap.getpixel((int(x * s[0]), int(y * s[1])))[0] * self._flux * 0.003922

    def occludes(self, viewpoint, target, bias=1e-6):
        if not self._castsShadow:
            return False
        return super().occludes(viewpoint, target, bias=bias)

    def textureCoordinate(self, point):
        """Returns the fractional pixel coordinate of a texture image from a point in world space."""
        _, azimuth, polar = self.frame.fromWorldToSpherical(point)
        return (azimuth * 0.15915, polar * 0.31830)

class EmissionSpheroid(EmissionObject):
    def __init__(self, frame, radii, flux, textureMap=None):
        super().__init__(flux, textureMap=textureMap)
        self._primitive = Spheroid(frame, radii)
        self._description = 'EmissionSpheroid'

class UniformVolume(SceneObject):
    def __init__(self, primitive, extinction, scatAlbedo):
        super().__init__()
        self._primitive = primitive
        self._extinction = extinction
        self._scatAlbedo = scatAlbedo
        self._opaque = False

    @classmethod
    def spheroid(cls, frame, radii, extinction):
        primitive = Spheroid(frame, radii)
        return cls(primitive, extinction)

    @classmethod
    def sphere(cls, centre, radius, extinction, scatAlbedo):
        frame = Frame.principal(centre)
        primitive = Spheroid(frame, (radius, radius, radius))
        return cls(primitive, extinction, scatAlbedo)

    def intersect(self, ray):
        hits = self._primitive.intersect(ray)
        if not hits:
            return [], []
        for hit in hits:
            hit['object'] = self
        #
        interaction = dict()
        interaction['object'] = self
        interaction['depth'] = hits[0]['depth']
        interaction['startPoint'] = hits[0]['position']
        try:
            interaction['endPoint'] = hits[1]['position']
        except:
            # print(hits)
            return [], []
        interaction['length'] = (interaction['endPoint'] - interaction['startPoint']).length
        return hits, [interaction]

    def transmit(self, pathStart, pathEnd, pathLength, inI, scatI, scatPhase):
        phaseFactor = self.phaseFunction(scatPhase)
        if type(self._extinction) is not np.ndarray:
            transmittance = math.exp(-self._extinction * pathLength)
        else:
            transmittance = np.exp(-self._extinction * pathLength)
        return transmittance * inI + self._scatAlbedo * phaseFactor * (1 - transmittance) * scatI

class _SegmentVolume(SceneObject):
    def __init__(self, refPoint, scale, extinction, scatAlbedo):
        self._refPoint = refPoint
        self._scale = scale
        self._extinction = extinction
        self._segCount = len(extinction)
        self._scaleFactor = self._segCount / scale
        self._scatAlbedo = scatAlbedo
        super().__init__()
        self._opaque = False

    def _arrayMap(self, point):
        # maps a point in world space to an index for segments array, based on its radial distance from a reference
        return int(((point - self._refPoint).length - 1e-6) * self._scaleFactor)

    def intersect(self, ray):
        hits = self._primitive.intersect(ray)
        if not hits:
            return [], []
        for hit in hits:
            hit['object'] = self
        #
        interaction = dict()
        interaction['object'] = self
        interaction['depth'] = hits[0]['depth']
        interaction['startPoint'] = hits[0]['position']
        try:
            interaction['endPoint'] = hits[1]['position']
        except:
            # print(hits)
            return [], []
        interaction['length'] = (interaction['endPoint'] - interaction['startPoint']).length
        return hits, [interaction]

    def transmit(self, pathStart, pathEnd, pathLength, inI, scatI, scatPhase):
        # returns the intensity emerging along a line of sight's path
        m1 = self._arrayMap(pathStart)
        m2 = self._arrayMap(pathEnd)
        if m2 < m1:
            pathExtinction = np.flip(self._extinction[m2:m1 + 1])
            dl = pathLength / (1 + m1 - m2)
        else:
            pathExtinction = self._extinction[m1:m2 + 1]
            dl = pathLength / (1 + m2 - m1)
        transmittance = np.exp(-1 * dl * np.sum(pathExtinction, 0))
        phaseFactor = self.phaseFunction(scatPhase)
        scatterance = dl * np.sum(pathExtinction * np.exp(-1 * dl * np.cumsum(pathExtinction, 0)), 0)
        try:
            return (transmittance * inI) + (self._scatAlbedo * phaseFactor * scatterance * scatI)
        except:
            print(scatterance)
            1/0

class LinearSegmentVolume(_SegmentVolume):
    def __init__(self, refPoint, length, axis, segments, primitive, scatAlbedo):
        super().__init__(refPoint, length, segments, scatAlbedo)
        self._axis = axis.norm
        self._primitive = primitive

    @classmethod
    def roundedCylinder(cls, frame, radius, length, segments, scatAlbedo):
        primitive = RoundedCylinder(frame, radius, length)
        return cls(frame.origin, length, frame.w, segments, primitive, scatAlbedo)

    @classmethod
    def cone(cls, frame, angle, length, segments, scatAlbedo):
        primitive = Cone(frame, angle, length)
        return cls(frame.origin, length, frame.w, segments, primitive, scatAlbedo)

    @classmethod
    def roundedCone(cls, frame, angle, radius, length, segments, scatAlbedo):
        primitive = RoundedCone(frame, angle, length, radius)
        return cls(frame.origin, length, frame.w, segments, primitive, scatAlbedo)

    def _arrayMap(self, point):
        # maps a point in world space to an index for segments array, based on its projected distance from a reference
        return int(abs((point - self._refPoint).dot(self._axis)) * self._scaleFactor)

class RadialSegmentVolume(_SegmentVolume):
    def __init__(self, primitive, refPoint, radius, extinction, scatAlbedo):
        super().__init__(refPoint, radius, extinction, scatAlbedo)
        self._primitive = primitive
        self._testPoint = primitive.frame.origin + radius * primitive.frame.v

    @classmethod
    def cone(cls, tip, axis, angle, length, extinction, scatAlbedo):
        primitive = Cone(Frame.withW(axis, tip), angle, length)
        return cls(primitive, tip, length, extinction, scatAlbedo)

    @classmethod
    def sphere(cls, centre, radius, extinction, scatAlbedo):
        primitive = Spheroid(Frame.principal(centre).rotateInPlace(Vec3.k(), 0.79), (radius, radius, radius))
        return cls(primitive, centre, radius, extinction, scatAlbedo)

    @classmethod
    def spheroid(cls, frame, spheroidRadii, radius, extinction, scatAlbedo):
        primitive = Spheroid(frame, spheroidRadii)
        return cls(primitive, frame.origin, radius, extinction, scatAlbedo)

    @classmethod
    def cylinder(cls, centre, axis, radius, length, extinction, scatAlbedo):
        axis = axis.norm
        frame = Frame.withW(axis, origin=centre - (0.5 * length * axis))
        primitive = Cylinder(frame, radius, length)
        return cls(primitive, centre, radius, extinction, scatAlbedo)

    def pathTransmittance(self, pathLength, pathStart, pathEnd):
        mStart = self._arrayMap(pathStart)
        mEnd = self._arrayMap(pathEnd)
        pathDir = (pathEnd - pathStart).norm
        refDir = self._refPoint - pathStart
        # get directional distance along path from pathStart to point of closest approach:
        distToCA = pathDir.dot(refDir)
        # geometry 1: closest approach is between start and end of path
        if 0 <= distToCA <= pathLength:
            #return 0
            # get path's closest approach distance to refPoint:
            ca = pathStart + distToCA * pathDir
            mCA = self._arrayMap(ca)
            #dL = distToCA / (1 + mStart - mCA)
            dL = pathLength / (2 + mStart + mEnd - mCA - mCA)
            integratedCrossSections = np.sum(self._segments[mCA:mStart + 1], 0)
            integratedCrossSections += np.sum(self._segments[mCA:mEnd + 1], 0)
            return np.exp(-1 * dL * integratedCrossSections)
        # geometry 2: closest approach is outside of path
        else:
            m1, m2 = sorted((mStart, mEnd))
            dL = pathLength / (1 + m2 - m1)
            integratedCrossSections = np.sum(self._segments[m1:m2 + 1], 0)
            return np.exp(-1 * dL * integratedCrossSections)
        mid = 0.5 * (pathStart + pathEnd)
        f = 1 - ((mid - self._testPoint).length / self._scale)
        f = max(0, f ** 2)
        return np.exp(-1 * dL * integratedCrossSections * f)

    def transmit(self, pathStart, pathEnd, pathLength, inI, scatI, scatPhase):
        # returns the intensity emerging along a line of sight's path
        mStart = self._arrayMap(pathStart)
        mEnd = self._arrayMap(pathEnd)
        pathDir = (pathEnd - pathStart).norm
        refDir = self._refPoint - pathStart
        # get directional distance along path from pathStart to point of closest approach:
        distToCA = pathDir.dot(refDir)
        # geometry 1: closest approach is between start and end of path
        if 0 <= distToCA <= pathLength:
            # get path's closest approach distance:
            ca = pathStart + distToCA * pathDir
            # from this, get the closest approach's array index:
            mCA = self._arrayMap(ca)
            pathExtinction = np.append(np.flip(self._extinction[mCA:mStart + 1]), self._extinction[mCA:mEnd + 1])
            dl = pathLength / (2 + mStart + mEnd - mCA - mCA)
        # geometry 2: closest approach is outside of path
        else:
            if mEnd < mStart:
                pathExtinction = np.flip(self._extinction[mEnd:mStart + 1])
                dl = pathLength / (1 + mStart - mEnd)
            else:
                pathExtinction = self._extinction[mStart:mEnd + 1]
                dl = pathLength / (1 + mEnd - mStart)
        nonEmptyPath = pathExtinction > 0
        transmittance = np.exp(-1 * dl * np.sum(pathExtinction[nonEmptyPath], 0))
        phaseFactor = self.phaseFunction(scatPhase)
        scatterance = dl * np.sum(pathExtinction[nonEmptyPath] * np.exp(-1 * dl * np.cumsum(pathExtinction[nonEmptyPath], 0)), 0)
        return transmittance * inI + self._scatAlbedo * phaseFactor * scatterance * scatI
