import math
import numpy as np
from ray import Ray
from vec3 import Vec3
from frame import Frame
from matplotlib import pyplot as plt

class Sensor:
    def __init__(self, columns, rows, physicalWidth, physicalHeight, frame, qe, fullWell, darkCurrent, readNoise):
        """

        :param columns: number of columns of pixels.
        :param rows: number of rows of pixels.
        :param physicalWidth: physical size of the detector along a row's length.
        :param physicalHeight: physical size of the detector along a column's length.
        :param frame: the detector's frame with origin at principal point, u axis in decreasing column direction, v axis
         in increasing row direction, w axis is principal axis.
        :param qe: quantum efficiency of detector.
        :param fullWell: detector's full well capacity in electrons.
        """
        self._resolution = (columns, rows)
        self._dimensions = (physicalWidth, physicalHeight)
        self._halfDimensions = (0.5 * physicalWidth, 0.5 * physicalHeight)
        self._frame = frame
        xAxis = frame.u.negated
        yAxis = frame.v
        self._xAxis = xAxis
        self._yAxis = yAxis
        self._topLeft = frame.origin - (0.5 * physicalWidth * xAxis) - (0.5 * physicalHeight * yAxis)
        self._topRight = self._topLeft + (physicalWidth * xAxis)
        self._bottomLeft = self._topRight + (physicalHeight * yAxis)
        self.quantumEfficiency = qe
        self.fullWell = fullWell
        self.Nd = darkCurrent
        self.Nr = readNoise

    def fromImageToWorld(self, xfract, yfract):
        return self._topLeft + (self._xAxis * self.width * xfract) + (self._yAxis * self.height * yfract)

    def fromWorldToImage(self, point, cropToImageBoundaries=True):
        point = self.frame.fromWorld(point)
        point.z = 0
        x = (0.5 * self.width) - point.x
        y = (0.5 * self.height) + point.y
        if type(point.x) is not np.ndarray:
            if cropToImageBoundaries and (x < 0 or x >= self.width or y < 0 or y >= self.height):
                return None
        return x / self.width, y / self.height

    def photonResponse(self, spectralPhotonFlux):
        # electron rate in detector due to incident spectral photon rate
        spectralElectronRate = spectralPhotonFlux * self.quantumEfficiency
        totalElectronRate = np.sum(spectralElectronRate, 2)
        return totalElectronRate

    def electronResponse(self, electronRate, integrationTime):
        # electron count, including noise, due to electron rate over integration period
        countedSignalElectrons = np.random.poisson(electronRate * integrationTime)
        countedThermalElectrons = np.random.poisson(self.Nd * integrationTime)
        totalCountedElectrons = np.random.normal(countedSignalElectrons + countedThermalElectrons, self.Nr)
        return np.clip(totalCountedElectrons, 0, self.fullWell)

    def digitiseElectronResponse(self, electronResponse, bitDepth=8):
        normalisedPixelResponse = electronResponse / self.fullWell
        levels = 2 ** bitDepth
        return np.round(normalisedPixelResponse * (levels - 1), 0)

    def integrateFlux(self, spectralPhotonFlux, integrationTime):
        electronRate = self.photonResponse(spectralPhotonFlux)
        electronCount = self.electronResponse(electronRate, integrationTime)
        # TODO: remove hard-coding of bit depth
        digitalImage = self.digitiseElectronResponse(electronCount, bitDepth=10)
        return digitalImage

    def pixelResponse(self, electronRate, integrationTime):
        # TODO: remove this function
        countedSignalElectrons = np.random.poisson(electronRate * integrationTime)
        countedThermalElectrons = np.random.poisson(self.Nd * integrationTime)
        totalCountedElectrons = np.random.normal(countedSignalElectrons + countedThermalElectrons, self.Nr)
        return np.clip(totalCountedElectrons, 0, self.fullWell)

    def digitisePixel(self, pixelResponse, bitDepth=8):
        # TODO: remove this function
        normalisedPixelResponse = pixelResponse / self.fullWell
        levels = 2 ** bitDepth
        value = np.round(normalisedPixelResponse * (levels - 1), 0)
        return value

    @property
    def width(self):
        return self._dimensions[0]

    @property
    def height(self):
        return self._dimensions[1]

    @property
    def resolution(self):
        return self._resolution

    @property
    def principalPoint(self):
        return self._frame.origin

    @principalPoint.setter
    def principalPoint(self, newPrincipalPoint):
        self._frame.origin = newPrincipalPoint
        self._updateCorners()

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, newFrame):
        self._frame = newFrame
        self._xAxis = self.frame.u.negated
        self._yAxis = self.frame.v
        self._updateCorners()

    def _updateCorners(self):
        offset = self._halfDimensions
        self._topLeft = self._frame.origin - (offset[0] * self._xAxis) - (offset[1] * self._yAxis)
        self._topRight = self._topLeft + (self.width * self._xAxis)
        self._bottomLeft = self._topLeft + (self.height * self._yAxis)

class Sensor_old:
    # TODO: remove this deprecated class
    def __init__(self, pixelsX, pixelsY, lengthX, lengthY, origin, axes, QE, fullWell):
        self._resolution = (pixelsX, pixelsY)
        self._dimensions = (lengthX, lengthY)
        self._origin = origin
        xAxis = axes[0].norm
        yAxis = axes[1].norm
        self._principalPoint = origin + (0.5 * lengthX * xAxis) + (0.5 * lengthY * yAxis)
        self._topRightCorner = origin + (lengthX * xAxis)
        self._bottomLeftCorner = origin + (lengthY * yAxis)
        self._xAxis = axes[0]
        self._yAxis = axes[1]
        self.quantumEfficiency = QE
        self.fullWell = fullWell
        self.readNoise = 15

    def fromImageToWorld(self, xfract, yfract):
        return self._origin + (self._xAxis * xfract * self.width) + (self._yAxis * yfract * self.height)

    def pixelValue(self, photonCount, noise=False):
        if noise:
            if photonCount < 1e6:
                photonCount += np.random.poisson(photonCount ** 0.5)
        electrons = self.quantumEfficiency * photonCount
        if noise:
            electrons += np.random.normal(self.readNoise)
        if electrons > self.fullWell:
            electrons = self.fullWell
        return int(255 * (electrons / self.fullWell))

    @property
    def pixelsX(self):
        return self._resolution[0]

    @property
    def pixelsY(self):
        return self._resolution[1]

    @property
    def resolution(self):
        return self._resolution

    @property
    def width(self):
        return self._dimensions[0]

    @property
    def height(self):
        return self._dimensions[1]

    @property
    def origin(self):
        return self._origin

    @property
    def principalPoint(self):
        return self._principalPoint

    @principalPoint.setter
    def principalPoint(self, newPrincipalPoint):
        r = self._origin - self._principalPoint
        self._principalPoint = newPrincipalPoint
        self._origin = self._principalPoint + r
        self._topRightCorner = self._origin + (self.width * self._xAxis)
        self._bottomLeftCorner = self._origin + (self.height * self._yAxis)

    @property
    def topRight(self):
        return self._topRightCorner

    @property
    def bottomLeft(self):
        return self._bottomLeftCorner

    @property
    def axes(self):
        return (self._xAxis, self._yAxis)

    @axes.setter
    def axes(self, newAxes):
        self._xAxis, self._yAxis = newAxes
        self._origin = self._principalPoint - (0.5 * self.width * self._xAxis) - (0.5 * self.height * self._yAxis)
        self._topRightCorner = self._origin + (self.width * self._xAxis)
        self._bottomLeftCorner = self._origin + (self.height * self._yAxis)

class Polariser:
    # TODO: remove deprecated this class (unused i think)
    def __init__(self, frame):
        self._frame = frame

    def transmit(self, ray):
        """Returns the fraction of the input ray's intensity that transmits through the polariser."""
        polarised = ray.polarisation
        if polarised == 0:
            return 0.5
        unpolarised = 1 - polarised
        transmissionAxis = ray.direction.cross(self.axis.cross(ray.direction)).norm
        #offset = transmissionAxis.signedAngleWith(ray.polarisationAxis, ray.direction)
        offset = transmissionAxis.angleWith(ray.polarisationAxis)
        return (0.5 * unpolarised) + (polarised * (math.cos(offset) ** 2))

    @property
    def axis(self):
        return self._frame.v

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, newFrame):
        self._frame = newFrame

def tempSNR(eRate, intTime, Nd, Nr):
    N = eRate * intTime
    return N / ((N + (Nd * intTime) + (Nr * Nr)) ** 0.5)

def captPower(rad, filterWidth, IFOV, apDiam, opticsTrans):
    aperture = 3.14159 * ((0.5 * apDiam) ** 2)
    pixSolidAng = IFOV ** 2
    power = rad * opticsTrans * aperture * pixSolidAng * filterWidth
    return power

def pixPowerToElect(power, filterCent, qe):
    photonEnergy = 6.626e-34 * 3e8 / (filterCent * 1e-9)
    electronRate = qe * power / photonEnergy
    return electronRate

s = Sensor(2048, 2048, 1, 1, Frame.universal(), 0.6, 13500, 10, 13)

'''rad = 1e-5
pow1 = captPower(rad, 100, 1.53e-3, 1.1e-3, 0.6)
e1 = pixPowerToElect(pow1, 600, 0.6)
snr1 = tempSNR(e1, 1e-3, 10, 13)
print(snr1, e1 * 1e-3)'''

'''minIFOV = 3.142 / 2048
spinT = 4
rad = np.linspace(1e-9, 1e-3, 10000)
fig, ax1 = plt.subplots()
ax1.set_xlabel('scene radiance W / (m^2 sr nm)')
ax1.set_ylabel('pixel SNR')
IFOV = minIFOV
t = spinT * IFOV / 6.283
pow = captPower(rad, 200, IFOV, 1.1e-3, 0.8)
e = pixPowerToElect(pow, 612, 0.6)
snr = tempSNR(e, t, 10, 13)
ax1.plot(rad, snr, color='#414286', label='1')
IFOV = 2 * minIFOV
t = spinT * IFOV / 6.283
pow = captPower(rad, 200, IFOV, 1.1e-3, 0.8)
e = pixPowerToElect(pow, 612, 0.6)
snr = tempSNR(e, t, 10, 13)
ax1.plot(rad, snr, '#2A768E', label='1/2')
IFOV = 3 * minIFOV
t = spinT * IFOV / 6.283
pow = captPower(rad, 200, IFOV, 1.1e-3, 0.8)
e = pixPowerToElect(pow, 612, 0.6)
snr = tempSNR(e, t, 10, 13)
ax1.plot(rad, snr, '#20A585', label='1/3')
IFOV = 4 * minIFOV
t = spinT * IFOV / 6.283
pow = captPower(rad, 200, IFOV, 1.1e-3, 0.8)
e = pixPowerToElect(pow, 612, 0.6)
snr = tempSNR(e, t, 10, 13)
ax1.plot(rad, snr, '#71CE56', label='1/4')
ax1.legend()
plt.show()'''

#plt.plot(rad, snr)
#plt.show()
