import os
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt


class SpectralDensityCurve:
    __array_priority__ = 1000
    wavelengths = np.linspace(400, 800, 401)  # nm

    def __init__(self, curve):
        self._curve = curve  # nm^-1

    @classmethod
    def zero(cls):
        return cls(np.zeros_like(cls.wavelengths))

    @classmethod
    def uniform(cls, value):
        return cls(np.full_like(cls.wavelengths, value))

    @classmethod
    def fromExpression(cls, expression):
        curve = expression(SpectralDensityCurve.wavelengths)
        return cls(curve)

    @classmethod
    def load(cls, filename):
        path = '../data/resources/processed/spectra/np_arrays/{0}.npy'.format(filename)
        curve = np.load(path)
        return cls(curve)

    @classmethod
    def fromImage(cls, image, minY, maxY, ylog=False):
        yRange = maxY - minY
        curve = np.empty(401)
        image = Image.open('../data/resources/processed/spectra/images/{0}.png'.format(image)).convert('RGB')
        columns, rows = image.size
        if ylog:
            yVals = np.flip(np.geomspace(minY, maxY, rows))
        else:
            yVals = np.flip(np.linspace(minY, maxY, rows))
        for sample in range(401):
            fract = sample / 401
            column = int(fract * columns)
            for row in range(rows):
                pixel = image.getpixel((column, row))
                if pixel[0] == 0:
                    curve[sample] = maxY - yRange * row / rows
                    curve[sample] = yVals[row]
                    break
        return cls(curve)

    def valueAtWavelength(self, wavelength):
        if not 400 <= wavelength <= 800:
            return 0
        n = wavelength - 400
        return self._curve[n]

    def clipToWavelengths(self, minWavelength, maxWavelength, clipValue=0):
        minWavelength = int(max(minWavelength, 400))
        maxWavelength = int(min(maxWavelength, 800))
        minN = minWavelength - 400
        maxN = 1 + maxWavelength - 400
        curve = np.copy(self._curve)
        curve[:minN] = clipValue
        curve[maxN:] = clipValue
        return SpectralDensityCurve(curve)

    def display(self, xscale=None, yscale=None):
        plt.plot(self.wavelengths, self.curve)
        plt.xlim(400, 800)
        if xscale is not None:
            plt.xscale(xscale)
        if yscale is not None:
            plt.yscale(yscale)
        plt.show()

    def save(self, filename):
        path = '../data/resources/processed/spectra/np_arrays/{0}.npy'.format(filename)
        np.save(path, self.curve)

    @property
    def curve(self):
        return self._curve

    @property
    def integrated(self):
        return sum(self._curve)

    @property
    def normalised(self):
        return SpectralDensityCurve(self.curve / np.max(self.curve))

    def __neg__(self):
        return SpectralDensityCurve(-self._curve)

    def __add__(self, other):
        return SpectralDensityCurve(self._curve + other.curve)

    def __sub__(self, other):
        return SpectralDensityCurve(self._curve - other.curve)

    def __mul__(self, other):
        return SpectralDensityCurve(self._curve * other.curve)

    def __rmul__(self, other):
        return SpectralDensityCurve(other * self._curve)

    def __truediv__(self, divisor):
        return SpectralDensityCurve(self._curve / divisor)

    def __pow__(self, power):
        return SpectralDensityCurve(self._curve ** power)

class PhotonRateSDC(SpectralDensityCurve):
    _c = 2.998e8
    _h = 6.626e-34
    _k = 1.381e-23
    photonPerEnergy = SpectralDensityCurve.wavelengths / _h / _c / 1e9  # J^-1

    def __init__(self, spectralPhotonRate):
        """

        :param spectralPhotonRate: numpy array spectral photon rate curve in units photons s^-1 m^-2 sr^-1 nm^-1
        """
        super().__init__(spectralPhotonRate)

    @classmethod
    def allZeros(cls):
        """

        :return: PhotonRateSDC object with a spectral curve of all zeros
        """
        return cls(np.zeros_like(cls.wavelengths))

    @classmethod
    def fromSpectralRadiance(cls, spectralRadiance):
        """

        :param spectralRadiance: numpy array spectral radiance curve in units W m^-2 sr^-1 nm^-1
        :return: PhotonRateSDC object with a photon rate curve derived from the input radiance curve
        """
        spectralPhotonRate = spectralRadiance * PhotonRateSDC.photonPerEnergy  # photons s^-1 m^-2 sr^-1 nm^-1
        return cls(spectralPhotonRate)

    @classmethod
    def fromExpression(cls, expression):
        spectralRadiance = expression(SpectralDensityCurve.wavelengths)  # W m^-2 sr^-1 nm^-1
        return cls.fromSpectralRadiance(spectralRadiance)

    @classmethod
    def blackbody(cls, T):
        w = cls.wavelengths * 1e-9
        c = cls._c
        h = cls._h
        k = cls._k
        spectralRadiance = 1e-9 * 2 * h * c ** 2 / ((w ** 5) * (np.exp(h * c / (w * k * T)) - 1))  # W m^-2 sr^-1 nm^-1
        #print(np.sum(spectralRadiance))
        return cls.fromSpectralRadiance(spectralRadiance)

    @classmethod
    def solar(cls):
        # for comparison, solar spectrum with absorption features:
        # https: // www.nrel.gov / grid / solar - resource / spectra - astm - e490.html
        return cls.blackbody(5778)

    def clipToWavelengths(self, minWavelength, maxWavelength, clipValue=0):
        minWavelength = int(max(minWavelength, 400))
        maxWavelength = int(min(maxWavelength, 800))
        minN = minWavelength - 400
        maxN = 1 + maxWavelength - 400
        curve = np.copy(self._curve)
        curve[:minN] = 0
        curve[maxN:] = 0
        return PhotonRateSDC(curve)

    @property
    def toSpectralRadiance(self):
        return SpectralDensityCurve(self.curve / PhotonRateSDC.photonPerEnergy)

    @property
    def totalPhotons(self):
        return sum(self._curve)

    @property
    def normalised(self):
        return PhotonRateSDC(self.curve / np.max(self.curve))

    @property
    def hasZeroPhotons(self):
        return not np.any(self.curve)

    def __add__(self, other):
        return PhotonRateSDC(self._curve + other.curve)

    def __mul__(self, other):
        return PhotonRateSDC(self._curve * other.curve)

    def __rmul__(self, other):
        return PhotonRateSDC(other * self._curve)

    def __truediv__(self, other):
        return PhotonRateSDC(self._curve / other)

#t = SpectralDensityCurve.fromImage('comet_lines_rel_flux', 0, 1)
#t.save('comet_lines_rel_flux')
#t = SpectralDensityCurve.load('comet_lines_rel_flux')
#t.display()

#q = SpectralDensityCurve.load('cmv4000_qe')
#p = PhotonRateSDC.solar()
#p.display()
#e = p * q
#e.display()

#plt.plot(q.wavelengths, p.curve, q.wavelengths, e.curve)
#plt.show()

"""fig, ax1 = plt.subplots()
ax1.set_xlabel('wavelength [nm]')
ax1.set_ylabel('spectral flux [s^-1 nm^-1]')
ax1.plot(q.wavelengths, p.curve, 'c', label='incident photons')
ax1.plot(q.wavelengths, e.curve, 'b', label='generated electrons')
ax1.legend()
ax2 = ax1.twinx()
ax2.set_ylabel('quantum efficiency')
ax2.set_ylim((0, 1))
ax2.plot(q.wavelengths, q.curve, 'r--', label='quantum efficiency')
ax2.legend(bbox_to_anchor=(1, 0.1))
plt.show()"""