import math
import numpy as np


class Vec3:
    def __init__(self, xyz: tuple) -> None:
        self._x = xyz[0]
        self._y = xyz[1]
        self._z = xyz[2]
        self._length = None
        self._norm = None
        self._reciprocated = None

    @classmethod
    def zeroVector(cls):
        return cls((0, 0, 0))

    @classmethod
    def i(cls):
        return cls((1, 0, 0))

    @classmethod
    def j(cls):
        return cls((0, 1, 0))

    @classmethod
    def k(cls):
        return cls((0, 0, 1))

    @classmethod
    def perpendicularTo(cls, v):
        """Returns a normalised arbitrary vector perpendicular to this one."""
        try:
            z = (-v.x - v.y) / v.z
            return cls((1, 1, z)).norm
        except ZeroDivisionError:
            try:
                x = (-v.z - v.y) / v.x
                return cls((x, 1, 1)).norm
            except ZeroDivisionError:
                try:
                    y = (-v.x - v.z) / v.y
                    return cls((1, y, 1)).norm
                except ZeroDivisionError:
                    return cls.zeroVector()

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, newX):
        self._x = newX
        self.expire()

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, newY):
        self._y = newY
        self.expire()

    @property
    def z(self):
        return self._z

    @z.setter
    def z(self, newZ):
        self._z = newZ
        self.expire()

    @property
    def length(self):
        if self._length is None:
            self._length = self.dot(self) ** 0.5
        return self._length

    @property
    def norm(self):
        if self._norm is None:
            try:
                self._norm = self * (1 / self.length)
            except ZeroDivisionError:
                self._norm = Vec3.zeroVector()
        return self._norm

    @property
    def tup(self):
        return (self.x, self.y, self.z)

    @property
    def negated(self):
        return -1 * self

    @property
    def reciprocated(self):
        if not self._reciprocated:
            x = 1 / self._x if self._x != 0 else 1e20
            y = 1 / self._y if self._y != 0 else 1e20
            z = 1 / self._z if self._z != 0 else 1e20
            self._reciprocated = Vec3((x, y, z))
        return self._reciprocated

    @property
    def isZero(self):
        if self.x == 0 and self.y == 0 and self.z == 0:
            return True
        return False

    @property
    def onPlaneXY(self):
        return Vec3((self.x, self.y, 0))

    @property
    def onPlaneXZ(self):
        return Vec3((self.x, 0, self.z))

    @property
    def onPlaneYZ(self):
        return Vec3((0, self.y, self.z))

    def expire(self):
        self._length = None
        self._norm = None
        self._reciprocated = None

    def dot(self, other: 'Vec3') -> float:
        """
        Returns the dot product of this vector and the input vector.

        :param other: vector to dot with this vector
        :return: scalar dot product
        """
        return self._x * other.x + self._y * other.y + self._z * other.z

    def cross(self, other: 'Vec3') -> 'Vec3':
        """
        Returns the cross product of this vector and the input vector.

        :param other: vector (RHS) to cross with this vector (LHS)
        :return: vector cross product
        """
        a = self._y * other.z - self._z * other.y
        b = self._z * other.x - self._x * other.z
        c = self._x * other.y - self._y * other.x
        return Vec3((a, b, c))

    def hadmul(self, other: 'Vec3') -> 'Vec3':
        """
        Returns the elementwise Hadarmard multiplication of this vector and the input vector.

        :param other: vector to elementwise multiply with this vector
        :return: Hadamard product
        """
        return Vec3((self.x * other.x, self.y * other.y, self.z * other.z))

    def rotate(self, around: 'Vec3', by: float) -> 'Vec3':
        """Returns this vector rotated anticlockwise around the input vector by given amount (in radians).

        :param around: axis of anticlockwise rotation
        :param by: angle of rotation (radians)
        :return: the rotated vector
        """
        axis = around.norm
        angle = by
        W = Mat3((0, -1 * axis.z, axis.y), (axis.z, 0, -1 * axis.x), (-1 * axis.y, axis.x, 0))
        T1 = math.sin(angle) * W
        #T2 = 2 * (math.sin(angle * 0.5) ** 2) * W.multM3(W)
        T2 = 2 * (math.sin(angle * 0.5) ** 2) * W * W
        #M = W.Identity + T1 + T2
        M = Mat3.identity() + T1 + T2
        #a = M.x00 * self._x + M.x01 * self._y + M.x02 * self._z
        #b = M.x10 * self._x + M.x11 * self._y + M.x12 * self._z
        #c = M.x20 * self._x + M.x21 * self._y + M.x22 * self._z
        #return Vec3((a, b, c))
        return M * self

    def angleWith(self, other: 'Vec3') -> float:
        """Returns inner angle of this vector with another.

        :param other: the vector forming an angle with this vector
        :return: the inner angle between the two vectors
        """
        if type(self.x) is np.ndarray or type(other.x) is np.ndarray:
            dotResult = self.norm.dot(other.norm)
            dotResult[dotResult < -1] = -1
            dotResult[dotResult > 1] = 1
            return np.arccos(dotResult)
        else:
            try:
                return math.acos(self.norm.dot(other.norm))
            except:
                return 0

    def signedAngleWith(self, otherVector, rotationAxis):
        """Returns the angle, from -pi to pi, between this vector and another, measured around mutually perpendicular rotationAxis."""
        b = otherVector
        n = rotationAxis
        rotationAxis = rotationAxis.norm
        dot = self.dot(otherVector)
        det = self._x * b.y * n.z + self._z + b.x * n.y + self._y * b.z * n.x - self._z * b.y * n.x - self._x * b.z * n.y - self._y * b.x * n.z
        try:
            return math.atan2(det, dot)
        except TypeError:
            return np.arctan2(det, dot)

    def anticlockAngleWith(self, otherVector, rotationAxis):
        """Returns the anticlockwise angle, from 0 to 2pi, from this vector to another, measured around mutually perpendicular rotationAxis."""
        angle = self.signedAngleWith(otherVector, rotationAxis)
        if type(angle) is not np.ndarray:
            if angle < 0:
                angle = 2 * math.pi + angle
        else:
            angle[angle < 0] = 2 * math.pi + angle[angle < 0]
        return angle

    def parallelWith(self, otherVector, epsilon=1e-7):
        return self.angleWith(otherVector) < epsilon

    def antiparallelWith(self, otherVector, epsilon=1e-7):
        return self.angleWith(otherVector) > math.pi - epsilon

    def projectOnto(self, otherVector):
        return otherVector * self.dot(otherVector) / (otherVector.length ** 2)

    def projectPerpTo(self, otherVector):
        return self - self.projectOnto(otherVector)

    def intersectWith(self, other, selfOrigin, otherOrigin):
        """DEPRECATED (MOVE TO RAY CLASS)?"""
        n1 = self.norm
        n2 = other.norm
        p1 = selfOrigin
        p2 = otherOrigin
        b = p1 - p2
        ax1 = n1.cross(b)
        ax2 = n2.cross(b)
        if ax1.cross(ax2).isZero:
            return None
        theta1 = n1.angleWith(b)
        theta2 = n2.angleWith(b)
        theta3 = math.pi - (theta1 + theta2)
        l1 = math.sin(theta2) * (b.length / math.sin(theta3))
        return p1 + (l1 * n1)

    def multM3asCol_deprecated(self, matrix):
        """Returns the result of matrix * vector, with vector in column form."""
        M = matrix
        a = (M.x00 * self._x) + (M.x01 * self._y) + (M.x02 * self._z)
        b = (M.x10 * self._x) + (M.x11 * self._y) + (M.x12 * self._z)
        c = (M.x20 * self._x) + (M.x21 * self._y) + (M.x22 * self._z)
        return Vec3((a, b, c))

    def npMask(self, mask):
        return Vec3((self.x[mask], self.y[mask], self.z[mask]))

    def __eq__(self, other):
        return self._x == other.x and self._y == other.y and self._z == other.z

    def __ne__(self, other):
        return self._x != other.x or self._y != other.y or self._z != other.z

    def __abs__(self):
        return self.length

    def __add__(self, other):
        return Vec3((self._x + other.x, self._y + other.y, self._z + other.z))

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return Vec3((self._x - other.x, self._y - other.y, self._z - other.z))

    def __rsub__(self, other):
        return Vec3((other.x - self._x, other.y - self._y, other.z - self._z))

    def __mul__(self, other):
        return Vec3((self._x * other, self._y * other, self._z * other))

    def __rmul__(self, other):
        return Vec3((self._x * other, self._y * other, self._z * other))

    def __truediv__(self, other):
        return Vec3((self._x / other, self._y / other, self._z / other))

    def __setitem__(self, index, value):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        elif index == 2:
            self.z = value
        else:
            raise IndexError('vector index out of range')

    def __getitem__(self, index):
        try:
            return self.tup[index]
        except:
            raise IndexError('vector index out of range')

    def __str__(self):
        return str((self._x, self._y, self._z))

    def __repr__(self):
        return 'Vec3{0}'.format(self)


class Vec4:
    def __init__(self, values):
        self.x0 = values[0]
        self.x1 = values[1]
        self.x2 = values[2]
        self.x3 = values[3]

    @classmethod
    def zeroVector(cls):
        return cls((0, 0, 0, 0))

    def __add__(self, other):
        return Vec4((self.x0 + other.x0, self.x1 + other.x1, self.x2 + other.x2, self.x3 + other.x3))

    def __mul__(self, other):
        return Vec4((self.x0 * other, self.x1 * other, self.x2 * other, self.x3 * other))

    def __rmul__(self, other):
        return self * other


class Mat3:
    def __init__(self, row1, row2, row3):
        if row1 is None:
            row1 = (0, 0, 0)
        if row2 is None:
            row2 = (0, 0, 0)
        if row3 is None:
            row3 = (0, 0, 0)
        self.x00 = row1[0]
        self.x01 = row1[1]
        self.x02 = row1[2]
        self.x10 = row2[0]
        self.x11 = row2[1]
        self.x12 = row2[2]
        self.x20 = row3[0]
        self.x21 = row3[1]
        self.x22 = row3[2]

    @classmethod
    def identity(cls):
        return cls((1, 0, 0), (0, 1, 0), (0, 0, 1))

    @classmethod
    def fromColumns(cls, column1, column2, column3):
        row1 = (column1[0], column2[0], column3[0])
        row2 = (column1[1], column2[1], column3[1])
        row3 = (column1[2], column2[2], column3[2])
        return cls(row1, row2, row3)

    @property
    def determ(self):
        return (self.x00 * ((self.x11 * self.x22) - (self.x12 * self.x21))) - (self.x01 * ((self.x10 * self.x22) - (self.x12 * self.x20))) + (self.x02 * ((self.x10 * self.x21) - (self.x11 * self.x20)))

    @property
    def transpose(self):
        top = (self.x00, self.x10, self.x20)
        middle = (self.x01, self.x11, self.x21)
        bottom = (self.x02, self.x12, self.x22) 
        return Mat3(top, middle, bottom)

    @property
    def inverse(self):
        a = np.array([[self.x00, self.x01, self.x02], [self.x10, self.x11, self.x12], [self.x20, self.x21, self.x22]])
        b = np.linalg.inv(a)
        return Mat3(b[0], b[1], b[2])

    @property
    def row1(self):
        return self.x00, self.x01, self.x02

    @property
    def row2(self):
        return self.x10, self.x11, self.x12

    @property
    def row3(self):
        return self.x20, self.x21, self.x22

    @property
    def column1(self):
        return self.x00, self.x10, self.x20

    @property
    def column2(self):
        return self.x01, self.x11, self.x21

    @property
    def column3(self):
        return self.x02, self.x12, self.x22

    def hadmul(self, other):
        row1 = (self.x00 * other.x00, self.x01 * other.x01, self.x02 * other.x02)
        row2 = (self.x10 * other.x10, self.x11 * other.x11, self.x12 * other.x12)
        row3 = (self.x20 * other.x20, self.x21 * other.x21, self.x22 * other.x22)
        return Mat3(row1, row2, row3)

    def __add__(self, other):
        t = (self.x00 + other.x00, self.x01 + other.x01, self.x02 + other.x02)
        m = (self.x10 + other.x10, self.x11 + other.x11, self.x12 + other.x12)
        b = (self.x20 + other.x20, self.x21 + other.x21, self.x22 + other.x22)
        return Mat3(t, m, b)

    def __sub__(self, other):
        t = (self.x00 - other.x00, self.x01 - other.x01, self.x02 - other.x02)
        m = (self.x10 - other.x10, self.x11 - other.x11, self.x12 - other.x12)
        b = (self.x20 - other.x20, self.x21 - other.x21, self.x22 - other.x22)
        return Mat3(t, m, b)

    def __mul__(self, other):
        if type(other) is Vec3:
            a = (self.x00 * other.x) + (self.x01 * other.y) + (self.x02 * other.z)
            b = (self.x10 * other.x) + (self.x11 * other.y) + (self.x12 * other.z)
            c = (self.x20 * other.x) + (self.x21 * other.y) + (self.x22 * other.z)
            return Vec3((a, b, c))
        if type(other) is Mat3:
            M = other
            top0 = M.x00 * self.x00 + M.x01 * self.x10 + M.x02 * self.x20
            top1 = M.x00 * self.x01 + M.x01 * self.x11 + M.x02 * self.x21
            top2 = M.x00 * self.x02 + M.x01 * self.x12 + M.x02 * self.x22
            middle0 = M.x10 * self.x00 + M.x11 * self.x10 + M.x12 * self.x20
            middle1 = M.x10 * self.x01 + M.x11 * self.x11 + M.x12 * self.x21
            middle2 = M.x10 * self.x02 + M.x11 * self.x12 + M.x12 * self.x22
            bottom0 = M.x20 * self.x00 + M.x21 * self.x10 + M.x22 * self.x20
            bottom1 = M.x20 * self.x01 + M.x21 * self.x11 + M.x22 * self.x21
            bottom2 = M.x20 * self.x02 + M.x21 * self.x12 + M.x22 * self.x22
            top = (top0, top1, top2)
            middle = (middle0, middle1, middle2)
            bottom = (bottom0, bottom1, bottom2)
            return Mat3(top, middle, bottom)
        row1 = other * self.x00, other * self.x01, other * self.x02
        row2 = other * self.x10, other * self.x11, other * self.x12
        row3 = other * self.x20, other * self.x21, other * self.x22
        return Mat3(row1, row2, row3)

    def __rmul__(self, scalar):
        """Returns the result of multiplying a scalar by this matrix."""
        row1 = scalar * self.x00, scalar * self.x01, scalar * self.x02
        row2 = scalar * self.x10, scalar * self.x11, scalar * self.x12
        row3 = scalar * self.x20, scalar * self.x21, scalar * self.x22
        return Mat3(row1, row2, row3)

    def __pow__(self, power, modulo=None):
        result = self
        for _ in range(power - 1):
            result = result * result
        return result

    def __str__(self):
        r1 = [self.x00, self.x01, self.x02]
        r2 = [self.x10, self.x11, self.x12]
        r3 = [self.x20, self.x21, self.x22]
        return '[{0},\n{1},\n{2}]'.format(r1, r2, r3)

    def __repr__(self):
        return 'Mat3{0}'.format(self)


class Mat4:
    def __init__(self, row1, row2, row3, row4):
        if row1 is None:
            row1 = (0, 0, 0, 0)
        if row2 is None:
            row2 = (0, 0, 0, 0)
        if row3 is None:
            row3 = (0, 0, 0, 0)
        if row4 is None:
            row4 = (0, 0, 0, 0)
        self.x00 = row1[0]
        self.x01 = row1[1]
        self.x02 = row1[2]
        self.x03 = row1[3]
        self.x10 = row2[0]
        self.x11 = row2[1]
        self.x12 = row2[2]
        self.x13 = row2[3]
        self.x20 = row3[0]
        self.x21 = row3[1]
        self.x22 = row3[2]
        self.x23 = row3[3]
        self.x30 = row4[0]
        self.x31 = row4[1]
        self.x32 = row4[2]
        self.x33 = row4[3]

    def __mul__(self, other):
        y0 = self.x00 * other.x0 + self.x01 * other.x1 + self.x02 * other.x2 + self.x03 * other.x3
        y1 = self.x10 * other.x0 + self.x11 * other.x1 + self.x12 * other.x2 + self.x13 * other.x3
        y2 = self.x20 * other.x0 + self.x21 * other.x1 + self.x22 * other.x2 + self.x23 * other.x3
        y3 = self.x30 * other.x0 + self.x31 * other.x1 + self.x32 * other.x2 + self.x33 * other.x3
        return Vec4((y0, y1, y2, y3))

    def __str__(self):
        r1 = [self.x00, self.x01, self.x02, self.x03]
        r2 = [self.x10, self.x11, self.x12, self.x13]
        r3 = [self.x20, self.x21, self.x22, self.x23]
        r4 = [self.x30, self.x31, self.x32, self.x33]
        return '[{0},\n{1},\n{2},\n{3}]'.format(r1, r2, r3, r4)

    def __repr__(self):
        return 'Mat4{0}'.format(self)